#ifndef BORN_H
#define BORN_H 1

#include <cmath>
#include <vector>

#include "base/Operator.h"
#include "base/DataInCoreFloat.h"
#include "base/inCoreFloat.h"
#include "base/Vector.h"
#include "Grad.h"
#include "Interpolation.h"
#include "Clean.h"
typedef long long llong;

/*
  Born Modeling Operator
  
  Forward is Born Operator
  Adjoint is Imaging Operator

  First Created: 2015-11-05, Yinbin Ma
  Last Modified: 2015-11-05
*/

extern "C" void AddISPC(float *pin,float *pout,int n2);
extern "C" void ScaleISPC(float *pin,float *pout,int n2);
extern "C" void ProdISPC(float *coef, float *pin,float *pout,int n2);
extern "C" void ProdToISPC(float *coef, float *pin,float *pout,int n2);
extern "C" void LeapFrogISPC(float *coef,float *p1,float *px,float *pz,float *p2,float *p3,int n);
extern "C" void TimeCoreISPC(float *pre, float *cur, float *next, float *pin, float *pout, int n);
namespace SEP{

class BornEngine: public Operator{
  public:
    BornEngine(Space *mspace, Space *uspace);
    ~BornEngine(){}
    void SetPar(llong nb, float dt, float dx, llong sample);
    void SetBG(Vector *BG);
    void SetModel(Vector *MK, Vector *MR);
    
    bool FMarch(Vector *pre, Vector *cur, Vector *next);
    bool BMarch(Vector *pre, Vector *cur, Vector *next);
    
    void TimeCoreForward(bool add, Vector *model, Vector *data);
    void TimeCoreAdjoint(bool add, Vector *model, Vector *data);
    void SpaceFCoreForward(bool add, Vector *model, Vector *data);
    void SpaceFCoreAdjoint(bool add, Vector *model, Vector *data);
    void SpaceACoreForward(bool add, Vector *model, Vector *data);
    void SpaceACoreAdjoint(bool add, Vector *model, Vector *data);
    //void BornMX();
    //void BornMZ();
    virtual bool Forward(bool add, Vector *model, Vector *data);
    virtual bool Adjoint(bool add, Vector *model, Vector *data);
    bool StarForward(bool add, Vector *model, Vector *data);
    bool StarAdjoint(bool add, Vector *model, Vector *data);   
  private:
    void init();
    llong nt, nz, nx, nb, mblock,sample;
    float dx, dt, dt2,dx2, dtdx2;
    float *Dtemp;
    
    inCoreFloat *mspace, *uspace;
    Vector *MK, *VTX, *MRX, *MRZ, *MR;
    Vector *temp, *tempX, *tempZ, *tempuf, *tempub, *temp1, *temp2;
    Vector *BG;
    
    GradX *DX;
    GradZ *DZ;
    GradXT *DXT;
    GradZT *DZT;
    IntX *interpX;
    IntZ *interpZ;    
    CleanVec Zero;
        
};


class BornT: public Operator{
  public:
    BornT(Space *mspace, Space *uspace);
    ~BornT(){}
    void SetEngine(BornEngine *BornMod);
    virtual bool Forward(bool add, Vector *model, Vector *data);
    virtual bool Adjoint(bool add, Vector *model, Vector *data);    
    
  private:
    BornEngine *BornMod;

};

class BornMR: public Operator{
  public:
    BornMR(Space *mspace, Space *uspace);
    ~BornMR(){}
    void SetEngine(BornEngine *BornMod);
    virtual bool Forward(bool add, Vector *model, Vector *data);
    virtual bool Adjoint(bool add, Vector *model, Vector *data);    
    
  private:
    BornEngine *BornMod;

};
} // end of namespace
#endif
