#include <iostream>
#include <cstdlib>
#include <stdbool.h>
#include <vector>
#include <cmath>
#include<string>
#include "seplib.h"

#include <omp.h>
#include "Laplacian.h"
#include "Source.h"
#include "SensorPad.h"
#include "Grad.h"
#include "Interpolation.h"
#include "Born.h"
#include "Prop.h"
#include "Engine.h"
#include "CreateModel.h"
#include "HessInv.h"
#include "Wrapper.h"
#include "NCG.h"
#include "utilities.h"
#include "mkl_rci.h"
#include "mkl_blas.h"
#include "mkl_spblas.h"
#include "mkl_service.h"
extern "C" {
#include "sep3d.h"
}
using namespace std;
using namespace SEP;



/*
Beta version for Multiparameter FWI

Generated by Yinbin: 2015-07-06

Assume: 
	1) source term
	2) initial condition with gradient = 0
	3) K[GPa], t[ms], rho[g/cm^3]
*/

int main(int argc, char *argv[]) {
  cout << "========================================"<<endl;
  cout << "                                        "<<endl;
  cout << "========================================"<<endl;
  cout << "              Test Modulus.        "<<endl;
  initpar(argc,argv);       // Initialize the sep library
  omp_set_num_threads(16);   // Initialize the parallel environment
  
  // Define dimension of interest
  int nx, nz, nt;
  int sx, sz, ns, oxs, dxs;
  int boundary;
  float dx, dz, dt;
  int nb, bsize;
  int sample;
  int mblock;
  int num;
  int type=1;

   // Get parameters
  getch("nb","d",&nb);              cout << "	number of cores = " << nb << endl;
  getch("num","d",&num);     
  getch("bsize","d",&bsize);        cout << "	jobs on each core = " << bsize << endl;
  ns = nb*bsize;
  getch("nx","d",&nx);              cout << "	nx = " << nx << endl;
  getch("nz","d",&nz);              cout << "	nz = " << nz << endl;
  getch("nt","d",&nt);              cout << "	nt = " << nt << endl;
  getch("boundary","d",&boundary);  cout << "	boundary = " << boundary << endl;
  mblock=nx*nz;
  getch("dx","f",&dx);              cout << "	dx = " << dx <<"[m]"<< endl;
  dz = dx;
  getch("sz","d",&sz);              cout << "	sz = " << sz << endl;
  getch("oxs","d",&oxs);            cout << "	oxs = " << oxs << endl;
  getch("dxs","d",&dxs);            cout << "	dxs = " << dxs << endl;
  getch("dt","f",&dt);              cout << "	dt = " << dt << "[ms]"<< endl;
  getch("ns","d",&ns);              cout << "	ns = " << ns << endl;
  getch("sample","d",&sample);      cout << "	sample = " << sample << endl;
 
  // Set up space 
  inCoreFloat1D sspace(nt);               // space for source
  inCoreFloat2D mspace(nz,nx);            // space for velocity, reflectivity
  inCoreFloat2D bmspace(2*nz,nx);         // space for velocity, reflectivity
  inCoreFloat3D uspace(nt,nz,nx);         // space for wavefield
  inCoreFloat2D dspace(nt,nx);            // space for data
 
  
  // Set up vectors
  Vector *shot     = new Vector(&sspace);
  Vector *VT       = new Vector(&mspace); 
  Vector *RT       = new Vector(&mspace);
  Vector *VM       = new Vector(&mspace); 
  Vector *RM       = new Vector(&mspace);
  Vector *dV       = new Vector(&mspace); 
  Vector *dR       = new Vector(&mspace);
  Vector *HV       = new Vector(&mspace); 
  Vector *HR       = new Vector(&mspace);
  Vector *u1       = new Vector(&uspace);
  Vector *MKT      = new Vector(&mspace); 
  Vector *MRT      = new Vector(&mspace);
  Vector *MKM      = new Vector(&mspace); 
  Vector *MRM      = new Vector(&mspace); 
 
  // Creating pointers
  float *__restrict__ shot_val  = ((dataInCoreFloat*)shot->getDataContainer())->getFloatPtr(); 
  float *__restrict__ VT_val    = ((dataInCoreFloat*)VT->getDataContainer())->getFloatPtr(); 
  float *__restrict__ VM_val    = ((dataInCoreFloat*)VM->getDataContainer())->getFloatPtr(); 
  float *__restrict__ RT_val    = ((dataInCoreFloat*)RT->getDataContainer())->getFloatPtr(); 
  float *__restrict__ RM_val    = ((dataInCoreFloat*)RM->getDataContainer())->getFloatPtr(); 
  float *__restrict__ MKT_val   = ((dataInCoreFloat*)MKT->getDataContainer())->getFloatPtr(); 
  float *__restrict__ MKM_val   = ((dataInCoreFloat*)MKM->getDataContainer())->getFloatPtr(); 
  float *__restrict__ MRT_val   = ((dataInCoreFloat*)MRT->getDataContainer())->getFloatPtr(); 
  float *__restrict__ MRM_val   = ((dataInCoreFloat*)MRM->getDataContainer())->getFloatPtr(); 
  float *__restrict__ dV_val    = ((dataInCoreFloat*)dV->getDataContainer())->getFloatPtr(); 
  float *__restrict__ dR_val    = ((dataInCoreFloat*)dR->getDataContainer())->getFloatPtr();  
  float *__restrict__ HV_val    = ((dataInCoreFloat*)HV->getDataContainer())->getFloatPtr(); 
  float *__restrict__ HR_val    = ((dataInCoreFloat*)HR->getDataContainer())->getFloatPtr();    
  float *__restrict__ u1_val    = ((dataInCoreFloat*)u1->getDataContainer())->getFloatPtr();   
  // Read in velocity and density
  cout << "========================== Reading in velocity and density model ==========================" <<endl;  
  sreed("wavelet",shot_val,sizeof(float)*nt); 

  for(int i=0;i<mblock;i++){
    VT_val[i] = 1.5;
    VM_val[i] = 1.5;
    RT_val[i] = 1.;
    RM_val[i] = 1.;
  }


  for(int i=0;i<mblock;i++){
  MKT_val[i] = 1./(RT_val[i]*VT_val[i]*VT_val[i]);
  MKM_val[i] = 1./(RM_val[i]*VM_val[i]*VM_val[i]);  
  MRT_val[i] = 1./RT_val[i];
  MRM_val[i] = 1./RM_val[i];
  }

  int water = 0*nx;
  float *weight = new float[mblock];
  for(int iz=0;iz < nz;iz++){
    for(int ix=0;ix<nx;ix++){
      int pos = iz*nx+ix;
      weight[pos] = 1.;
    }
  }

  /*
  Models *CModel =  new Models(nz, nx, 2.0, 1.5);
  CModel->Model0(VM,RM);
  CModel->Model1(VT,RT);
  */
  //Constructing operators
  cout << "========================== Constructing operator ==========================" <<endl; 
  std::vector<int> sxv;
  std::vector<int> szv;
  for(int i=0; i < nb*bsize; i++){
    sxv.push_back(60+30*i);
    szv.push_back(60);
  }
  Engine *Run = new Engine();
  Run->SetSpace(&mspace, &dspace, &uspace);
  Run->SetShot(shot,sxv,szv);
  Run->SetPropPar(nb,bsize,boundary,dt,dx,dz,sample,water,weight);
  Run->SetTrueModel(MKT,MRT);
  
  cout << "========================== Collecting data ==========================" <<endl; 
  float *dataTrue;
  float *dataRes;  
  cout << "    True data collected.  " <<endl;
  Run->SetModel(MKM,MRM);
  Run->GradientFunc(dV,dR);  
  cout << "    Migration data collected.  " <<endl;
  srite("GK",dV_val,sizeof(float)*nz*nx);
  srite("GR",dR_val,sizeof(float)*nz*nx);


  Hess *HDot = new Hess(&mspace, &bmspace, &bmspace, Run);
  HDot->DotTest(true);


 
  cout << "========================== Hessian action on (K, rho) ==========================" <<endl; 
  cout << "    Compute perturbation  " <<endl;
  dV->zeroElement();
  dR->zeroElement();
  HV->zeroElement();
  HR->zeroElement();
  float pert = 0.01;
  for(int iz=100; iz < nz; iz+=50){
    for(int ix=100; ix < nx; ix+=100){
      int pos = iz*nx+ix;
      dV_val[pos] = 0.001;    
      dR_val[pos] = 0.;    
    }
  }
  Run->GaussHessianFunc(dV,dR,HV,HR);
  dV->zeroElement();
  dR->zeroElement();
  for(int i=0;i<mblock;i++){
    HV_val[i] = HV_val[i]*100000.;
    HR_val[i] = HR_val[i]*100000.;
  }    
  for(int i=0;i<mblock;i++){
    dV_val[i] = HV_val[i];
    dR_val[i] = HR_val[i];
  }    
  srite("KRHKK",dV_val,sizeof(float)*nz*nx);
  srite("KRHKR",dR_val,sizeof(float)*nz*nx);  

  dV->zeroElement();
  dR->zeroElement();
  HV->zeroElement();
  HR->zeroElement();
  for(int iz=100; iz < nz; iz+=50){
    for(int ix=100; ix < nx; ix+=100){
      int pos = iz*nx+ix;
      dV_val[pos] = 0. ;    
      dR_val[pos] = 0.001;    
    }
  }
  Run->GaussHessianFunc(dV,dR,HV,HR);
  dV->zeroElement();
  dR->zeroElement();
  for(int i=0;i<mblock;i++){
    HV_val[i] = HV_val[i]*100000.;
    HR_val[i] = HR_val[i]*100000.;
  }    
  for(int i=0;i<mblock;i++){
    dV_val[i] = HV_val[i];
    dR_val[i] = HR_val[i];
  }    
  srite("KRHRK",dV_val,sizeof(float)*nz*nx);
  srite("KRHRR",dR_val,sizeof(float)*nz*nx);  

  return 0;
  cout << "========================== Hessian action on (V, rho) ==========================" <<endl; 
  cout << "    Compute perturbation  " <<endl;
  dV->zeroElement();
  dR->zeroElement();
  HV->zeroElement();
  HR->zeroElement();
  for(int iz=100; iz < nz; iz+=50){
    for(int ix=120; ix < nx; ix+=100){
      int pos = iz*nx+ix;
      dV_val[pos] = -2.*0.001;    
      dR_val[pos] = 0.;    
    }
  }
  Run->HessianFunc(dV,dR,HV,HR);
  dV->zeroElement();
  dR->zeroElement();
  for(int i=0;i<mblock;i++){
    HV_val[i] = HV_val[i]*100000.;
    HR_val[i] = HR_val[i]*100000.;
  }    
  for(int i=0;i<mblock;i++){
    dV_val[i] = 2*HV_val[i];
    dR_val[i] = HV_val[i]+HR_val[i];
  }    
  srite("VRHVV",dV_val,sizeof(float)*nz*nx);
  srite("VRHVR",dR_val,sizeof(float)*nz*nx);  

  dV->zeroElement();
  dR->zeroElement();
  HV->zeroElement();
  HR->zeroElement();
  for(int iz=100; iz < nz; iz+=50){
    for(int ix=100; ix < nx; ix+=100){
      int pos = iz*nx+ix;
      dV_val[pos] = -0.001;    
      dR_val[pos] = -0.001;    
    }
  }
  Run->HessianFunc(dV,dR,HV,HR);
  dV->zeroElement();
  dR->zeroElement();
  for(int i=0;i<mblock;i++){
    HV_val[i] = HV_val[i]*100000.;
    HR_val[i] = HR_val[i]*100000.;
  }    
  for(int i=0;i<mblock;i++){
    dV_val[i] = 2*HV_val[i];
    dR_val[i] = HV_val[i]+HR_val[i];
  }    
  srite("VRHRV",dV_val,sizeof(float)*nz*nx);
  srite("VRHRR",dR_val,sizeof(float)*nz*nx);  

 

  cout << "========================== Hessian action on (V, I) ==========================" <<endl; 
  cout << "    Compute perturbation  " <<endl;
  dV->zeroElement();
  dR->zeroElement();
  HV->zeroElement();
  HR->zeroElement();
  for(int iz=100; iz < nz; iz+=50){
    for(int ix=120; ix < nx; ix+=100){
      int pos = iz*nx+ix;
      dV_val[pos] = 0.001;    
      dR_val[pos] = -0.001;   
    }
  }
  Run->HessianFunc(dV,dR,HV,HR);
  dV->zeroElement();
  dR->zeroElement();
  for(int i=0;i<mblock;i++){
    HV_val[i] = HV_val[i]*100000.;
    HR_val[i] = HR_val[i]*100000.;
  }    
  for(int i=0;i<mblock;i++){
    dV_val[i] = HV_val[i]-HR_val[i];
    dR_val[i] = HV_val[i]+HR_val[i];
  }    
  srite("VIHVV",dV_val,sizeof(float)*nz*nx);
  srite("VIHVI",dR_val,sizeof(float)*nz*nx);  

  dV->zeroElement();
  dR->zeroElement();
  HV->zeroElement();
  HR->zeroElement();
  for(int iz=100; iz < nz; iz+=50){
    for(int ix=100; ix < nx; ix+=100){
      int pos = iz*nx+ix;
      dV_val[pos] = 0.001;    
      dR_val[pos] = 0.001;    
    }
  }
  Run->HessianFunc(dV,dR,HV,HR);
  dV->zeroElement();
  dR->zeroElement();
  for(int i=0;i<mblock;i++){
    HV_val[i] = HV_val[i]*100000.;
    HR_val[i] = HR_val[i]*100000.;
  }    
  for(int i=0;i<mblock;i++){
    dV_val[i] = HV_val[i]-HR_val[i];
    dR_val[i] = HV_val[i]+HR_val[i];
  }  
  srite("VIHIV",dV_val,sizeof(float)*nz*nx);
  srite("VIHII",dR_val,sizeof(float)*nz*nx);  




  cout << "========================== Hessian action on (I, rho) ==========================" <<endl; 
  cout << "    Compute perturbation  " <<endl;
  dV->zeroElement();
  dR->zeroElement();
  HV->zeroElement();
  HR->zeroElement();
  for(int iz=100; iz < nz; iz+=50){
    for(int ix=120; ix < nx; ix+=100){
      int pos = iz*nx+ix;
      dV_val[pos] = 2.*0.001;    
      dR_val[pos] = 0.00;   
    }
  }
  Run->HessianFunc(dV,dR,HV,HR);
  dV->zeroElement();
  dR->zeroElement();
  for(int i=0;i<mblock;i++){
    HV_val[i] = HV_val[i]*100000.;
    HR_val[i] = HR_val[i]*100000.;
  }    
  for(int i=0;i<mblock;i++){
    dV_val[i] = 2.*HV_val[i];
    dR_val[i] = -1.*HV_val[i]+HR_val[i];
  }    
  srite("IRHII",dV_val,sizeof(float)*nz*nx);
  srite("IRHIR",dR_val,sizeof(float)*nz*nx);  

  dV->zeroElement();
  dR->zeroElement();
  HV->zeroElement();
  HR->zeroElement();
  for(int iz=100; iz < nz; iz+=50){
    for(int ix=100; ix < nx; ix+=100){
      int pos = iz*nx+ix;
      dV_val[pos] = -0.001;    
      dR_val[pos] = 0.001;    
    }
  }
  Run->HessianFunc(dV,dR,HV,HR);
  dV->zeroElement();
  dR->zeroElement();
  for(int i=0;i<mblock;i++){
    HV_val[i] = HV_val[i]*100000.;
    HR_val[i] = HR_val[i]*100000.;
  }    
  for(int i=0;i<mblock;i++){
    dV_val[i] = 2.*HV_val[i];
    dR_val[i] = -1.*HV_val[i]+HR_val[i];
  }   
  srite("IRHRI",dV_val,sizeof(float)*nz*nx);
  srite("IRHRR",dR_val,sizeof(float)*nz*nx);  


  return 0;
}
