#ifndef ENGINE_H
#define ENGINE_H 1



#include <cmath>
#include <vector>
#include <iostream>
#include "base/Operator.h"
#include "base/DataInCoreFloat.h"
#include "base/inCoreFloat.h"
#include "base/Vector.h"
#include "Clean.h"
#include "Source.h"
#include "SensorPad.h"
#include "Born.h"
#include "Prop.h"

typedef long long llong;

/*
  Prop Engine, Gradient and Hessian Operator
  Use carefully and have fun.
  
  First Created: 2015-11-04, Yinbin Ma
  Last Modified: 2015-11-04

*/

namespace SEP {


class Engine{
  public:
    Engine(){};
    ~Engine(){};
    void TestEcho();
    
    void SetSpace(Space *mspace, Space *dspace, Space *uspace);
    void SetShot(Vector *shot, std::vector<int> sxv, std::vector<int> szv);
    void SetPropPar(llong nb, llong bsize, llong boundary, float dt, float dx, float dz, int sample, int water=0, float *weight=NULL); // set up nb propagators
    void SetTrueModel(Vector *MK, Vector *MR);
    void SetModel(Vector *MK, Vector *MR);
    void CreateData();
    void GetData(float *data);
    void GetTrueData(float *data);
    void PutData(float *data);
    double OBJF();
    void GetFWave(Vector *wave, int i);
    void GetAdjWave(Vector *wave, int i);
    void GradientFunc(Vector *GradMK, Vector *GradMR);
    void HessianFunc(Vector *dMK, Vector *dMR, Vector *HMK, Vector *HMR);
    void GaussHessianFunc(Vector *dMK, Vector *dMR, Vector *HMK, Vector *HMR);
    int water;
  private:
    llong ns, nb, bsize, nt, nz, nx, boundary, sx, sz;
    int sample;
    std::vector<int> sxv, szv;
    //float dt, dx, dz, pi;
    float dt, dx, dz;
    float *dataTrue, *dataRes;
    float *weight;
    inCoreFloat *mspace, *dspace, *uspace;
    Vector *MK, *MR, *shot;
    std::vector<Vector *> DataVec, Data1Vec, VelVec, rhoVec;
    std::vector<Vector *> FU0Vec, AU0Vec, FU1Vec, AU1Vec, SVec, ShotVec;
    std::vector<Vector *> dMKVec, dMRVec, dmVVec, dmrVec, C1Vec, C2Vec;
    std::vector<BornEngine *>BornVec;
    std::vector<Prop *> LVec;
    std::vector<Source *> SourceVec;
    std::vector<SensorPad *> PadVec;
    
    void init();
    CleanVec Zero;

};

class Hess: public Operator{
  public:
    Hess(Space *Vspace, Space *mod, Space *data, Engine *_Run);
    virtual bool Forward(bool add, Vector *model, Vector *data);
    virtual bool Adjoint(bool add, Vector *model, Vector *data);
  private:
    Engine *Run;
    Vector *Vtemp1, *Vtemp2,*rtemp1, *rtemp2;
    int mblock;
    inCoreFloat *Vspace, *mspace;
};


} // end of namespace
#endif