#include "Grad.h"

namespace SEP {



GradX::GradX(Space *mod, Space *data, llong _nb){
  setDomain(mod);
  setRange(data);
  init();  nb = _nb;
}

void GradX::init(){
  coef[0] = 0.;
  coef[1] = 1225./1024;
  coef[2] = -245./3072;
  coef[3] = 49./5120;
  coef[4] = -5./7168;
  std::vector<long long> in_dims = ((inCoreFloat *)domain)->get_ns();
  nz = in_dims[0];
  nx = in_dims[1];
}

bool GradX::Forward(bool add, Vector *model, Vector *data){

  if(!add) Zero.Zero(data,nz,nx);
  float *__restrict__ m = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ d = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();
  
  // ISPC substitution
  GradXISPCForward(m,d,nz,nx,nb);
  /*
  llong position;
  for(int iz = nb; iz < nz-nb; iz++){
    for(int ix = nb; ix < nx - nb - 1; ix++){
      position = iz*nx+ix;
      d[position] += -1.*coef[4]*m[position-3];
      d[position] += -1.*coef[3]*m[position-2];
      d[position] += -1.*coef[2]*m[position-1];
      d[position] += -1.*coef[1]*m[position];
      d[position] += 1.*coef[1]*m[position+1];
      d[position] += 1.*coef[2]*m[position+2];
      d[position] += 1.*coef[3]*m[position+3];
      d[position] += 1.*coef[4]*m[position+4];      
    }
  }
  */
  return 0;
}

bool GradX::Adjoint(bool add, Vector *model, Vector *data){
  if(!add) Zero.Zero(model,nz,nx);
  float *__restrict__ m = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ d = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();
  
  // ISPC substitution
  GradXISPCAdjoint(d,m,nz,nx,nb);  
  /*
  llong position;
  for(int iz = nb; iz < nz-nb; iz++){
    for(int ix = nb; ix < nx - nb - 1; ix++){
      position = iz*nx+ix;
      m[position-3] += -1.*coef[4]*d[position];
      m[position-2] += -1.*coef[3]*d[position];
      m[position-1] += -1.*coef[2]*d[position];
      m[position]   += -1.*coef[1]*d[position];
      m[position+1] += 1.*coef[1]*d[position];
      m[position+2] += 1.*coef[2]*d[position];
      m[position+3] += 1.*coef[3]*d[position];
      m[position+4] += 1.*coef[4]*d[position];      
    }
  }
  */
  return 0;
}



GradZ::GradZ(Space *mod, Space *data, llong _nb){
  setDomain(mod);
  setRange(data);
  init();  nb = _nb;
}

void GradZ::init(){
  coef[0] = 0.;
  coef[1] = 1225./1024;
  coef[2] = -245./3072;
  coef[3] = 49./5120;
  coef[4] = -5./7168;
  std::vector<long long> in_dims = ((inCoreFloat *)domain)->get_ns();
  nz = in_dims[0];
  nx = in_dims[1];
}

bool GradZ::Forward(bool add, Vector *model, Vector *data){

  if(!add) Zero.Zero(data,nz,nx);
  float *__restrict__ m = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ d = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();
  // ISPC substitution
  GradZISPCForward(m,d,nz,nx,nb);
  /*   
  llong position;
  for(int iz = nb; iz < nz-nb - 1; iz++){
    for(int ix = nb; ix < nx - nb; ix++){
      position = iz*nx+ix;
      d[position] += -1.*coef[4]*m[position-3*nx];
      d[position] += -1.*coef[3]*m[position-2*nx];
      d[position] += -1.*coef[2]*m[position-1*nx];
      d[position] += -1.*coef[1]*m[position];
      d[position] += 1.*coef[1]*m[position+1*nx];
      d[position] += 1.*coef[2]*m[position+2*nx];
      d[position] += 1.*coef[3]*m[position+3*nx];
      d[position] += 1.*coef[4]*m[position+4*nx];      
    }
  }
  */
  return 0;
}

bool GradZ::Adjoint(bool add, Vector *model, Vector *data){
  if(!add) Zero.Zero(model,nz,nx);
  float *__restrict__ m = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ d = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();
 
  // ISPC substitution
  GradZISPCAdjoint(d,m,nz,nx,nb);
  /* 
  llong position;
  for(int iz = nb; iz < nz-nb - 1; iz++){
    for(int ix = nb; ix < nx - nb; ix++){
      position = iz*nx+ix;
      m[position-3*nx] += -1.*coef[4]*d[position];
      m[position-2*nx] += -1.*coef[3]*d[position];
      m[position-1*nx] += -1.*coef[2]*d[position];
      m[position]      += -1.*coef[1]*d[position];
      m[position+1*nx] += 1.*coef[1]*d[position];
      m[position+2*nx] += 1.*coef[2]*d[position];
      m[position+3*nx] += 1.*coef[3]*d[position];
      m[position+4*nx] += 1.*coef[4]*d[position];      
    }
  }
  */
  return 0;
}


GradXT::GradXT(Space *mod, Space *data, llong _nb){
  setDomain(mod);
  setRange(data);
  init();  nb = _nb;
}

void GradXT::init(){
  coef[0] = 0.;
  coef[1] = 1225./1024;
  coef[2] = -245./3072;
  coef[3] = 49./5120;
  coef[4] = -5./7168;
  std::vector<long long> in_dims = ((inCoreFloat *)domain)->get_ns();
  nz = in_dims[0];
  nx = in_dims[1];
}

bool GradXT::Forward(bool add, Vector *model, Vector *data){

  if(!add) Zero.Zero(data,nz,nx);
  float *__restrict__ m = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ d = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();

  // ISPC substitution
  GradXTISPCForward(m,d,nz,nx,nb);
  /*   
  llong position;
  for(int iz = nb; iz < nz-nb; iz++){
    for(int ix = nb; ix < nx - nb; ix++){
      position = iz*nx+ix;
      d[position] += -1.*coef[4]*m[position-4];
      d[position] += -1.*coef[3]*m[position-3];
      d[position] += -1.*coef[2]*m[position-2];
      d[position] += -1.*coef[1]*m[position-1];
      d[position] += 1.*coef[1]*m[position];
      d[position] += 1.*coef[2]*m[position+1];
      d[position] += 1.*coef[3]*m[position+2];
      d[position] += 1.*coef[4]*m[position+3];      
    }
  }
  */
  return 0;
}

bool GradXT::Adjoint(bool add, Vector *model, Vector *data){
  if(!add) Zero.Zero(model,nz,nx);
  float *__restrict__ m = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ d = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();

  // ISPC substitution
  GradXTISPCAdjoint(d,m,nz,nx,nb);
  /*   
  llong position;
  for(int iz = nb; iz < nz-nb; iz++){
    for(int ix = nb; ix < nx - nb; ix++){
      position = iz*nx+ix;
      m[position-4] += -1.*coef[4]*d[position];
      m[position-3] += -1.*coef[3]*d[position];
      m[position-2] += -1.*coef[2]*d[position];
      m[position-1] += -1.*coef[1]*d[position];
      m[position] += 1.*coef[1]*d[position];
      m[position+1] += 1.*coef[2]*d[position];
      m[position+2] += 1.*coef[3]*d[position];
      m[position+3] += 1.*coef[4]*d[position];      
    }
  }
  */
  return 0;
}



GradZT::GradZT(Space *mod, Space *data, llong _nb){
  setDomain(mod);
  setRange(data);
  init();  nb = _nb;
}

void GradZT::init(){
  coef[0] = 0.;
  coef[1] = 1225./1024;
  coef[2] = -245./3072;
  coef[3] = 49./5120;
  coef[4] = -5./7168;
  std::vector<long long> in_dims = ((inCoreFloat *)domain)->get_ns();
  nz = in_dims[0];
  nx = in_dims[1];
}
bool GradZT::Forward(bool add, Vector *model, Vector *data){

  if(!add) Zero.Zero(data,nz,nx);
  float *__restrict__ m = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ d = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();

  // ISPC substitution
  GradZTISPCForward(m,d,nz,nx,nb);
  /*   
  llong position;
  for(int iz = nb; iz < nz-nb; iz++){
    for(int ix = nb; ix < nx - nb; ix++){
      position = iz*nx+ix;
      d[position] += -1.*coef[4]*m[position-4*nx];
      d[position] += -1.*coef[3]*m[position-3*nx];
      d[position] += -1.*coef[2]*m[position-2*nx];
      d[position] += -1.*coef[1]*m[position-1*nx];
      d[position] += 1.*coef[1]*m[position];
      d[position] += 1.*coef[2]*m[position+1*nx];
      d[position] += 1.*coef[3]*m[position+2*nx];
      d[position] += 1.*coef[4]*m[position+3*nx];      
    }
  }
  */
  return 0;
}

bool GradZT::Adjoint(bool add, Vector *model, Vector *data){
  if(!add) Zero.Zero(model,nz,nx);
  float *__restrict__ m = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ d = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();

  // ISPC substitution
  GradZTISPCAdjoint(d,m,nz,nx,nb);
  /*   
  llong position;
  for(int iz = nb; iz < nz-nb; iz++){
    for(int ix = nb; ix < nx - nb; ix++){
      position = iz*nx+ix;
      m[position-4*nx] += -1.*coef[4]*d[position];
      m[position-3*nx] += -1.*coef[3]*d[position];
      m[position-2*nx] += -1.*coef[2]*d[position];
      m[position-1*nx]      += -1.*coef[1]*d[position];
      m[position] += 1.*coef[1]*d[position];
      m[position+1*nx] += 1.*coef[2]*d[position];
      m[position+2*nx] += 1.*coef[3]*d[position];
      m[position+3*nx] += 1.*coef[4]*d[position];      
    }
  }
  */
  return 0;
}





}