#ifndef GRAD_H
#define GRAD_H 1

#include "base/Operator.h"
#include "base/DataInCoreFloat.h"
#include "base/inCoreFloat.h"
#include "Clean.h"
#include <omp.h>

typedef long long llong;

/*
  Gradient operator
  
  GradX:  regular     -> X staggered grid
  GradZ:  regular     -> Z staggered grid
  GradXT: X staggered -> regular
  GradZT: Z staggered -> regular

  
  First Created: 2015-11-03, Yinbin Ma
  Last Modified: 2015-11-03
  
*/

extern "C" void GradXISPCForward(float *pin, float *pout,int n2, int n1, int nb);
extern "C" void GradXISPCAdjoint(float *pin, float *pout,int n2, int n1, int nb);
extern "C" void GradZISPCForward(float *pin, float *pout,int n2, int n1, int nb);
extern "C" void GradZISPCAdjoint(float *pin, float *pout,int n2, int n1, int nb);
extern "C" void GradXTISPCForward(float *pin, float *pout,int n2, int n1, int nb);
extern "C" void GradXTISPCAdjoint(float *pin, float *pout,int n2, int n1, int nb);
extern "C" void GradZTISPCForward(float *pin, float *pout,int n2, int n1, int nb);
extern "C" void GradZTISPCAdjoint(float *pin, float *pout,int n2, int n1, int nb);

namespace SEP{

class GradX: public Operator{
  public:
    GradX(Space *mod, Space *data, llong _nb=5);
    ~GradX(){}
    virtual bool Forward(bool add, Vector *model, Vector *data);
    virtual bool Adjoint(bool add, Vector *model, Vector *data);
    
  private:
    void init();
    double coef[10];
    llong nz, nx, nb;
    CleanVec Zero;
};


class GradZ: public Operator{
  public:
    GradZ(Space *mod, Space *data, llong _nb=5);
    ~GradZ(){}
    virtual bool Forward(bool add, Vector *model, Vector *data);
    virtual bool Adjoint(bool add, Vector *model, Vector *data);
    
  private:
    void init();
    double coef[10];
    llong nz, nx, nb;
    CleanVec Zero;
};


class GradXT: public Operator{
  public:
    GradXT(Space *mod, Space *data, llong _nb=5);
    ~GradXT(){}
    virtual bool Forward(bool add, Vector *model, Vector *data);
    virtual bool Adjoint(bool add, Vector *model, Vector *data);
    
  private:
    void init();
    double coef[10];
    llong nz, nx, nb;
    CleanVec Zero;
};


class GradZT: public Operator{
  public:
    GradZT(Space *mod, Space *data, llong _nb=5);
    ~GradZT(){}
    virtual bool Forward(bool add, Vector *model, Vector *data);
    virtual bool Adjoint(bool add, Vector *model, Vector *data);
    
  private:
    void init();
    double coef[10];
    llong nz, nx, nb;
    CleanVec Zero;
};









} // end of namespace
#endif