#include "Interpolation.h"

namespace SEP {

IntX::IntX(Space *mod, Space *data){
  setDomain(mod);
  setRange(data);
  init();
}
void IntX::init() {
  std::vector<long long> in_dims = ((inCoreFloat *)domain)->get_ns();
  nz = in_dims[0];
  nx = in_dims[1];
}

bool IntX::Forward(bool add, Vector *model, Vector *data) {
  if(!add) Zero.Zero(data,nz,nx);
  float *__restrict__ mval = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ dval = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();
  
 
  llong position;
  for(long long iz=0; iz<nz;iz++){
    for(long long ix=0; ix<nx-1;ix++){
      position = iz*nx+ix;
          dval[position] += (mval[position]+mval[position+1])/2;
    }  
  } 
  for(long long iz=0; iz<nz;iz++){
	  position = iz*nx+nx-1;
	  dval[position] += (mval[position]+mval[position])/2;
  }  
  return 0;
}

bool IntX::Adjoint(bool add, Vector *model, Vector *data) {
  if(!add) Zero.Zero(model,nz,nx);
  float *__restrict__ mval = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ dval = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();
  
  llong position;
  
  for(long long iz=0; iz<nz;iz++){
    for(long long ix=1; ix<nx;ix++){
      position = iz*nx+ix;
          mval[position] += (dval[position]+dval[position-1])/2;
    }  
  }   
  for(long long iz=0; iz<nz;iz++){
	  position = iz*nx;
      mval[position] += dval[position]/2.;
	  position = iz*nx+nx-1;
	  mval[position] += dval[position]/2;
  } 
  /*
  for(long long iz=0; iz<nz;iz++){
    for(long long ix=0; ix<nx-1;ix++){
      position = iz*nx+ix;
          mval[position] += dval[position]/2.;
          mval[position+1] += dval[position]/2.;
    }  
  } 
  for(long long iz=0; iz<nz;iz++){
	  position = iz*nx+nx-1;
      mval[position] += dval[position]/2.;
      mval[position] += dval[position]/2.;
  } 
  */ 
  return 0;
}


IntZ::IntZ(Space *mod, Space *data){
  setDomain(mod);
  setRange(data);
  init();
}
void IntZ::init() {
  std::vector<long long> in_dims = ((inCoreFloat *)domain)->get_ns();
  nz = in_dims[0];
  nx = in_dims[1];
}

bool IntZ::Forward(bool add, Vector *model, Vector *data) {
  if(!add) Zero.Zero(data,nz,nx);
  float *__restrict__ mval = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ dval = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();
  

  llong position;
  for(long long iz=0; iz<nz-1;iz++){
	for(long long ix=0; ix<nx;ix++){
      position = iz*nx+ix;
          dval[position] += (mval[position]+mval[position+nx])/2;
    }  
  } 
  for(long long ix=0; ix<nx;ix++){
    position = (nz-1)*nx+ix;
	  dval[position] += (mval[position]+mval[position])/2;
  }  
  return 0;
}

bool IntZ::Adjoint(bool add, Vector *model, Vector *data) {
  if(!add) Zero.Zero(model,nz,nx);
  float *__restrict__ mval = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ dval = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();
  
  llong position;
  
  
  for(long long iz=1; iz<nz;iz++){
	for(long long ix=0; ix<nx;ix++){
      position = iz*nx+ix;
          mval[position] += (dval[position]+dval[position-nx])/2;
    }  
  } 
  for(long long ix=0; ix<nx;ix++){
    position = ix;
    mval[position] += dval[position]/2.;
    position = (nz-1)*nx+ix;
	mval[position] += dval[position]/2;
  }  
  
  /*
  for(long long iz=0; iz<nz-1;iz++){
	for(long long ix=0; ix<nx;ix++){
      position = iz*nx+ix;
          mval[position] += dval[position]/2.;
          mval[position+nx] += dval[position]/2.;
    }  
  } 
  for(long long ix=0; ix<nx;ix++){
    position = (nz-1)*nx+ix;
	mval[position] += dval[position]/2.;
	mval[position] += dval[position]/2.;
  } 
  */ 
  return 0;
}


}