/*C version of propagation kernel

  void LapISPCForward(uniform float pin[], uniform float pout[],uniform int n2, uniform int n1)
  void LapISPCAdjoint(uniform float pin[], uniform float pout[],uniform int n2, uniform int n1)
  void GradXISPCForward(uniform float pin[], uniform float pout[],uniform int n2, uniform int n1, uniform int nb)
  void GradXISPCAdjoint(uniform float pin[], uniform float pout[],uniform int n2, uniform int n1, uniform int nb)
  void GradZISPCForward(uniform float pin[], uniform float pout[],uniform int n2, uniform int n1, uniform int nb)
  void GradZISPCAdjoint(uniform float pin[], uniform float pout[],uniform int n2, uniform int n1, uniform int nb)
  void GradXTISPCForward(uniform float pin[], uniform float pout[],uniform int n2, uniform int n1, uniform int nb)
  void GradXTISPCAdjoint(uniform float pin[], uniform float pout[],uniform int n2, uniform int n1, uniform int nb)
  void GradZTISPCForward(uniform float pin[], uniform float pout[],uniform int n2, uniform int n1, uniform int nb)
  void GradZTISPCAdjoint(uniform float pin[], uniform float pout[],uniform int n2, uniform int n1, uniform int nb)
*/
export void LapISPCForward(uniform float pin[], uniform float pout[],uniform int n2, uniform int n1){
//uniform int i1,i2,i3;
//uniform int i;
uniform float a0=-205.0*2/72;
uniform float a1=8./5.;
uniform float a2=-1./5.;
uniform float a3=8./315;
uniform float a4=-1./560;

for(uniform int i2=4;i2 < n2-4; i2++){
   foreach(i1=4...n1-4){
      int i = i1+i2*n1;
      pout[i] += a4*(pin[i-n1*4]+pin[i-4]+pin[i+4]+pin[i+n1*4]);
      pout[i] += a3*(pin[i-n1*3]+pin[i-3]+pin[i+3]+pin[i+n1*3]);
      pout[i] += a2*(pin[i-n1*2]+pin[i-2]+pin[i+2]+pin[i+n1*2]);  
      pout[i] += a1*(pin[i-n1]+pin[i-1]+pin[i+1]+pin[i+n1])+a0*pin[i];      
}
}
}


/*C version of propagation kernel*/
export void LapISPCAdjoint(uniform float pin[], uniform float pout[],uniform int n2, uniform int n1){
//uniform int i1,i2,i3;
//uniform int i;
uniform float a0=-205.0*2/72;
uniform float a1=8./5.;
uniform float a2=-1./5.;
uniform float a3=8./315;
uniform float a4=-1./560;


for(uniform int i2=4;i2 < n2-4; i2++){
   foreach(i1=4...n1-4){
      int i = i1+i2*n1;
      float temp0 = a0*pin[i];
      float temp1 = a1*pin[i];
      float temp2 = a2*pin[i];
      float temp3 = a3*pin[i];
      float temp4 = a4*pin[i];
      pout[i-4] += temp4;
      pout[i+4] += temp4;
      pout[i-3] += temp3;
      pout[i+3] += temp3;
      pout[i-2] += temp2;
      pout[i+2] += temp2;
      pout[i] += temp0;
      pout[i-1] += temp1;
      pout[i+1] += temp1;
      pout[i-4*n1] += temp4;
      pout[i+4*n1] += temp4;
      pout[i-3*n1] += temp3;
      pout[i+3*n1] += temp3;
      pout[i-2*n1] += temp2;
      pout[i+2*n1] += temp2;
      pout[i-n1] += temp1;
      pout[i+n1] += temp1;
}

}

}

/* C version of GradX function */
export void GradXISPCForward(uniform float pin[], uniform float pout[],uniform int n2, uniform int n1, uniform int nb){
//uniform int i1,i2,i3;
//uniform int i;
uniform float a0=0.;
uniform float a1=9./8.;
uniform float a2=-1./24.;


for(uniform int i2=nb;i2 < n2-nb; i2++){
   foreach(i1=nb...n1-nb-1){
      int position = i2*n1+i1;
      pout[position] += -1.*a2*pin[position-1];
      pout[position] += -1.*a1*pin[position];
      pout[position] += 1.*a1*pin[position+1];
      pout[position] += 1.*a2*pin[position+2]; 
}

}

} // end of GradXISPCForward


/* C version of GradX function */
export void GradXISPCAdjoint(uniform float pin[], uniform float pout[],uniform int n2, uniform int n1, uniform int nb){
//uniform int i1,i2,i3;
//uniform int i;
uniform float a0=0.;
uniform float a1=9./8.;
uniform float a2=-1./24.;


for(uniform int i2=nb;i2 < n2-nb; i2++){
   foreach(i1=nb...n1-nb-1){
      int position = i2*n1+i1;
      float temp1 = a1*pin[position];
      float temp2 = a2*pin[position];
      pout[position-1] -= temp2;
      pout[position]   -= temp1;
      pout[position+1] += temp1;  
      pout[position+2] += temp2;        
   }
}
} // end of GradXISPCAdjoint


/* C version of GradX function */
export void GradZISPCForward(uniform float pin[], uniform float pout[],uniform int n2, uniform int n1, uniform int nb){
//uniform int i1,i2,i3;
//uniform int i;
uniform float a0=0.;
uniform float a1=9./8.;
uniform float a2=-1./24.;


for(uniform int i2=nb;i2 < n2-nb-1; i2++){
   foreach(i1=nb...n1-nb){
      int position = i2*n1+i1;
      pout[position] += -1.*a2*pin[position-1*n1];
      pout[position] += -1.*a1*pin[position];
      pout[position] += 1.*a1*pin[position+1*n1];
      pout[position] += 1.*a2*pin[position+2*n1]; 
}

}

} // end of GradZISPCForward

/* C version of GradX function */
export void GradZISPCAdjoint(uniform float pin[], uniform float pout[],uniform int n2, uniform int n1, uniform int nb){
//uniform int i1,i2,i3;
//uniform int i;
uniform float a0=0.;
uniform float a1=9./8.;
uniform float a2=-1./24.;

for(uniform int i2=nb;i2 < n2-nb-1; i2++){
   foreach(i1=nb...n1-nb){
      int position = i2*n1+i1;
      float temp1 = a1*pin[position];
      float temp2 = a2*pin[position];
      pout[position-1*n1] -= temp2;
      pout[position]      -= temp1;
      pout[position+1*n1] += temp1;  
      pout[position+2*n1] += temp2;
       
   }
}
} // end of GradXISPCAdjoint

/* C version of GradX function */
export void GradXTISPCForward(uniform float pin[], uniform float pout[],uniform int n2, uniform int n1, uniform int nb){
//uniform int i1,i2,i3;
//uniform int i;
uniform float a0=0.;
uniform float a1=9./8.;
uniform float a2=-1./24.;


for(uniform int i2=nb;i2 < n2-nb; i2++){
   foreach(i1=nb...n1-nb){
      int position = i2*n1+i1;
      pout[position] += -1.*a2*pin[position-2];
      pout[position] += -1.*a1*pin[position-1];
      pout[position] += 1.*a1*pin[position];
      pout[position] += 1.*a2*pin[position+1];
  
}

}

} // end of GradXISPCForward


/* C version of GradX function */
export void GradXTISPCAdjoint(uniform float pin[], uniform float pout[],uniform int n2, uniform int n1, uniform int nb){
//uniform int i1,i2,i3;
//uniform int i;
uniform float a0=0.;
uniform float a1=9./8.;
uniform float a2=-1./24.;


for(uniform int i2=nb;i2 < n2-nb; i2++){
   foreach(i1=nb...n1-nb){
      int position = i2*n1+i1;
      float temp1 = a1*pin[position];
      float temp2 = a2*pin[position];
      pout[position-2] -= temp2;
      pout[position-1] -= temp1;
      pout[position+0] += temp1;  
      pout[position+1] += temp2;
     
   }
}
} // end of GradXISPCAdjoint


/* C version of GradX function */
export void GradZTISPCForward(uniform float pin[], uniform float pout[],uniform int n2, uniform int n1, uniform int nb){
//uniform int i1,i2,i3;
//uniform int i;
uniform float a0=0.;
uniform float a1=9./8.;
uniform float a2=-1./24.;


for(uniform int i2=nb;i2 < n2-nb; i2++){
   foreach(i1=nb...n1-nb){
      int position = i2*n1+i1;
      pout[position] += -1.*a2*pin[position-2*n1];
      pout[position] += -1.*a1*pin[position-1*n1];
      pout[position] += 1.*a1*pin[position+0*n1];
      pout[position] += 1.*a2*pin[position+1*n1];

}

}

} // end of GradZISPCForward

/* C version of GradX function */
export void GradZTISPCAdjoint(uniform float pin[], uniform float pout[],uniform int n2, uniform int n1, uniform int nb){
//uniform int i1,i2,i3;
//uniform int i;
uniform float a0=0.;
uniform float a1=9./8.;
uniform float a2=-1./24.;


for(uniform int i2=nb;i2 < n2-nb; i2++){
   foreach(i1=nb...n1-nb){
      int position = i2*n1+i1;
      float temp1 = a1*pin[position];
      float temp2 = a2*pin[position];
      pout[position-2*n1] -= temp2;
      pout[position-1*n1] -= temp1;
      pout[position+0*n1] += temp1;  
      pout[position+1*n1] += temp2;
      
   }
}
} // end of GradXISPCAdjoint

