#include "Laplacian.h"

namespace SEP {

Laplacian4::Laplacian4(Space *mod, Space *data){
  setDomain(mod);
  setRange(data);
  init();
}
void Laplacian4::init(){
  //−1/560	8/315	−1/5	8/5	−205/72	8/5	−1/5	8/315	−1/560
  order = 4;
  coef[0] = -205./72;
  coef[1] = 8./5;
  coef[2] = -1./5;
  coef[3] = 8./315;
  coef[4] = -1./560;
  std::vector<long long> in_dims = ((inCoreFloat *)domain)->get_ns();
  n2 = in_dims[0];
  n1 = in_dims[1];
}

bool Laplacian4::Forward(bool add, Vector *model, Vector *data){
  if(!add) Zero.Zero(data,n2,n1);
  float *__restrict__ m = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ d = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();
  LapISPCForward(m,d,n2,n1);
  /*
    swap i1 <-> i2, n1<->n2, I am just to lazy to change it.
  for(i1 = order; i1 < n1 - order; i1++){
    for(i2=order, position = i1*n2+order; i2 < n2- order; i2++, position++){
      d[position] += coef[4]*(m[position-n2*4]+m[position+n2*4]+m[position+4]+m[position-4]);
      d[position] += coef[3]*(m[position-n2*3]+m[position+n2*3]+m[position+3]+m[position-3]); 
      d[position] += coef[2]*(m[position-n2*2]+m[position+n2*2]+m[position+2]+m[position-2]);  
      d[position] += coef[1]*(m[position-n2*1]+m[position+n2*1]+m[position+1]+m[position-1]);
      d[position] += 2.*coef[0]*m[position];
    }
  }
  */
  return 0;
} // end of Forward

bool Laplacian4::Adjoint(bool add, Vector *model, Vector *data){
  if(!add) Zero.Zero(model,n2,n1);
  float *__restrict__ m = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ d = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();  
  LapISPCAdjoint(d,m,n2,n1);
  /*
  swap i1 <-> i2, n1<->n2, I am just to lazy to change it.
  for(i1 = order; i1 < n1 - order; i1++){
    for(i2=order, position = i1*n2+order; i2 < n2- order; i2++, position++){
      m[position]      += coef[0]*d[position];
      m[position+n2*1] += coef[1]*d[position];
      m[position+n2*2] += coef[2]*d[position];
      m[position+n2*3] += coef[3]*d[position];  
      m[position+n2*4] += coef[4]*d[position];  
      m[position-n2*1] += coef[1]*d[position];
      m[position-n2*2] += coef[2]*d[position];
      m[position-n2*3] += coef[3]*d[position];  
      m[position-n2*4] += coef[4]*d[position]; 
      m[position]      += coef[0]*d[position];
      m[position+1]    += coef[1]*d[position];
      m[position+2]    += coef[2]*d[position];
      m[position+3]    += coef[3]*d[position];  
      m[position+4]    += coef[4]*d[position];  
      m[position-1]    += coef[1]*d[position];
      m[position-2]    += coef[2]*d[position];
      m[position-3]    += coef[3]*d[position];  
      m[position-4]    += coef[4]*d[position];
    }
  }
  */
  return 0;
} // end of Adjoint



}