/*
  ISPC module to accelerate the code
  
  AddISPC(pin, pout, n):
    pout = pout + pin

  ScaleISPC(pin,pout,n):
    pout = pout*pin

  ProdISPC(coef,pin,pout,n):
    pout = pin*coef 

  ProdToISPC(coef,pin,pout,n):
    pout += pin*coef 

*/

/*

*/
export void AddISPC(uniform float pin[], uniform float pout[], uniform int n){
  foreach(i=0...n){
    pout[i] += pin[i];
  }
}

export void ScaleISPC(uniform float pin[], uniform float pout[], uniform int n){
  foreach(i=0...n){
    pout[i] *= pin[i];
  }
}

export void ProdISPC(uniform float coef[], uniform float pin[], uniform float pout[], uniform int n){
  foreach(i=0...n){
    pout[i] = pin[i]*coef[i];
  }
}

export void ProdToISPC(uniform float coef[], uniform float pin[], uniform float pout[], uniform int n){
  foreach(i=0...n){
    pout[i] += pin[i]*coef[i];
  }
}


export void ProdScaleToISPC(uniform float coef, uniform float pin[], uniform float pout[], uniform int n){
  foreach(i=0...n){
    pout[i] += pin[i]*coef;
  }
}

export void LeapFrogISPC(uniform float coef[], uniform float p1[], uniform float px[], uniform float pz[], uniform float p2[], uniform float p3[], uniform int n){
  foreach(i=0...n){
    p3[i] = p2[i]*2.+ coef[i]*(px[i]+pz[i])-p1[i];
  }
}

export void TimeCoreISPC(uniform float pre[], uniform float cur[], uniform float next[], uniform float pin[], uniform float pout[], uniform int n){
  foreach(i=0...n){
    pout[i] += pin[i]*(pre[i]-2.*cur[i]+next[i]);
  }
}