#include <iostream>
#include <cstdlib>
#include <stdbool.h>
#include <vector>
#include <cmath>
#include<string>
#include "seplib.h"

#include <omp.h>
#include "Laplacian.h"
#include "Source.h"
#include "SensorPad.h"
#include "Grad.h"
#include "Interpolation.h"
#include "Born.h"
#include "Prop.h"
#include "Engine.h"
#include "CreateModel.h"
#include "HessInv.h"
#include "Wrapper.h"
#include "NCG.h"
#include "utilities.h"
#include "mkl_rci.h"
#include "mkl_blas.h"
#include "mkl_spblas.h"
#include "mkl_service.h"
extern "C" {
#include "sep3d.h"
}
#define amp 0.01
#define decay 100
using namespace std;
using namespace SEP;



/*
Beta version for Multiparameter FWI

Generated by Yinbin: 2015-07-06

Assume: 
	1) source term
	2) initial condition with gradient = 0
	3) K[GPa], t[ms], rho[g/cm^3]
*/

void InjectKRK(float *VT_val,float *RT_val,int nz,int nx,int pz,int px){
  //VT_val[pz*nx+px] = VT_val[pz*nx+px]*sqrt(1.02);
  
  for(int iz=0;iz<nz;iz++){
    for(int ix=0;ix<nx;ix++){
      int pos = iz*nx+ix;
      float dist = 1.*(iz-pz)*(iz-pz)+1.*(ix-px)*(ix-px);
      if(dist < 400.){
        VT_val[pos] = VT_val[pos]*sqrt(1+exp(-1.*dist/decay)*amp);
      }
  }
  
}

}
void InjectKRR(float *VT_val,float *RT_val,int nz,int nx,int pz,int px){
  //VT_val[pz*nx+px] = VT_val[pz*nx+px]/sqrt(1.02);
  //RT_val[pz*nx+px] = RT_val[pz*nx+px]*1.02;

  for(int iz=0;iz<nz;iz++){
    for(int ix=0;ix<nx;ix++){
      int pos = iz*nx+ix;
      float dist = 1.*(iz-pz)*(iz-pz)+1.*(ix-px)*(ix-px)/1.5;
      if(dist < 400.){
        VT_val[pos] = VT_val[pos]/sqrt(1+exp(-1.*dist/decay)*amp);
        RT_val[pos] = RT_val[pos]*(1+exp(-1.*dist/decay)*amp);
      }
  }
}

}

void InjectVRV(float *VT_val,float *RT_val,int nz,int nx,int pz,int px){
  for(int iz=0;iz<nz;iz++){
    for(int ix=0;ix<nx;ix++){
      int pos = iz*nx+ix;
      float dist = 1.*(iz-pz)*(iz-pz)+1.*(ix-px)*(ix-px);
      if(dist < 400.){
        VT_val[pos] = VT_val[pos]*(1+exp(-1.*dist/decay)*amp);
      }
  }
}
}
void InjectVRR(float *VT_val,float *RT_val,int nz,int nx,int pz,int px){
  for(int iz=0;iz<nz;iz++){
    for(int ix=0;ix<nx;ix++){
      int pos = iz*nx+ix;
      float dist = 1.*(iz-pz)*(iz-pz)+1.*(ix-px)*(ix-px)/1.5;
      if(dist < 400.){
        RT_val[pos] = RT_val[pos]*(1-exp(-1.*dist/decay)*amp);
      }
  }
}
}

void InjectIVV(float *VT_val,float *RT_val,int nz,int nx,int pz,int px){
  for(int iz=0;iz<nz;iz++){
    for(int ix=0;ix<nx;ix++){
      int pos = iz*nx+ix;
      float dist = 1.*(iz-pz)*(iz-pz)+1.*(ix-px)*(ix-px);
      if(dist < 400.){
        VT_val[pos] = VT_val[pos]*(1+exp(-1.*dist/decay)*amp);
        RT_val[pos] = RT_val[pos]/(1+exp(-1.*dist/decay)*amp);
      }
  }
}
}
void InjectIVI(float *VT_val,float *RT_val,int nz,int nx,int pz,int px){
  for(int iz=0;iz<nz;iz++){
    for(int ix=0;ix<nx;ix++){
      int pos = iz*nx+ix;
      float dist = 1.*(iz-pz)*(iz-pz)+1.*(ix-px)*(ix-px)/1.5;
      if(dist < 200.){
        RT_val[pos] = RT_val[pos]*(1+exp(-1.*dist/decay)*amp);
      }
  }
}
}

int main(int argc, char *argv[]) {
  cout << "========================================"<<endl;
  cout << "                                        "<<endl;
  cout << "========================================"<<endl;
  cout << "              Test Modulus.        "<<endl;
  initpar(argc,argv);       // Initialize the sep library
  omp_set_num_threads(12);   // Initialize the parallel environment
  
  // Define dimension of interest
  int nx, nz, nt;             // size of computational grid
  int sx, sz, ns, oxs, dxs;   // source geometry
  int boundary;               // size of zero boudary
  float dx, dz, dt;           // propagation parameter
  int nb, bsize;              // nb is the number of cores, bsize is shot per core
  int sample;                 // data sampling
  int mblock;                 // size of velocity model
  int type;                   // type of inversion, 1 for (V,r=const), 2 for (V=const,I), 10 for joint inversion
  int niter;                  // number of iteration for CG method

   // Get parameters
  getch("nb","d",&nb);               
  getch("type","d",&type);      
  getch("bsize","d",&bsize);        ns = nb*bsize;
  getch("nx","d",&nx);              
  getch("nz","d",&nz);              mblock=nx*nz;         
  getch("nt","d",&nt);              
  getch("boundary","d",&boundary);  
  getch("dx","f",&dx);              dz = dx;
  getch("sz","d",&sz);              
  getch("oxs","d",&oxs);            
  getch("dxs","d",&dxs);            
  getch("dt","f",&dt);              
  getch("ns","d",&ns);             
  getch("sample","d",&sample);      
  getch("niter","d",&niter);  
  // Summary of jobs
  cout << "========================== Beginning of FWI for Marmousi model ==========================" <<endl;  
  cout << "Number of cores = "   << nb << endl;
  cout << "Jobs on each core = " << bsize << endl;
  cout << "Number of shots = "   << ns << endl;
  if(type==1)  cout << "Velocity estimation for constant rho" << endl;
  if(type==2)  cout << "Impedance estimation for constant velocity" << endl;
  if(type==10) cout << "Joint parameter estimation" << endl;  
  cout << "nx = " << nx << endl;
  cout << "nz = " << nz << endl;
  cout << "nt = " << nt << endl;
  cout << "boundary = " << boundary << endl;
  cout << "dx = " << dx <<"[m]"<< endl;  
  cout << "sz = " << sz << endl;
  cout << "oxs = " << oxs << endl;
  cout << "dxs = " << dxs << endl;
  cout << "dt = " << dt << "[ms]"<< endl;
  cout << "sample = " << sample << endl;
  cout << "Number of CG steps = " << niter << endl;
  
  // Set up space 
  inCoreFloat1D *sspace = new inCoreFloat1D(nt);               // space for source
  inCoreFloat2D *mspace = new inCoreFloat2D(nz,nx);            // space for velocity, reflectivity
  inCoreFloat2D *bmspace= new inCoreFloat2D(2*nz,nx);          // space for velocity, reflectivity
  inCoreFloat3D *uspace = new inCoreFloat3D(nt,nz,nx);         // space for wavefield
  inCoreFloat2D *dspace = new inCoreFloat2D(nt,nx);            // space for data
 
  
  // Set up vectors
  Vector *shot     = new Vector(sspace);          // shot vector
  Vector *VT       = new Vector(mspace);          // true velocity
  Vector *RT       = new Vector(mspace);          // true density
  Vector *VM       = new Vector(mspace);          // starting velocity
  Vector *RM       = new Vector(mspace);          // starting density
  Vector *u1       = new Vector(uspace);          // wave field
  Vector *MKT      = new Vector(mspace);          // K_inv for the true model
  Vector *MRT      = new Vector(mspace);          // rho_inv for the true model
  Vector *MKM      = new Vector(mspace);          // K_inv for the starting model
  Vector *MRM      = new Vector(mspace);          // rho_inv for the starting model
  Vector *Grad1    = new Vector(mspace);
  Vector *Grad2    = new Vector(mspace);  
  Vector *Delta1   = new Vector(mspace);
  Vector *Delta2   = new Vector(mspace);   
 
  // Creating pointers
  float *__restrict__ shot_val  = ((dataInCoreFloat*)shot->getDataContainer())->getFloatPtr(); 
  float *__restrict__ VT_val    = ((dataInCoreFloat*)VT->getDataContainer())->getFloatPtr(); 
  float *__restrict__ VM_val    = ((dataInCoreFloat*)VM->getDataContainer())->getFloatPtr(); 
  float *__restrict__ RT_val    = ((dataInCoreFloat*)RT->getDataContainer())->getFloatPtr(); 
  float *__restrict__ RM_val    = ((dataInCoreFloat*)RM->getDataContainer())->getFloatPtr(); 
  float *__restrict__ MKT_val   = ((dataInCoreFloat*)MKT->getDataContainer())->getFloatPtr(); 
  float *__restrict__ MKM_val   = ((dataInCoreFloat*)MKM->getDataContainer())->getFloatPtr(); 
  float *__restrict__ MRT_val   = ((dataInCoreFloat*)MRT->getDataContainer())->getFloatPtr(); 
  float *__restrict__ MRM_val   = ((dataInCoreFloat*)MRM->getDataContainer())->getFloatPtr();   
  float *__restrict__ u1_val    = ((dataInCoreFloat*)u1->getDataContainer())->getFloatPtr();  
  float *__restrict__ Grad1_val    = ((dataInCoreFloat*)Grad1->getDataContainer())->getFloatPtr();  
  float *__restrict__ Grad2_val    = ((dataInCoreFloat*)Grad2->getDataContainer())->getFloatPtr();   
  float *__restrict__ Delta1_val    = ((dataInCoreFloat*)Delta1->getDataContainer())->getFloatPtr();  
  float *__restrict__ Delta2_val    = ((dataInCoreFloat*)Delta2->getDataContainer())->getFloatPtr();  
  // Read in velocity and density
  cout << "========================== Reading in velocity and density model ==========================" <<endl;  
  sreed("wavelet",shot_val,sizeof(float)*nt); 

  for(int i=0;i<mblock;i++){
    VT_val[i] = 1.5;
    VM_val[i] = 1.5;
    RT_val[i] = 1.;
    RM_val[i] = 1.;
  }
  for(int i=150*nx;i<mblock;i++){
    VT_val[i] = 1.5;
    RT_val[i] = 1.05;  
    VM_val[i] = 1.5;
    RM_val[i] = 1.05;  
  }

  if(type==10){
    cout << " (K, rho) perturbation." <<endl; 
    int flag = 0;
    for(int pz = 100; pz < 190; pz+=50){
      for(int px = 100; px < 490; px += 90){
        if(flag%2==0){
          InjectKRK(VT_val,RT_val,nz,nx,pz,px);
        }
        else{
          InjectKRR(VT_val,RT_val,nz,nx,pz,px);
        }
        flag++;
      }
    }
  }
  if(type==20){
    cout << " (V, rho) perturbation." <<endl; 
    int flag = 0;
    InjectVRV(VT_val,RT_val,nz,nx,100,180);
    InjectVRR(VT_val,RT_val,nz,nx,100,285);
    InjectVRV(VT_val,RT_val,nz,nx,100,285);
    InjectVRR(VT_val,RT_val,nz,nx,100,390);
  }
  if(type==30){
    cout << " (V, I) perturbation." <<endl; 
    int flag = 0;
    for(int pz = 100; pz < 190; pz+=50){
      for(int px = 100; px < 490; px += 90){
        if(flag%2==0){
          InjectIVV(VT_val,RT_val,nz,nx,pz,px);
        }
        else{
          InjectIVI(VT_val,RT_val,nz,nx,pz,px);
        }
        flag++;
      }
    }
  }


  for(int i=0;i<mblock;i++){
    MKT_val[i] = 1./(RT_val[i]*VT_val[i]*VT_val[i]);
    MKM_val[i] = 1./(RM_val[i]*VM_val[i]*VM_val[i]);  
    MRT_val[i] = 1./RT_val[i];
    MRM_val[i] = 1./RM_val[i];
  }
  float *modk = new float[mblock];
  float *modr = new float[mblock];
  float *modv = new float[mblock];
  float *modi = new float[mblock];
  for(int i=0;i<mblock;i++){
    modk[i] = log(MKM_val[i]/MKT_val[i]);
    modr[i] = log(RT_val[i]/RM_val[i]);
    modv[i] = log(VT_val[i]/VM_val[i]);
    modi[i] = log(VT_val[i]*RT_val[i]/VM_val[i]/RM_val[i]);
    
  }
  srite("ModK",modk,sizeof(float)*nz*nx);
  srite("ModR",modr,sizeof(float)*nz*nx); 
  srite("ModI",modi,sizeof(float)*nz*nx); 
  srite("ModV",modv,sizeof(float)*nz*nx); 
  //return 0;
  int water =0;
  // creating weighting function
  float *weight = new float[mblock];
  for(int iz=0;iz < nz;iz++){
    double coef = exp(-1.*(iz-70)*(iz-70)/100.);
    coef = 1.;
    if(iz >= 70) coef = 1.;
    for(int ix=0;ix<nx;ix++){
      int pos = iz*nx+ix;
      weight[pos] = coef;
    }
  }
  //srite("weight",weight,sizeof(float)*nx*nz);

  //Constructing operators
  cout << "========================== Constructing operator ==========================" <<endl; 
  float *cmodl = new float[2*mblock];
  for(int i=0; i < mblock; i++){
    cmodl[i] =MKM_val[i];
    cmodl[i+mblock] =MRM_val[i];
  }   
  std::vector<int> sxv;
  std::vector<int> szv;
  for(int i=0; i < nb*bsize; i++){
    sxv.push_back(oxs+dxs*i);
    szv.push_back(sz);
  }
  Engine *Run = new Engine();
  Run->SetSpace(mspace, dspace, uspace);
  Run->SetShot(shot,sxv,szv);
  Run->SetPropPar(nb,bsize,boundary,dt,dx,dz,sample,water,weight);
  Run->SetTrueModel(MKT,MRT);

  
  
  // Collecting data
  cout << "========================== Collecting data ==========================" <<endl; 
  float *dataTrue;
  float *dataRes = new float[nt*nx*ns];     
  Run->SetModel(MKM,MRM);
  //Run->CreateData();
  cout << "========================== Computating gradient ==========================" <<endl; 
  
  Run->GradientFunc(Grad1,Grad2);
  float temp1, temp2;
  if(type==10){
    cout << " (K, rho) gradients." <<endl; 
  }
  if(type==20){
    cout << " (V, rho) gradients." <<endl; 
    for(int i=0;i<mblock;i++){
      temp1 = Grad1_val[i];
      temp2 = Grad2_val[i];
      Grad1_val[i] = 2.*temp1;
      Grad2_val[i] = temp1 + temp2;
    }
  }
  if(type==30){
    cout << " (V, I) gradients." <<endl; 
    for(int i=0;i<mblock;i++){
      temp1 = Grad1_val[i];
      temp2 = Grad2_val[i];
      Grad1_val[i] = temp1 - temp2;
      Grad2_val[i] = temp1 + temp2;
    }  
  }

  srite("Grad1",Grad1_val,sizeof(float)*nz*nx);
  srite("Grad2",Grad2_val,sizeof(float)*nz*nx);  
  cout << "========================== End of Program, Done! ==========================" <<endl; 
  return 0;  
  

}
