#ifndef NCG_H
#define NCG_H 1

#include <cmath>
#include "base/Operator.h"
#include "base/DataInCoreFloat.h"
#include "base/inCoreFloat.h"
#include "base/Vector.h"
#include "Clean.h"
#include "Source.h"
#include "SensorPad.h"
#include "Born.h"
#include "Prop.h"
#include "Engine.h"
#include "Wrapper.h"
#include "utilities.h"
typedef long long llong;

/*
  Nonlinear Conjugate Gradient method
  Use carefully and have fun.
  
  First Created: 2015-12-02, Yinbin Ma
  Last Modified: 2015-12-02  
*/
namespace SEP {
int VSIZE;
double NCG_inner(float *d1, float *d2, int n){
  double result = 0.;
  for(int i=0;i<n;i++) result += d1[i]*d2[i];
  return result;
}
void bound(float *model){
 /*
  float v2=5.8*5.8;
  int flag= 0;
  for(int i=0;i<VSIZE;i++){
    if(model[i]*v2 < model[i+VSIZE]){
      model[i+VSIZE] = model[i]*v2;
      flag = 1;
    }
  }
  if(flag) std::cout <<"^^^^^^^^^^^^^^bounded^^^^^^^^^^^" <<std::endl;
  */ 
}

double NCGmax;


class Zoom{
  public:
    Zoom(int _mblock){
      mblock = _mblock;
      xtr = new float[mblock];
      gradtr = new float[mblock];
    }  

    int run(double *alpha, float *m, double *f, float *grad, int *nsteps,
             double low, double high, double flow, double fhigh, double glow, double ghi, 
             double f0, double g0, float *m0, float *d,
             double c1, double c2,Wrapper *Wrap){
      // input: 
      // output: alpha, m, f, grad, nsteps
      fail=0;
      //*nsteps = 0; 
      low2 = high; flow2 = fhigh; glow2 = ghi;
      while(fabs(low-high) >1.e-4 && *nsteps <100){
        std::cout <<"        ====ZOOM===="<<std::endl;
        //std::cout <<"low = "<<low<<", high="<<high<<std::endl;
        bisect = (low+high)/2.;    // bisection
        eta = glow + glow2 - 3*(flow - flow2)/(low-low2);
        //sgn = (low2 - low > 0) ? 1 : ((low2 - low < 0) ? -1 : 0);
        sgn = 1.;
        if(low2 < low) sgn = -1.;
        gamma  = sgn*sqrt(eta*eta-glow*flow2);
        interp = low2 - (low2-low)*(glow2+gamma-eta)/(glow2-flow+2*gamma);
        in_int = 0;
        if(low <= bisect){
          if(interp >=low && interp <= bisect)
            in_int = 1;
        }
        else{
          if(interp >= bisect && interp <= low)
            in_int = 1;
        }
        if(in_int)
          alphatr = interp;
        else
          alphatr = bisect;
        memcpy(xtr,m0,mblock*sizeof(float));
        ProdScaleToISPC((float)alphatr,d,xtr,mblock);  // xtr = m0 + alphatr*d
        bound(xtr);    // this is life
        Wrap->Evaluate(xtr, gradtr, &ftr);
        *nsteps = *nsteps+1;
        gtr = NCG_inner(gradtr, d, mblock);  //gtr=gradtr->innerProd(*d);
        std::cout <<"        alphatr=" << alphatr<<std::endl;
        std::cout <<"        dd  =" << NCG_inner(gradtr,gradtr,mblock) << std::endl;
        std::cout <<"        gtr = " << gtr <<", g0 = " << g0 << std::endl;
        std::cout <<"        ftr = " << ftr/NCGmax <<", f0 = " << f0/NCGmax <<",flow=" << flow/NCGmax<< std::endl;
        if(ftr > f0 + c1*alphatr*g0 || ftr >= flow){
          // overshot
          //std::cout <<"        over shot"<<std::endl;
          high=alphatr; fhigh = ftr;
          low2 = high; flow2 = ftr; flow2 = gtr;
        }
        else {
          if(fabs(gtr) <= -c2*g0){   //strong Wolfe condition satified
            *alpha = alphatr;
            memcpy(m,xtr,mblock*sizeof(float));        // m->linearCombo(0.,*xtr,1.);
            *f=ftr;
            memcpy(grad,gradtr,mblock*sizeof(float));  // grad->linearCombo(0.,*gradtr,1.);
            std::cout << "        Zoom successful, f= "<<*f/NCGmax << std::endl;
            return 0;
          }
          if(gtr*(high-low) >=0){
            high = low; fhigh = flow;
          }
          low2 = low; flow2=flow; glow2=flow;
          low=alphatr; flow=ftr; glow=gtr;
        }
        std::cout <<"        low = " << low <<", high = " << high<< std::endl;
      }
      // failed

      *alpha = alphatr;            
      memcpy(m,xtr,mblock*sizeof(float));        // m->linearCombo(0.,*xtr,1.);
      *f=ftr;
      memcpy(grad,gradtr,mblock*sizeof(float));  // grad->linearCombo(0.,*gradtr,1.);
      std::cout << "        Zoom failed,"<<*alpha <<",f ="<<*f/NCGmax << std::endl;
      return 1;
    }

   
  private:
    int mblock; 
    int fail, nsteps,in_int;
    double low2, flow2, glow2, bisect, eta, gamma,sgn,interp,alphatr,ftr,gtr;
    float *xtr, *gradtr;
     //   Func->Evaluate(mnew,&fnew,gradnew);
};



class LineSearch{
  public:
    LineSearch(int _mblock){
      mblock = _mblock;
      zoom_in = new Zoom(mblock);
      c1 = 0.01;
      c2 = 0.8;
      niter = 50;
      mnew = new float[mblock];       //Vector(mspace);
      gradnew = new float[mblock];    // Vector(mspace);
      alphamax=0.01;
    }
    void SetPar(double _alphamax){alphamax = _alphamax;}

    int run(double *alpha, float *m, double *f, float *grad, int *nsteps,
             float *m0, double f0, float *grad0, float *d, Wrapper * Wrap){        
      g0 = NCG_inner(grad0,d,mblock);   //grad0->innerProd(*d);  
      std::cout << "g0 = " << g0 << std::endl;
      if(g0 >=0){
        std::cout << "Line search: not a descent direction!" <<std::endl;
      }
      old = 0; fold=f0; gold=g0;
      alphanew =alphamax;
      float alphaf;
      for(int iter=0; iter < niter; iter++){

        std::cout << "    Line iteration: " << iter << std::endl;
        memcpy(mnew, m0, mblock*sizeof(float));                       //mnew->linearCombo(0.,*m0,1.);
        alphaf = (float)alphanew;
        ProdScaleToISPC(alphaf,d,mnew,mblock); //mnew->linearCombo(1.,*d,(float)alphanew);
        bound(mnew);    // this is life
        Wrap->Evaluate(mnew, gradnew, &fnew);  //Func->Evaluate(mnew,&fnew,gradnew); 
        if(isnan(fnew)) {
            std::cout <<"fuck"<< std::endl;
            
          *alpha = alphanew;
          memcpy(m,mnew,mblock*sizeof(float));          // m->linearCombo(0.,*mnew,1.); 
          *f = fnew; 
          memcpy(grad,gradnew,mblock*sizeof(float));    //grad->linearCombo(0.,*gradnew,1.);
          return 0;
          
          //alphanew = alphanew/2;
          continue;
        }
        *nsteps = *nsteps+1;
        std::cout << "    alphanew=" << alphanew<<", fnew = " << fnew/NCGmax << std::endl;
        gnew = NCG_inner(gradnew,d,mblock); // gradnew->innerProd(*d);
        //std::cout <<"gg = " << gradnew->innerProd(*gradnew)<<"gnew = " << gnew <<", g0 = " << g0 << std::endl;
        if(fnew > f0 + c1*alphanew*g0 || ((fnew >= fold)&&iter>0)){
          std::cout << "    overshot: zoom in norm order:" << old <<"to "<< alphanew<< std::endl;
          zoom_in->run(alpha,m,f,grad,nsteps,
                       old,alphanew,fold,fnew,gold,gnew,
                       f0,g0,m0,d,c1,c2,Wrap);
          //std::cout << "    Haha, Zoom successful, f= "<<*f << std::endl;
          //*alpha = alphanew;
          //std::cout << "mnorm" << m->innerProd(*m) <<std::endl;
          //std::cout << " Line search successful. case 1" << std::endl;
          return 0;
        }
        if(fabs(gnew) <= -c2*g0){
        //if(gnew <= -c2*g0){
          *alpha = alphanew;
          memcpy(m,mnew,mblock*sizeof(float));          // m->linearCombo(0.,*mnew,1.); 
          *f = fnew; 
          memcpy(grad,gradnew,mblock*sizeof(float));    //grad->linearCombo(0.,*gradnew,1.);
          return 0;
        }
        if(gnew >=0){
          std::cout << "    zoom in norm order:" << alphanew <<"to "<< old<< std::endl;
           zoom_in->run(alpha,m,f,grad,nsteps,
                       alphanew,old,fnew,fold,gnew,gold,
                       f0,g0,m0,d,c1,c2,Wrap);
          //std::cout << "    Haha, Zoom successful, f= "<<*f << std::endl;
          //*alpha = alphanew;
          //std::cout << " Line search successful. case 2" << std::endl;
          return 0;      
        }
        old = alphanew;
        fold = fnew;
        gold = gnew;
        alphanew = 2*alphanew;
      }  // end of for
      std::cout <<"    line search: minimizer not bracketed" << std::endl;
      *alpha = alphanew; 
      memcpy(m,mnew,mblock*sizeof(float));         //m->linearCombo(0.,*mnew,1.); 
      *f = fnew;
      memcpy(grad,gradnew,mblock*sizeof(float));   // grad->linearCombo(0.,*gradnew,1.);
    }  
    ~LineSearch(){}
  
  private:
    Zoom *zoom_in;
    float *mnew,*gradnew;
    double g0, old,gold;
    double c1; // Wolfe condition
    double c2; // curvature condition
    double alphamax;
    int niter, mblock;
    double fold, fnew, alphanew,gnew;
    
};

class NCG{
  public:
    NCG(int _msize, Wrapper *_Wrap){
      msize = _msize;
      VSIZE = msize;
      mblock = 2*msize;
      Wrap   = _Wrap;
      lsch = new LineSearch(mblock);
      niter = 50;
      normtol = 1.e-10;
      y = new float[mblock];
      gprev = new float[mblock];
      mprev = new float[mblock];
      dmodl = new float[mblock];

    }
    void SetPar(int _niter){niter = _niter;}
    int run(float *modl, float *grad, double *objf, int *nfunc, float *objs,float *modlVs, float *modlRs, float *grads, float *gradrs){
  
      Wrap->Evaluate(modl, grad, objf);
      memcpy(modlVs,modl,msize*sizeof(float));         // record
      memcpy(modlRs,&modl[msize],msize*sizeof(float)); // record
      memcpy(grads,grad,msize*sizeof(float)); // record
      memcpy(gradrs,&grad[msize],msize*sizeof(float)); // record
      gnorm = NCG_inner(grad,grad,mblock);
      NCGmax = *objf;
      //lsch->SetPar(2.*NCGmax/gnorm);
      std::cout << "starting from f="<< *objf<<", alphamax = " << 2.*NCGmax/gnorm <<", gnorm = "<<gnorm << std::endl;
      objs[0] = 1.;
      nfunc[0] = 0.;
      if(gnorm < normtol) {
        std::cout << "ncg: grad norm < tolorance at initial iteration\n" << std::endl;
        return 0;
      }
      
      ProdScaleToISPC(-1.,grad,dmodl,mblock);  // dmodl = -grad, first steepest descent
      std::cout << "Max iter = " << niter << std::endl;
      double t0, t1;
      for(int iter=0; iter < niter; iter++){
        nsteps = 0;
        std::cout << "==================================================" << std::endl;
        t0=timer();
        gtp = NCG_inner(grad,dmodl,mblock);
        if(gtp >=0 ){ 
          std::cout << "NCG: not a descent direction, restart from steepest descent" << std::endl;
          Wrap->Evaluate(modl, grad, objf);
          ProdScaleToISPC(-1.,grad,dmodl,mblock);
        }
        // Line search 
        memcpy(mprev, modl, mblock*sizeof(float));
        memcpy(gprev, grad, mblock*sizeof(float));
        fprev = *objf;
        lsch->run(&alpha, modl, objf, grad, &nsteps, mprev,fprev,gprev, dmodl, Wrap); // Line search

        if(isnan(*objf)) {std::cout <<"fuck"<< std::endl; return 0;}
        lsch->SetPar(alpha);
        memcpy(&modlVs[(iter+1)*msize],modl,msize*sizeof(float));         // record
        memcpy(&modlRs[(iter+1)*msize],&modl[msize],msize*sizeof(float)); // record
        memcpy(&grads[(iter+1)*msize],grad,msize*sizeof(float)); // record
        memcpy(&gradrs[(iter+1)*msize],&grad[msize],msize*sizeof(float)); // record
        gnorm = NCG_inner(grad,grad,mblock);
        objs[iter+1]= *objf;
        nfunc[iter+1] = nfunc[iter] + nsteps;
        t1=timer();
        std::cout << "NCG iteration: "<< iter << ", step = " << alpha << ", f = "<< *objf/NCGmax <<", gnorm = "<<gnorm << ", nsteps = " << nfunc[iter+1] << ", used " << t1-t0 <<" seconds."<< std::endl;
        if(gnorm <normtol){
          std::cout << " accuracy reached." << std::endl;
          return 0;
        }
        // Assume everything is fine
        memcpy(y, grad, mblock*sizeof(float));
        ProdScaleToISPC(-1.,gprev,y,mblock);
        //y = grad-gprev;
        // PR+
        nmgprevsq = NCG_inner(gprev,gprev,mblock);            // gprev->innerProd(*gprev);
        beta_pr = NCG_inner(grad, y,mblock)/nmgprevsq;        //grad->innerProd(*y)/nmgprevsq;
        beta_fr = NCG_inner(grad, grad,mblock)/nmgprevsq;     //grad->innerProd(*grad)/nmgprevsq;
        if(beta_pr < -beta_fr){
          beta=-beta_fr;
        }
        else if (beta_pr > beta_fr){
          beta = beta_fr;
        }
        else{
          beta = beta_pr;
        }
        for(int i=0;i<mblock;i++) dmodl[i] = beta*dmodl[i] - grad[i];//d->linearCombo(beta,*grad,-1.);          
        std::cout << "  beta  = " << beta <<std::endl;
      }

      std::cout << " max number of iteration reached." << std::endl;
    }
    ~NCG() {}

  private:
    // reference to external objects
    float *y, *gprev,*mprev,*dmodl;
    LineSearch *lsch;
    Wrapper *Wrap;
    double beta, beta_fr, beta_pr, gtp,nmgprevsq,alpha,fprev, gnorm;
    double normtol;
    int niter, nsteps;
    int mblock,msize;
    
};


}
#endif