#include "Prop.h"

namespace SEP {

Prop::Prop(Space *mod, Space *data){
  setDomain(mod);
  setRange(data);
  init();
}

void Prop::init() {
  uspace = (inCoreFloat *)domain;
  std::vector<long long> in_dims = uspace->get_ns();
  nt = in_dims[0];
  nz = in_dims[1];
  nx = in_dims[2];
  mblock = nx*nz;
  Utemp = new Vector((inCoreFloat *)domain);
}

void Prop::SetPropPar(BornEngine *_BornMod, llong _nb, float dt, float dx, Vector *_Vel, int _sample){
 
 
  //Set up parameters
  BornMod = _BornMod;
  nb = _nb;
  sample = _sample;
  dt2 = dt*dt;

  // Set up temporary vectors
  mspace = (inCoreFloat *)_Vel->getSpace();
  VT2   = new Vector(_Vel);
  pre   = new Vector(_Vel);
  cur   = new Vector(_Vel);
  next  = new Vector(_Vel);
  qian  = new Vector(_Vel);
  hou   = new Vector(_Vel);
  tqian = new Vector(_Vel);
  thou  = new Vector(_Vel);
  pre   = new Vector(_Vel);
  
  // Set up operators


  taper = new Taper(_Vel, 50, 50, 0.97);  

  // Interpolation coefficient  
  Ls1 = new float[sample];
  Ls2 = new float[sample];
  Lr1 = new float[sample];
  Lr2 = new float[sample];
  float sqt1 = 2./(sample+1.);
  float sqt2 = 2./(sample+1.);
  sqt1= 1.;///sqrt(sample);
  //sqt2=1./sqrt(sample);
  sqt1= 1./sample;
  sqt2=1./sample;
  for(int i =0; i < sample; i++){
    Ls1[i] = (sample-i)*1./sample;
    Ls2[i] = 1-Ls1[i];
    Lr1[i] = (sample-i-1)*1./sample;
    Lr2[i] = 1-Lr1[i];
    Ls1[i] = Ls1[i]*sqt1;
    Ls2[i] = Ls2[i]*sqt1;
    Lr1[i] = Lr1[i]*sqt2;
    Lr2[i] = Lr2[i]*sqt2;  
  }

} 

void Prop::SetModel(Vector *MK, Vector *MR){
  BornMod->SetModel(MK, MR);
  float *__restrict__ mkval = ((dataInCoreFloat*)MK->getDataContainer())->getFloatPtr();
  float *__restrict__ mrval = ((dataInCoreFloat*)MR->getDataContainer())->getFloatPtr();
  float *__restrict__ cval = ((dataInCoreFloat*)VT2->getDataContainer())->getFloatPtr();
  for(llong i=0;i<mblock; i++){
    cval[i] = dt2/mkval[i];
    //if(i==10202) std::cout << "******* cval = " << cval[i] <<std::endl;
  }
}
void Prop::SetBG(Vector *BG){
  BornMod->SetBG(BG);
}

bool Prop::Forward(bool add, Vector *model, Vector *data){
  if(!add) Zero.Zero(data,nt,mblock);
  // pre processing
  llong cur_pos = 0;
  Zero.Zero(pre,nz,nx);
  Zero.Zero(cur,nz,nx);
  Zero.Zero(next,nz,nx);
  Zero.Zero(Utemp,nt,mblock);
  CleanBound(model,Utemp);
  // write u(0), u(1)
  PushData(add,next,data,cur_pos);  // write u(-1)
  cur_pos = cur_pos + nz*nx;
  PushData(add,next,data,cur_pos);  // write u(0)  
  
  // write u(2) to u(nt-1-2)
  for(llong it=1; it < nt-2; it++){
    // in order to get it+1
    
    // Extract source
    Zero.Zero(qian, nz, nx);  Zero.Zero(hou, nz, nx);  // source
    if(it > 1) PullData(false,qian,Utemp,cur_pos);
    cur_pos = cur_pos + mblock;
    PullData(false,hou,Utemp,cur_pos);
    
    Zero.Zero(tqian, nz, nx); Zero.Zero(thou, nz, nx);  // wave field
    for(llong is=0; is<sample; is++){
      BornMod->FMarch(pre,cur,next);                     // Time marching
      
      // ISPC substitution
      float *__restrict__ qianval = ((dataInCoreFloat*)qian->getDataContainer())->getFloatPtr();
      float *__restrict__ houval = ((dataInCoreFloat*)hou->getDataContainer())->getFloatPtr();
      float *__restrict__ tqianval = ((dataInCoreFloat*)tqian->getDataContainer())->getFloatPtr();
      float *__restrict__ thouval = ((dataInCoreFloat*)thou->getDataContainer())->getFloatPtr();
      float *__restrict__ nextval = ((dataInCoreFloat*)next->getDataContainer())->getFloatPtr();
      ProdScaleToISPC(Ls1[is],qianval,nextval,mblock);
      ProdScaleToISPC(Ls2[is],houval,nextval,mblock);
      //next->linearCombo(1.,qian,Ls1[is]);               // Inject source
      //next->linearCombo(1.,hou, Ls2[is]);               // Inject source
      taper->tapeISPC(cur, next);                       // Taper boundary condition
      ProdScaleToISPC(Lr1[is],nextval,tqianval,mblock);
      ProdScaleToISPC(Lr2[is],nextval,thouval,mblock);     
      //tqian->linearCombo(1.,next,Lr1[is]);              // Save wave field
      //thou->linearCombo(1.,next,Lr2[is]);               // Save wave field
      temp = pre; pre = cur; cur = next; next = temp;
    }
    PushData(true,tqian,data,cur_pos-mblock);  
    if(it<nt-3) PushData(true,thou,data,cur_pos);      
  }
  //std::cout << "should be:" << (nt-2)*mblock <<", actually " << cur_pos << std::endl; 
  return 0;  
}

bool Prop::Adjoint(bool add, Vector *model, Vector *data){
  if(!add) Zero.Zero(model,nt,mblock);
  // pre processing
  llong cur_pos = (nt-1)*nz*nx;
  Zero.Zero(pre,nz,nx);
  Zero.Zero(cur,nz,nx);
  Zero.Zero(next,nz,nx);
  Zero.Zero(Utemp,nt,mblock);

  
  // write u(0), u(1)
  PushData(add,next,data,cur_pos);  // write u(-1)
  cur_pos = cur_pos - nz*nx;
  PushData(add,next,data,cur_pos);  // write u(0)  
  
  // write u(2) to u(nt-1-2)
  for(llong it=1; it < nt-2; it++){
    // in order to get it+1
    
    // Extract source
    Zero.Zero(qian, nz, nx);  Zero.Zero(hou, nz, nx);  // source
    if(it > 1) PullData(false,qian,data,cur_pos);
    cur_pos = cur_pos - mblock;
    PullData(false,hou,data,cur_pos);
    
    Zero.Zero(tqian, nz, nx); Zero.Zero(thou, nz, nx);  // wave field
    for(llong is=0; is<sample; is++){
      BornMod->BMarch(pre,cur,next);                     // Time marching
      // ISPC substitution
      float *__restrict__ qianval = ((dataInCoreFloat*)qian->getDataContainer())->getFloatPtr();
      float *__restrict__ houval = ((dataInCoreFloat*)hou->getDataContainer())->getFloatPtr();
      float *__restrict__ tqianval = ((dataInCoreFloat*)tqian->getDataContainer())->getFloatPtr();
      float *__restrict__ thouval = ((dataInCoreFloat*)thou->getDataContainer())->getFloatPtr();
      float *__restrict__ nextval = ((dataInCoreFloat*)next->getDataContainer())->getFloatPtr();
      ProdScaleToISPC(Ls1[is],qianval,nextval,mblock);
      ProdScaleToISPC(Ls2[is],houval,nextval,mblock);
      //next->linearCombo(1.,qian,Ls1[is]);               // Inject source
      //next->linearCombo(1.,hou, Ls2[is]);               // Inject source
      taper->tapeISPC(cur, next);                       // Taper boundary condition
      ProdScaleToISPC(Lr1[is],nextval,tqianval,mblock);
      ProdScaleToISPC(Lr2[is],nextval,thouval,mblock);    
      //tqian->linearCombo(1.,next,Lr1[is]);              // Save wave field
      //thou->linearCombo(1.,next,Lr2[is]);               // Save wave field
      temp = pre; pre = cur; cur = next; next = temp;

    }
    PushData(true,tqian,Utemp,cur_pos+mblock);  
    if(it<nt-3) PushData(true,thou,Utemp,cur_pos);      
  }
  CleanBound(Utemp,model);  
  //std::cout << "should be:" << mblock <<", actually " << cur_pos << std::endl; 
  return 0;  
}

bool Prop::ForwardB(bool add, Vector *model, Vector *data,float *mask){
  if(!add) Zero.Zero(data,nt,mblock);
  // pre processing
  llong cur_pos = 0;
  Zero.Zero(pre,nz,nx);
  Zero.Zero(cur,nz,nx);
  Zero.Zero(next,nz,nx);
  Zero.Zero(Utemp,nt,mblock);
  CleanBound(model,Utemp);
  // write u(0), u(1)
  PushData(add,next,data,cur_pos);  // write u(-1)
  cur_pos = cur_pos + nz*nx;
  PushData(add,next,data,cur_pos);  // write u(0)  
  
  // write u(2) to u(nt-1-2)
  for(llong it=1; it < nt-2; it++){
    // in order to get it+1
    
    // Extract source
    Zero.Zero(qian, nz, nx);  Zero.Zero(hou, nz, nx);  // source
    //if(it > 1) PullData(false,qian,Utemp,cur_pos);
    cur_pos = cur_pos + mblock;
    //PullData(false,hou,Utemp,cur_pos);
    
    Zero.Zero(tqian, nz, nx); Zero.Zero(thou, nz, nx);  // wave field
    for(llong is=0; is<sample; is++){
      BornMod->FMarch(pre,cur,next);                     // Time marching
      // ISPC substitution
      float *__restrict__ qianval = ((dataInCoreFloat*)qian->getDataContainer())->getFloatPtr();
      float *__restrict__ houval = ((dataInCoreFloat*)hou->getDataContainer())->getFloatPtr();
      float *__restrict__ tqianval = ((dataInCoreFloat*)tqian->getDataContainer())->getFloatPtr();
      float *__restrict__ thouval = ((dataInCoreFloat*)thou->getDataContainer())->getFloatPtr();
      float *__restrict__ nextval = ((dataInCoreFloat*)next->getDataContainer())->getFloatPtr();
      ProdScaleToISPC(Ls1[is],qianval,nextval,mblock);
      ProdScaleToISPC(Ls2[is],houval,nextval,mblock);
      //next->linearCombo(1.,qian,Ls1[is]);               // Inject source
      //next->linearCombo(1.,hou, Ls2[is]);               // Inject source
      taper->tapeISPC(cur, next);                       // Taper boundary condition
      ProdScaleToISPC(Lr1[is],nextval,tqianval,mblock);
      ProdScaleToISPC(Lr2[is],nextval,thouval,mblock);     
      //tqian->linearCombo(1.,next,Lr1[is]);              // Save wave field
      //thou->linearCombo(1.,next,Lr2[is]);               // Save wave field
      temp = pre; pre = cur; cur = next; next = temp;
      float *__restrict__ preval= ((dataInCoreFloat*)pre->getDataContainer())->getFloatPtr();
      float *__restrict__ curval= ((dataInCoreFloat*)cur->getDataContainer())->getFloatPtr();
      float *__restrict__ uval = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
      for(int k=0;k<nx*nz;k++){
        //preval[k] = mask[k]*preval[k]+(1-mask[k])*uval[cur_pos-nx*nz+k];
        curval[k] = mask[k]*curval[k]+(1-mask[k])*uval[cur_pos+k];
      }
    }
    PushData(true,tqian,data,cur_pos-mblock);  
    if(it<nt-3) PushData(true,thou,data,cur_pos);      
  }
  //std::cout << "should be:" << (nt-2)*mblock <<", actually " << cur_pos << std::endl; 
  return 0;  
}

bool Prop::BackB(bool add, Vector *model, Vector *data, float *mask){
  if(!add) Zero.Zero(model,nt,mblock);
  // pre processing
  llong cur_pos = (nt-1)*nz*nx;
  Zero.Zero(pre,nz,nx);
  Zero.Zero(cur,nz,nx);
  Zero.Zero(next,nz,nx);
  Zero.Zero(Utemp,nt,mblock);

  
  // write u(0), u(1)
  PushData(add,next,data,cur_pos);  // write u(-1)
  cur_pos = cur_pos - nz*nx;
  PushData(add,next,data,cur_pos);  // write u(0)  
  
  // write u(2) to u(nt-1-2)
  for(llong it=1; it < nt-2; it++){
    // in order to get it+1
    
    // Extract source
    Zero.Zero(qian, nz, nx);  Zero.Zero(hou, nz, nx);  // source
    //if(it > 1) PullData(false,qian,data,cur_pos);
    cur_pos = cur_pos - mblock;
 
    Zero.Zero(tqian, nz, nx); Zero.Zero(thou, nz, nx);  // wave field
    for(llong is=0; is<sample; is++){
      BornMod->BMarch(pre,cur,next);                     // Time marching
      // ISPC substitution
      float *__restrict__ qianval = ((dataInCoreFloat*)qian->getDataContainer())->getFloatPtr();
      float *__restrict__ houval = ((dataInCoreFloat*)hou->getDataContainer())->getFloatPtr();
      float *__restrict__ tqianval = ((dataInCoreFloat*)tqian->getDataContainer())->getFloatPtr();
      float *__restrict__ thouval = ((dataInCoreFloat*)thou->getDataContainer())->getFloatPtr();
      float *__restrict__ nextval = ((dataInCoreFloat*)next->getDataContainer())->getFloatPtr();
      ProdScaleToISPC(Ls1[is],qianval,nextval,mblock);
      ProdScaleToISPC(Ls2[is],houval,nextval,mblock);
      //next->linearCombo(1.,qian,Ls1[is]);               // Inject source
      //next->linearCombo(1.,hou, Ls2[is]);               // Inject source
      taper->tapeISPC(cur, next);                       // Taper boundary condition
      ProdScaleToISPC(Lr1[is],nextval,tqianval,mblock);
      ProdScaleToISPC(Lr2[is],nextval,thouval,mblock);    
      //tqian->linearCombo(1.,next,Lr1[is]);              // Save wave field
      //thou->linearCombo(1.,next,Lr2[is]);               // Save wave field
      temp = pre; pre = cur; cur = next; next = temp;
     float *__restrict__ preval= ((dataInCoreFloat*)pre->getDataContainer())->getFloatPtr();
     float *__restrict__ curval= ((dataInCoreFloat*)cur->getDataContainer())->getFloatPtr();
     float *__restrict__ uval = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();
      for(int k=0;k<nx*nz;k++){
        preval[k] = mask[k]*preval[k]+(1-mask[k])*uval[cur_pos+nx*nz+k];
        curval[k] = mask[k]*curval[k]+(1-mask[k])*uval[cur_pos+k];
      }
    }
    PushData(true,tqian,Utemp,cur_pos+mblock);  
    if(it<nt-3) PushData(true,thou,Utemp,cur_pos);      
  }
  CleanBound(Utemp,model);  
  //std::cout << "should be:" << mblock <<", actually " << cur_pos << std::endl; 
  return 0;  
}

void Prop::PullData(bool add, Vector * mvec, Vector * uvec, llong position) {
  float *__restrict__ mval = ((dataInCoreFloat*)mvec->getDataContainer())->getFloatPtr();
  float *__restrict__ uval = ((dataInCoreFloat*)uvec->getDataContainer())->getFloatPtr();
  float *__restrict__ vt2val = ((dataInCoreFloat*)VT2->getDataContainer())->getFloatPtr();

  if(add){
    // ISPC substitution
    //ProdToISPC(vt2val,&uval[position],mval,mblock);
    
    for(long long i = 0; i < nz*nx; i++) {
      mval[i] += uval[i + position]*vt2val[i];
    }
    
  }
  else { 
    // ISPC substitution
    //ProdISPC(vt2val,&uval[position],mval,mblock);
    
    for(long long i = 0; i < nz*nx; i++) {
      mval[i] = uval[i + position]*vt2val[i];
    }
    
  }
}

void Prop::PushData(bool add, Vector * mvec, Vector * uvec, llong position) {
  float *__restrict__ mval = ((dataInCoreFloat*)mvec->getDataContainer())->getFloatPtr();
  float *__restrict__ uval = ((dataInCoreFloat*)uvec->getDataContainer())->getFloatPtr();
  if(add){
    // ISPC substitution
    AddISPC(mval,&uval[position],mblock); 
    /*
    for(long long i = 0; i < nz*nx; i++) {
      uval[i + position] += mval[i];
    }
    */
  }
  else {
    // Memory copy subsitution
    memcpy(&uval[position],mval,sizeof(float)*mblock);
    /*
    for(long long i = 0; i < nz*nx; i++) {
      uval[i + position] = mval[i];
    }
    */
    
  }
}




void Prop::CleanBound(Vector *mvec, Vector *uvec){
  float *__restrict__ mval = ((dataInCoreFloat*)mvec->getDataContainer())->getFloatPtr();
  float *__restrict__ uval = ((dataInCoreFloat*)uvec->getDataContainer())->getFloatPtr();
  for(llong it=0; it < nt; it++){
    for(long long iz = nb; iz < nz-nb; iz++) {
    // ISPC substitution
    int start=it*nx*nz+iz*nx+nb;
    AddISPC(&mval[start],&uval[start],nx-2*nb);
    
    /*
      for(llong ix = nb; ix < nx - nb; ix++) {
        llong position = it*nx*nz + iz*nx+ix;
        uval[position] += mval[position];
      }
    */
    }
  }
}

void Prop::Dot(){
      Vector m1(domain);//=new Vector(domain);
      Vector d1(range);//=new Vector(range);
      Vector m2(domain);//=new Vector(domain);
      Vector d2(range);//=new Vector(range);
      float *__restrict__ m1val = ((dataInCoreFloat*)m1.getDataContainer())->getFloatPtr();
      float *__restrict__ d1val = ((dataInCoreFloat*)d1.getDataContainer())->getFloatPtr();
      float *__restrict__ m2val = ((dataInCoreFloat*)m2.getDataContainer())->getFloatPtr();
      float *__restrict__ d2val = ((dataInCoreFloat*)d2.getDataContainer())->getFloatPtr();
      m1.random();
      d2.random();
      //for(llong i=0; i<2*mblock;i++)m1val[i] = 0.;
      //for(llong i=(nt-2)*mblock; i<nt*mblock;i++)d2val[i] = 0.;
      Forward(false,&m1,&d1);
  
      double dot1=0; //d1.innerProd(d2);
      for(llong i=mblock*1; i<(nt-1)*mblock;i++)dot1+=d1val[i]*d2val[i];

      Adjoint(false,&m2,&d2);
      double dot2=0; //m2.innerProd(m1);
      for(llong i=mblock*1; i<(nt-1)*mblock;i++)dot2+=m1val[i]*m2val[i];
      std::cout<<"Dot product test with add=false: \n"<<dot1<<"\n"<<dot2<<"\n";

      Forward(true,&m1,&d1);
      double dot3=0; //d1.innerProd(d2);
      for(llong i=mblock*1; i<(nt-1)*mblock;i++)dot3+=d1val[i]*d2val[i];
      Adjoint(true,&m2,&d2);
      double dot4=0.; //m2.innerProd(m1);
      for(llong i=mblock*1; i<(nt-1)*mblock;i++)dot4+=m1val[i]*m2val[i];


     std::cout<<"Dot product test with add=true: \n"<<dot3<<"\n"<<dot4<<"\n";

    }
}