#ifndef PROP_H
#define PROP_H 1

#include "base/Operator.h"
#include "base/DataInCoreFloat.h"
#include "base/inCoreFloat.h"
#include "base/Vector.h"

#include "Taper.h"
#include "Clean.h"
#include <cmath>
#include <iostream>

#include "Born.h"

typedef long long llong;

/*
  Wave Propagation module.
  
  The most important module.
  Designed for single shot propagation.
  Use carefully and have fun.
  
  First Created: 2015-11-04, Yinbin Ma
  Last Modified: 2015-11-04

*/
extern "C" void AddISPC(float *pin,float *pout,int n2);
extern "C" void ScaleISPC(float *pin,float *pout,int n2);
extern "C" void ProdISPC(float *coef, float *pin,float *pout,int n2);
extern "C" void ProdToISPC(float *coef, float *pin,float *pout,int n2);
extern "C" void ProdScaleToISPC(float coef, float *pin,float *pout,int n2);
namespace SEP{

class Prop: public Operator{
  public:
    Prop(Space *mod, Space *data);
    ~Prop(){}
    void SetPropPar(BornEngine *BornMod, llong _nb, float dt, float dx, Vector *Vel, int _sample=1);
    void SetModel(Vector *MK, Vector *MR);
    void SetBG(Vector *BG);
    virtual bool Forward(bool add, Vector *model, Vector *data);
    virtual bool Adjoint(bool add, Vector *model, Vector *data);
    virtual bool ForwardB(bool add, Vector *model, Vector *data, float *mask);
    virtual bool BackB(bool add, Vector *model, Vector *data, float *mask);
    void Dot();
    
  private:
    void init();
    llong nt, nz, nx, nb, sample, mblock;
    float dt2;
    float *Ls1, *Ls2, *Lr1, *Lr2;
        
    Vector *pre, *cur, *next,*temp, *qian, *hou, *tqian, *thou;
    Vector *Utemp;
    Vector *VT2;

    inCoreFloat *mspace, *uspace;
    
    BornEngine *BornMod;
    Taper *taper;
    CleanVec Zero;
    
    void PullData(bool add, Vector *mvec, Vector *uvec, llong position);
    void PushData(bool add, Vector *mvec, Vector *uvec, llong position);
    void CleanBound(Vector *mvec, Vector *uvec);               // use for high precision dot test
};




} // end of namespace




#endif