#include <iostream>
#include <cstdlib>
#include <stdbool.h>
#include <vector>
#include <cmath>
#include<string>
#include "seplib.h"

#include <omp.h>
#include "Laplacian.h"
#include "Source.h"
#include "SensorPad.h"
#include "Grad.h"
#include "Interpolation.h"
#include "Born.h"
#include "Prop.h"
#include "Engine.h"
#include "CreateModel.h"
#include "HessInv.h"
#include "Wrapper.h"
#include "NCG.h"
#include "utilities.h"
#include "mkl_rci.h"
#include "mkl_blas.h"
#include "mkl_spblas.h"
#include "mkl_service.h"
extern "C" {
#include "sep3d.h"
}
using namespace std;
using namespace SEP;



/*
Beta version for Multiparameter FWI

Generated by Yinbin: 2015-07-06

Assume: 
	1) source term
	2) initial condition with gradient = 0
	3) K[GPa], t[ms], rho[g/cm^3]
*/

int main(int argc, char *argv[]) {
  cout << "========================================"<<endl;
  cout << "                                        "<<endl;
  cout << "========================================"<<endl;
  cout << "              Test Modulus.        "<<endl;
  initpar(argc,argv);       // Initialize the sep library
  omp_set_num_threads(16);   // Initialize the parallel environment
  
  // Define dimension of interest
  int nx, nz, nt;
  int sx, sz, ns, oxs, dxs;
  int boundary;
  float dx, dz, dt;
  int nb, bsize;
  int sample;
  int mblock;
  int type;

   // Get parameters
  getch("nb","d",&nb);              cout << "	number of cores = " << nb << endl;
  getch("bsize","d",&bsize);        cout << "	jobs on each core = " << bsize << endl;
  ns = nb*bsize;
  getch("nx","d",&nx);              cout << "	nx = " << nx << endl;
  getch("nz","d",&nz);              cout << "	nz = " << nz << endl;
  getch("nt","d",&nt);              cout << "	nt = " << nt << endl;
  getch("boundary","d",&boundary);  cout << "	boundary = " << boundary << endl;
  mblock=nx*nz;
  getch("dx","f",&dx);              cout << "	dx = " << dx <<"[m]"<< endl;
  dz = dx;
  getch("sz","d",&sz);              cout << "	sz = " << sz << endl;
  getch("oxs","d",&oxs);            cout << "	oxs = " << oxs << endl;
  getch("dxs","d",&dxs);            cout << "	dxs = " << dxs << endl;
  getch("dt","f",&dt);              cout << "	dt = " << dt << "[ms]"<< endl;
  getch("ns","d",&ns);              cout << "	ns = " << ns << endl;
  getch("sample","d",&sample);      cout << "	sample = " << sample << endl;
  getch("type","d",&type);          cout << "	type = " << type << endl;
  // Set up space 
  inCoreFloat1D sspace(nt);               // space for source
  inCoreFloat2D mspace(nz,nx);            // space for velocity, reflectivity
  inCoreFloat2D bmspace(2*nz,nx);         // space for velocity, reflectivity
  inCoreFloat3D uspace(nt,nz,nx);         // space for wavefield
  inCoreFloat2D dspace(nt,nx);            // space for data
 
  
  // Set up vectors
  Vector *shot     = new Vector(&sspace);
  Vector *VT       = new Vector(&mspace); 
  Vector *RT       = new Vector(&mspace);
  Vector *VM       = new Vector(&mspace); 
  Vector *RM       = new Vector(&mspace);
  Vector *dV       = new Vector(&mspace); 
  Vector *dR       = new Vector(&mspace);
  Vector *HV       = new Vector(&mspace); 
  Vector *HR       = new Vector(&mspace);
  Vector *u1       = new Vector(&uspace);
  //Vector *u2       = new Vector(&uspace);
  //Vector *u3       = new Vector(&uspace);
 
 
  // Creating pointers
  float *__restrict__ shot_val  = ((dataInCoreFloat*)shot->getDataContainer())->getFloatPtr(); 
  float *__restrict__ VT_val    = ((dataInCoreFloat*)VT->getDataContainer())->getFloatPtr(); 
  float *__restrict__ VM_val    = ((dataInCoreFloat*)VM->getDataContainer())->getFloatPtr(); 
  float *__restrict__ RT_val    = ((dataInCoreFloat*)RT->getDataContainer())->getFloatPtr(); 
  float *__restrict__ RM_val    = ((dataInCoreFloat*)RM->getDataContainer())->getFloatPtr(); 
  float *__restrict__ dV_val    = ((dataInCoreFloat*)dV->getDataContainer())->getFloatPtr(); 
  float *__restrict__ dR_val    = ((dataInCoreFloat*)dR->getDataContainer())->getFloatPtr();  
  float *__restrict__ HV_val    = ((dataInCoreFloat*)HV->getDataContainer())->getFloatPtr(); 
  float *__restrict__ HR_val    = ((dataInCoreFloat*)HR->getDataContainer())->getFloatPtr();    
  float *__restrict__ u1_val    = ((dataInCoreFloat*)u1->getDataContainer())->getFloatPtr();   
  // Read in velocity and density
  cout << "========================== Reading in velocity and density model ==========================" <<endl;  
  sreed("wavelet",shot_val,sizeof(float)*nt); 
  sreed("VT",VT_val,sizeof(float)*nz*nx);
  sreed("VM",VM_val,sizeof(float)*nz*nx);
  sreed("RT",RT_val,sizeof(float)*nz*nx);
  sreed("RM",RM_val,sizeof(float)*nz*nx);
  int water = 0;//58*nx;

  float vwater = VM_val[50*nx+nx/2];
  float rwater = RM_val[50*nx+nx/2];
  cout << " velocity of water = " << vwater << ", density of water = " << rwater << endl; 
  /*
  Models *CModel =  new Models(nz, nx, 2.0, 1.5);
  CModel->Model0(VM,RM);
  CModel->Model1(VT,RT);
  water = 0;
  */
  //Constructing operators
  cout << "========================== Constructing operator ==========================" <<endl; 
  std::vector<int> sxv;
  std::vector<int> szv;
  for(int i=0; i < nb*bsize; i++){
    sxv.push_back(oxs+dxs*i);
    szv.push_back(sz);
  }
  Engine *Run = new Engine();
  Run->SetSpace(&mspace, &dspace, &uspace);
  Run->SetShot(shot,sxv,szv);
  Run->SetPropPar(nb,bsize,boundary,dt,dx,dz,sample,water);
  Run->SetTrueModel(VT,RT);
  Run->SetModel(VM,RM);
  //Run->CreateData();
  Run->GradientFunc(dV,dR);
  srite("GV",dV_val,sizeof(float)*nz*nx);
  srite("GR",dR_val,sizeof(float)*nz*nx);
  
  for(int i=0;i<nx*nz;i++){
    dV_val[i] = VT_val[i] - VM_val[i];
    dR_val[i] = RT_val[i] - RM_val[i];    
  }  
  Run->GaussHessianFunc(dV,dR,HV,HR);

  srite("GNV",HV_val,sizeof(float)*nz*nx);
  srite("GNR",HR_val,sizeof(float)*nz*nx);
  HV->zeroElement();
  HR->zeroElement();  
  Run->HessianFunc(dV,dR,HV,HR);
  srite("FNV",HV_val,sizeof(float)*nz*nx);
  srite("FNR",HR_val,sizeof(float)*nz*nx);
}
