#include <iostream>
#include <cstdlib>
#include <stdbool.h>
#include <vector>
#include <cmath>
#include<string>
#include "seplib.h"

#include <omp.h>
#include "Laplacian.h"
#include "Source.h"
#include "SensorPad.h"
#include "Grad.h"
#include "Interpolation.h"
#include "Born.h"
#include "Prop.h"
#include "Engine.h"
#include "CreateModel.h"
#include "HessInv.h"
#include "mkl_rci.h"
#include "mkl_blas.h"
#include "mkl_spblas.h"
#include "mkl_service.h"
extern "C" {
#include "sep3d.h"
}
using namespace std;
using namespace SEP;



/*
Beta version for Multiparameter FWI

Generated by Yinbin: 2015-07-06

Assume: 
	1) source term
	2) initial condition with gradient = 0
	3) K[GPa], t[ms], rho[g/cm^3]
*/

int main(int argc, char *argv[]) {
  cout << "========================================"<<endl;
  cout << "                                        "<<endl;
  cout << "========================================"<<endl;
  cout << "              Test Modulus.        "<<endl;
  initpar(argc,argv);       // Initialize the sep library
  omp_set_num_threads(8);   // Initialize the parallel environment
  
  // Define dimension of interest
  int nx, nz, nt;
  int sx, sz, ns, oxs, dxs;
  int boundary;
  float dx, dz, dt;
  int nb, bsize;
  int sample;
  int mblock;


   // Get parameters
  getch("nb","d",&nb);              cout << "	number of cores = " << nb << endl;
  getch("bsize","d",&bsize);        cout << "	jobs on each core = " << bsize << endl;
  ns = nb*bsize;
  getch("nx","d",&nx);              cout << "	nx = " << nx << endl;
  getch("nz","d",&nz);              cout << "	nz = " << nz << endl;
  getch("nt","d",&nt);              cout << "	nt = " << nt << endl;
  getch("boundary","d",&boundary);  cout << "	boundary = " << boundary << endl;
  mblock=nx*nz;
  getch("dx","f",&dx);              cout << "	dx = " << dx <<"[m]"<< endl;
  dz = dx;
  getch("sz","d",&sz);              cout << "	sz = " << sz << endl;
  getch("oxs","d",&oxs);            cout << "	oxs = " << oxs << endl;
  getch("dxs","d",&dxs);            cout << "	dxs = " << dxs << endl;
  getch("dt","f",&dt);              cout << "	dt = " << dt << "[ms]"<< endl;
  getch("ns","d",&ns);              cout << "	ns = " << ns << endl;
  getch("sample","d",&sample);      cout << "	sample = " << sample << endl;
 
  // Set up space 
  inCoreFloat1D sspace(nt);               // space for source
  inCoreFloat2D mspace(nz,nx);            // space for velocity, reflectivity
  inCoreFloat2D bmspace(2*nz,nx);         // space for velocity, reflectivity
  inCoreFloat3D uspace(nt,nz,nx);         // space for wavefield
  inCoreFloat2D dspace(nt,nx);            // space for data
 
  
  // Set up vectors
  Vector *shot     = new Vector(&sspace);
  Vector *VT       = new Vector(&mspace); 
  Vector *RT       = new Vector(&mspace);
  Vector *VM       = new Vector(&mspace); 
  Vector *RM       = new Vector(&mspace);
  Vector *dV       = new Vector(&mspace); 
  Vector *dR       = new Vector(&mspace);
  Vector *HV       = new Vector(&mspace); 
  Vector *HR       = new Vector(&mspace);
  Vector *u1       = new Vector(&uspace);
  Vector *u2       = new Vector(&uspace);
  Vector *u3       = new Vector(&uspace);
  Vector *MKT      = new Vector(&mspace); 
  Vector *MRT      = new Vector(&mspace);
  Vector *MKM      = new Vector(&mspace); 
  Vector *MRM      = new Vector(&mspace);  
 
  // Creating pointers
  float *__restrict__ shot_val  = ((dataInCoreFloat*)shot->getDataContainer())->getFloatPtr(); 
  float *__restrict__ VT_val    = ((dataInCoreFloat*)VT->getDataContainer())->getFloatPtr(); 
  float *__restrict__ VM_val    = ((dataInCoreFloat*)VM->getDataContainer())->getFloatPtr(); 
  float *__restrict__ RT_val    = ((dataInCoreFloat*)RT->getDataContainer())->getFloatPtr(); 
  float *__restrict__ RM_val    = ((dataInCoreFloat*)RM->getDataContainer())->getFloatPtr(); 
  float *__restrict__ dV_val    = ((dataInCoreFloat*)dV->getDataContainer())->getFloatPtr(); 
  float *__restrict__ dR_val    = ((dataInCoreFloat*)dR->getDataContainer())->getFloatPtr();  
  float *__restrict__ MKT_val   = ((dataInCoreFloat*)MKT->getDataContainer())->getFloatPtr(); 
  float *__restrict__ MKM_val   = ((dataInCoreFloat*)MKM->getDataContainer())->getFloatPtr(); 
  float *__restrict__ MRT_val   = ((dataInCoreFloat*)MRT->getDataContainer())->getFloatPtr(); 
  float *__restrict__ MRM_val   = ((dataInCoreFloat*)MRM->getDataContainer())->getFloatPtr(); 
  float *__restrict__ HV_val    = ((dataInCoreFloat*)HV->getDataContainer())->getFloatPtr(); 
  float *__restrict__ HR_val    = ((dataInCoreFloat*)HR->getDataContainer())->getFloatPtr();    
  float *__restrict__ u1_val    = ((dataInCoreFloat*)u1->getDataContainer())->getFloatPtr();   
  float *__restrict__ u2_val    = ((dataInCoreFloat*)u2->getDataContainer())->getFloatPtr();   
  float *__restrict__ u3_val    = ((dataInCoreFloat*)u3->getDataContainer())->getFloatPtr();   
  // Read in velocity and density
  cout << "========================== Reading in velocity and density model ==========================" <<endl;  
  sreed("wavelet",shot_val,sizeof(float)*nt); 
  for(int iz=0;iz<nz;iz++){
    for(int ix=0;ix<nx;ix++){
      VT_val[iz*nx+ix] = 1.5;
      VM_val[iz*nx+ix] = 1.5;
      RT_val[iz*nx+ix] = 1.;
      RM_val[iz*nx+ix] = 1.;
    }
  }  
  for(int i=0;i<mblock;i++){
  MKT_val[i] = 1./(RT_val[i]*VT_val[i]*VT_val[i]);
  MKM_val[i] = 1./(RM_val[i]*VM_val[i]*VM_val[i]);  
  MRT_val[i] = 1./RT_val[i];
  MRM_val[i] = 1./RM_val[i];
  }
  //Constructing operators
  cout << "========================== Constructing operator ==========================" <<endl; 
  std::vector<int> sxv;
  std::vector<int> szv;
  sx = 100;
  sz = 100;
  for(int i=0; i < nb*bsize; i++){
    sxv.push_back(sx);
    szv.push_back(sz);
  }

  BornEngine *BornMod = new BornEngine(&mspace,&uspace); 
  BornMod->SetPar(boundary, dt, dx,sample);
  
  Source *SourceInj = new Source(nt,nz,nx,sample);
  SourceInj->Injection(shot,u1,sx,sz);
  Prop *L = new Prop(&uspace, &uspace);
  L->SetPropPar(BornMod, boundary, dt, dx, MKT,sample);
  L->SetModel(MKT,MRT);
  
  L->Forward(false,u1,u2);
  cout << "u1 norm" << u1->innerProd(*u1) <<endl;
  cout << "u2 norm" << u2->innerProd(*u2) <<endl;
  srite("wave1",u2_val,sizeof(float)*nz*nx*nt);  

  //Constructing operators
  cout << "========================== Boundary operator ==========================" <<endl;  
  float *mask = new float[nx*nz];
  for(int i=0;i<nx*nz;i++) mask[i] = 1.;
  for(int iz=55;iz<65;iz++){
    for(int ix=0;ix<nx;ix++){
      mask[iz*nx+ix] = 0.;
    }
  }
  
  for(int iz=335;iz<345;iz++){
    for(int ix=0;ix<nx;ix++){
      mask[iz*nx+ix] = 0.;
    }
  }  
  for(int iz=0;iz<nz;iz++){
    for(int ix=55;ix<65;ix++){
      mask[iz*nx+ix] = 0.;
    }
  }
  
  for(int iz=0;iz<nz;iz++){
    for(int ix=335;ix<345;ix++){
      mask[iz*nx+ix] = 0.;
    }
  }  
  L->BackB(false,u3,u2,mask);
  srite("wave3",u3_val,sizeof(float)*nz*nx*nt);  
  return 0;
}
