#include "SensorPad.h"

namespace SEP {

SensorPad::SensorPad(Space *mod, Space *data, llong sz, llong bnx){
  setDomain(mod);
  setRange(data);
  init(sz, bnx);
}

void SensorPad::init(llong _sz, llong _bnx){
  inCoreFloat *wave = (inCoreFloat *)range;
  std::vector<long long> in_dims = wave->get_ns();
  nt = in_dims[0];
  nz = in_dims[1];
  nx = in_dims[2];
  sz = _sz;
  bnx = _bnx;
  sx = nx/2;
  nm = 0;  // empty sensor around source
  std::cout << "nt = " << nt << ", nx = " << nx << ", nz = "<< nz << ", sz = " << sz <<std::endl;
}

bool SensorPad::Forward(bool add, Vector *model, Vector *data){
  if(!add) Zero.Zero(data,nx*nz,nt);
  float *__restrict__ m = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ d = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();

  llong block = nx*nz;
  llong it, ix;  
  /*
  llong start = 0, end = nz;
  if(start < sx-nm) start = sx-nm;
  if(end > sx+nm) end = sx+nm;
    for(it=0; it<nt; it++){
    for(ix=start; ix<end; ix++){
      d[it*block + sz*nx+ix] += m[it*nx + ix];
    }
  }  
  */
  for(it=0; it<nt; it++){
    for(ix=bnx; ix<sx-nm; ix++){
      d[it*block + sz*nx+ix] += m[it*nx + ix];
    }
    for(ix=sx+nm; ix<nx-bnx; ix++){
      d[it*block + sz*nx+ix] += m[it*nx + ix];
    }
  }  
  
  return 0;
}


bool SensorPad::Adjoint(bool add, Vector *model, Vector *data){
  if(!add) Zero.Zero(model,nt,nx);
  float *__restrict__ m = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ d = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();
  
  llong block = nx*nz;
  llong it, ix;
  /*
  llong start = 0, end = nz;
  if(start < sx-nm) start = sx-nm;
  if(end > sx+nm) end = sx+nm;
      for(it=0; it<nt; it++){
    for(ix=start; ix<end; ix++){
      m[it*nx + ix] += d[it*block + sz*nx+ix];
    }
  } 
  */
  
  for(it=0; it<nt; it++){
    for(ix=bnx; ix<sx-nm; ix++){
      m[it*nx + ix] += d[it*block + sz*nx+ix];
    }
    for(ix=sx+nm; ix<nx-bnx; ix++){
      m[it*nx + ix] += d[it*block + sz*nx+ix];
    }
  }
  
  return 0;
}


}