#include <iostream>
#include <cstdlib>
#include <stdbool.h>
#include <vector>
#include <cmath>
#include<string>
#include "seplib.h"

#include <omp.h>
#include "Laplacian.h"
#include "Source.h"
#include "SensorPad.h"
#include "Grad.h"
#include "Interpolation.h"
#include "Born.h"
#include "Prop.h"
#include "Engine.h"
#include "CreateModel.h"
#include "HessInv.h"
#include "mkl_rci.h"
#include "mkl_blas.h"
#include "mkl_spblas.h"
#include "mkl_service.h"
extern "C" {
#include "sep3d.h"
}
using namespace std;
using namespace SEP;



/*
Beta version for Multiparameter FWI

Generated by Yinbin: 2015-07-06

Assume: 
	1) source term
	2) initial condition with gradient = 0
	3) K[GPa], t[ms], rho[g/cm^3]
*/
/*

void ConvertFtoD(float *input, double *output, int size){
  for(int i=0; i < size; i++){
    output[i] = (double)input[i];
  }
}
void ConvertDtoF(double *input, float *output, int size){
  for(int i=0; i < size; i++){
    output[i] = (float)input[i];
  }
}

double OBJF(float *resid,int n){
  double temp = 0.;
  for(int i=0; i < n; i++){
    temp += resid[i]*resid[i];
  }
}
*/
int main(int argc, char *argv[]) {
  cout << "========================================"<<endl;
  cout << "                                        "<<endl;
  cout << "========================================"<<endl;
  cout << "              Test Modulus.        "<<endl;
  initpar(argc,argv);       // Initialize the sep library
  omp_set_num_threads(8);   // Initialize the parallel environment
  
  // Define dimension of interest
  int nt=500, nx=301, nz = 201;
  int boundary = 5;
  int sample = 3;
  float dt=1.5;
  float dx=10., dz = 10.;
  int mblock = nx*nz;
  int nb=8;
  int bsize=3;

  // Set up space 
  inCoreFloat1D sspace(nt);               // space for source
  inCoreFloat2D mspace(nz,nx);            // space for velocity, reflectivity
  inCoreFloat2D bmspace(2*nz,nx);            // space for velocity, reflectivity
  inCoreFloat3D uspace(nt,nz,nx);         // space for wavefield
  inCoreFloat2D dspace(nt,nx);            // space for data
  
  // Set up vectors
  Vector *shot     = new Vector(&sspace);
  Vector *in1      = new Vector(&mspace);    in1->zeroElement();      // migration velocity
  Vector *out1     = new Vector(&mspace);    out1->zeroElement();
  Vector *Vel      = new Vector(&mspace); 
  Vector *rho      = new Vector(&mspace);
  Vector *Vel0     = new Vector(&mspace); 
  Vector *rho0     = new Vector(&mspace);
  Vector *Vel1     = new Vector(&mspace); 
  Vector *rho1     = new Vector(&mspace);
  Vector *dV     = new Vector(&mspace); 
  Vector *dr     = new Vector(&mspace);
  Vector *HV     = new Vector(&mspace); 
  Vector *Hr     = new Vector(&mspace);
  Vector *u1       = new Vector(&uspace);
  Vector *u2       = new Vector(&uspace);
  Vector *u3       = new Vector(&uspace);
  Laplacian4 *Laplacian = new Laplacian4(&mspace, &mspace);
  Laplacian->DotTest(true);

  float *__restrict__ in_val = ((dataInCoreFloat*)in1->getDataContainer())->getFloatPtr();
  float *__restrict__ out_val = ((dataInCoreFloat*)out1->getDataContainer())->getFloatPtr();
  for(int i=0; i < 1000; i ++){
    in1->random();
    Laplacian->Forward(false,in1,out1);
  }
  for(int i=0; i < 1000; i ++){
    in1->random();
    Laplacian->Adjoint(false,out1,in1);
  }
  
  // Test for Source
  cout << "========================== Testing Source Injection Function ==========================" <<endl; 
  Source *SS = new Source(nt,nz,nx,sample);
  float *__restrict__ shot_val = ((dataInCoreFloat*)shot->getDataContainer())->getFloatPtr(); 
  sreed("wavelet",shot_val,sizeof(float)*nt); 
  SS->Injection(shot,u1,100,100);

  // Test for Sensor
  cout << "========================== Testing Sensor Padding Function ==========================" <<endl;   
  SensorPad *Sensor = new SensorPad(&dspace, &uspace,60);
  Sensor->DotTest(true);
  
  // Test for Gradient
  cout << "========================== Testing Gradient Function ==========================" <<endl;   
  GradX *DX = new GradX(&mspace, &mspace);
  DX->DotTest(true);  
  GradZ *DZ = new GradZ(&mspace, &mspace);
  DZ->DotTest(true);    
  GradXT *DXT = new GradXT(&mspace, &mspace);
  DXT->DotTest(true);  
  GradZT *DZT = new GradZT(&mspace, &mspace);
  DZT->DotTest(true);  
  
  // Test for Gradient
  cout << "========================== Testing Interpolation Function ==========================" <<endl;
  IntX *IntPX = new IntX(&mspace,&mspace);
  IntPX->DotTest(true);
  IntZ *IntPZ = new IntZ(&mspace,&mspace);
  IntPZ->DotTest(true);   

  // Test for Born Approximation
  cout << "========================== Testing Born Approximation ==========================" <<endl;  
  float *__restrict__ vval = ((dataInCoreFloat*)Vel->getDataContainer())->getFloatPtr();
  float *__restrict__ rval = ((dataInCoreFloat*)rho->getDataContainer())->getFloatPtr();  
  rho->random(); 
  for(int i=0; i < nx*nz;i++){
    vval[i] = 2.;
    rval[i] = 1.+0.0*rval[i];
  } 
  u1->random(); 
  BornEngine *BornMod = new BornEngine(&mspace,&uspace); 
  BornMod->SetPar(boundary, dt, dx,sample);
  BornMod->SetModel(Vel,rho);
  BornMod->SetBG(u1);


  BornT *BT = new BornT(&mspace, &uspace);
  BT->SetEngine(BornMod);
  BT->DotTest(true);
  BornMR *BMR = new BornMR(&mspace, &uspace);
  BMR->SetEngine(BornMod);
  BMR->DotTest(true);
  BornMod->DotTest(true);

  // Test for wave propagation
  cout << "========================== Testing Wave Propagation ==========================" <<endl;  


  Prop *L = new Prop(&uspace, &uspace);
  L->SetPropPar(BornMod, boundary, dt, dx, Vel,sample);
  L->SetModel(Vel,rho);
  //L->DotTest(true);
  //L->Dot();

  cout << "========================== Testing Sensitivity kernel ==========================" <<endl;
  float *__restrict__ u1val = ((dataInCoreFloat*)u1->getDataContainer())->getFloatPtr();      
  float *__restrict__ u2val = ((dataInCoreFloat*)u2->getDataContainer())->getFloatPtr();  
  float *__restrict__ u3val = ((dataInCoreFloat*)u3->getDataContainer())->getFloatPtr();   
  float *__restrict__ uval = ((dataInCoreFloat*)u2->getDataContainer())->getFloatPtr(); 
  /*
  u1->zeroElement();   
  for(int i=0; i < nt; i++){
    u1val[i*mblock + 100*nx+150] = shot_val[i];
  }
  L->Forward(false, u1, u2);
  u1->zeroElement();
  for(int i=0; i < nt; i++){
    u1val[i*mblock + 200*nx+150] = u2val[i*mblock + 200*nx+150];
  }
  L->Adjoint(false,u3,u1);
  BornMod->SetBG(u2);
  BornMod->BornRhoAdjoint(false,Vel1,u3);
  */
  
  // Test for wave propagation
  cout << "========================== Testing Forward Wave Propagation ==========================" <<endl;    
  //L->Forward(false, u1, u2);
  //cout << "wave norm = " << u2->innerProd(*u2) <<endl;




  // Test for Gradient
  cout << "========================== Testing Gradient ==========================" <<endl; 
  float *__restrict__ GVval = ((dataInCoreFloat*)Vel1->getDataContainer())->getFloatPtr();
  float *__restrict__ Grval = ((dataInCoreFloat*)rho1->getDataContainer())->getFloatPtr();

  Models *CModel =  new Models(nz, nx, 2.0, 1.5);
  CModel->Model0(Vel0,rho0);
  CModel->Model1(Vel1,rho1);
  dV->linearCombo(0.,*Vel1,1.);dV->linearCombo(1.,*Vel0,-1.);
  dr->linearCombo(0.,*rho1,1.);dr->linearCombo(1.,*rho0,-1.);
  L->SetModel(Vel1,rho1);
  L->Forward(false,u1,u2);
   
  srite("ModV",GVval,sizeof(float)*nx*nz);
  srite("Modr",Grval,sizeof(float)*nx*nz);
  std::vector<int> sxv;
  std::vector<int> szv;
  for(int i=0; i < 8*3; i++){
    sxv.push_back(80+6*i);
    szv.push_back(65);
  }
  Engine *Run = new Engine();
  Run->SetSpace(&mspace, &dspace, &uspace);
  Run->SetShot(shot,sxv,szv);
  Run->SetPropPar(nb,bsize,boundary,dt,dx,dz,sample);
  Run->SetTrueModel(Vel1,rho1);
  Run->SetModel(Vel0,rho0);
  cout << "    Migration data collected.  " <<endl;
  cout << "========================== Computating gradient ==========================" <<endl; 
  dV->zeroElement(); dr->zeroElement();
  Run->GradientFunc(dV,dr);
  
  
  cout << "========================== Computating Hessian ==========================" <<endl;  
  HessInv *HInv = new HessInv(&mspace, Run);
  HInv->Apply(dV,dr,HV,Hr,10);
  
  srite("waveF",u2val,sizeof(float)*nt*nz*nx);
  srite("waveA",u3val,sizeof(float)*nt*nz*nx);
  srite("wave",uval,sizeof(float)*nt*nz*nx);
  //srite("data",dataRes,sizeof(float)*5*nx*nt);
  srite("GradV",GVval,sizeof(float)*nx*nz);    
  srite("Gradr",Grval,sizeof(float)*nx*nz);
  
  // Test for Hessian
  cout << "========================== Testing Hessian ==========================" <<endl; 
  float *__restrict__ HVval = ((dataInCoreFloat*)HV->getDataContainer())->getFloatPtr();
  float *__restrict__ Hrval = ((dataInCoreFloat*)Hr->getDataContainer())->getFloatPtr();

  srite("HessV",HVval,sizeof(float)*nx*nz);    
  srite("Hessr",Hrval,sizeof(float)*nx*nz);

  //Hess *HessOP = new Hess(&mspace,&bmspace,&bmspace,Run);
  //HessOP->DotTest(true);
  
  // Test for the inverse of Hessian
  cout << "========================== Testing Inverse of Hessian ==========================" <<endl; 

  float *__restrict__ dVval = ((dataInCoreFloat*)dV->getDataContainer())->getFloatPtr();
  float *__restrict__ drval = ((dataInCoreFloat*)dr->getDataContainer())->getFloatPtr();
/*
  MKL_INT nm=nx*nz, n=2*nx*nz, rci_request, itercount, expected_itercount=100,i;
  MKL_INT length = 128;
  //double *expected_sol = new double[n];
  //ConvertFtoD(x_val,expected_sol,n);
  double *rhs = new double[n];
  float *resid = new float[n];
  
  ConvertFtoD(GVval,rhs,nm);
  ConvertFtoD(Grval,&rhs[nm],nm);
  double *solution = new double[n];for(int i=0; i<n;i++)solution[i]=0.;
  double *tmp = new double[n*4];for(int i=0; i<n*4;i++)tmp[i]=0.;
  double *dpar= new double[128];
  MKL_INT *ipar = new MKL_INT[128];
  dcg_init(&n,solution,rhs,&rci_request,ipar,dpar,tmp);
  ipar[8] = 1;
  ipar[9] = 0;
  dpar[0] = 1.e-6;
  dcg_check(&n,solution,rhs,&rci_request,ipar,dpar,tmp);
  dV->zeroElement(); dr->zeroElement();
  HV->zeroElement(); Hr->zeroElement();
  double rtol = 0.25;
  //double t0 = timer();
  int niter = 40;
  for(int iter = 0 ; iter < niter; iter++){
    cout<<"iter = " << iter ;
    
    dcg(&n,solution,rhs,&rci_request,ipar,dpar,tmp);
    ConvertDtoF(tmp,dVval,nm);
    ConvertDtoF(&tmp[nm],drval,nm);
    if(rci_request == 0){
      break;
    }
    Run->HessianFunc(dV,dr,HV,Hr);
    //Ktmp->linearCombo(1.,*dK,1.e-4);
    //mul->Forward(false,xtmp,btmp);
    ConvertFtoD(HVval,&tmp[n],nm);
    ConvertFtoD(Hrval,&tmp[n+nm],nm);

    ConvertDtoF(&tmp[2*n],resid,n);
    std::cout << ", HV norm = " << HV->innerProd(*HV) << std::endl;
    std::cout << ", residue = " << OBJF(resid,n) << std::endl;
    HV->zeroElement(); Hr->zeroElement();
    
  }
  */
  srite("InvV",dVval,sizeof(float)*nx*nz);    
  srite("Invr",drval,sizeof(float)*nx*nz);

  
  /*
  // Test for wave propagation
  cout << "========================== Testing Propagation Engine ==========================" <<endl;  
  std::vector<int>sxv(10);
  std::vector<int>szv(10);  
  Engine *Run =  new Engine();
  Run->SetSpace(&mspace,&dspace,&uspace);
  Run->SetShot(shot,sxv,szv);
  Run->SetPropPar(6,4,boundary, dt, dx, dz, sample);
  Run->SetModel(Vel,rho);
  Run->CreateData();
  */

  return 0;
}
