#ifndef TIMEMARCH_H
#define TIMEMARCH_H 1
#include "base/Operator.h"
#include "base/DataInCoreFloat.h"
#include "base/inCoreFloat.h"
#include "Laplacian.h"
#include "Grad.h"
#include "Interpolation.h"
typedef long long llong;

/*
  Created by Yinbin Ma
  Time marching operator Unext = F(Ucur, Uprev) without taper/source injection.
  Use carefully and have fun.
  
  First Created: 2015-10-30
  Last Modified: 2015-10-30

*/
extern "C" void AddISPC(float *pin,float *pout,int n2);
extern "C" void ScaleISPC(float *pin,float *pout,int n2);
extern "C" void ProdISPC(float *coef, float *pin,float *pout,int n2);
extern "C" void ProdToISPC(float *coef, float *pin,float *pout,int n2);
extern "C" void LeapFrogISPC(float *coef,float *p1,float *px,float *pz,float *p2,float *p3,int n);

namespace SEP{

class TimeMarch{
  public:
    TimeMarch(llong nb, float dt, float dx, Vector *Vel);
    ~TimeMarch(){};
    void SetModel(Vector *Vel, Vector *rho);
    bool FMarch(Vector *pre, Vector *cur, Vector *next);
    bool BMarch(Vector *pre, Vector *cur, Vector *next);
    
  private:
    void init(llong nb, float dt, float dx, Vector *Vel);
    inCoreFloat *mspace;
    Vector *VTX, *rhoInvX, *rhoInvZ, *rhoReg, *rhoInv;
    Vector *temp, *tempX, *tempZ;
    llong nz, nx, nb, mblock;
    float dx, dt, dtdx2;
    GradX *DX;
    GradZ *DZ;
    GradXT *DXT;
    GradZT *DZT;
    IntX *interpX;
    IntZ *interpZ;
    Laplacian4 *Lap;
};


}
#endif