#include "ADMMAlg.h"
#include <cmath>
#include <iostream>
#include <algorithm>


namespace SEP {

void Shrink::init(double _threshold) {
    threshold = _threshold;
    inCoreFloat *temp = (inCoreFloat *)z.getSpace();
    d = ((dataInCoreFloat*)z.getDataContainer())->getFloatPtr();
    n123 = temp->get_n123();
}


void Shrink::run() {
  double temp;
  float pos = 1., neg = 1.;
  double zero = 0.;
  long long zeros = 0;
  for(long long i = 0; i < n123; i++) {
    temp = d[i];
    if (temp >0) {
      d[i] = pos*std::max(zero,temp - threshold);
    }
    else {
      d[i] = neg*std::max(zero,-temp - threshold);
    }
    if(d[i] < 1.e-9 and d[i] > -1.e-9) zeros++;
  }
  std::cout <<"zeros(z) = " << 100.*((double)zeros)/(double)n123 <<std::endl;
}


/*
Below is ADMM algorithm components
*/


void ADMM_Solver::init(double _lambda, double _rho) {
        lambda = _lambda; rho = _rho;
        niter = 0;
        mspace   = (inCoreFloat *)m.getSpace();
        dspace   = (inCoreFloat *)L.getRange();
        zspace   = (inCoreFloat *)A.getRange();

        z = new Vector(zspace);
        u = new Vector(zspace);

        rhos = std::sqrt(rho);
        maxiter = 10; rtol = 1.e-6; nrtol = 1.e-6; maxCG = 1; maxInner = 1;//Default
        rnorm = 1.; nrnorm = 1.;
}

void ADMM_Solver::SetPar(int _maxiter,double _rtol, double _nrtol) {
  maxiter = _maxiter;
  rtol    = _rtol;
  nrtol   = _nrtol;
}
void ADMM_Solver::SetSPPar(int _maxInner, int _maxCG) {
  maxInner = _maxInner; 
  maxCG    = _maxCG;
}

bool ADMM_Solver::stop() {
  if (rnorm < rtol) return true;
  if (nrnorm < nrtol) return true;
  return false;
} 

void ADMM_Solver::run() {
      // Setup the operators
      ScaleOP scale(zspace,zspace,rhos);
      ChainOP   Reg(A,scale);
      // Setup the shrinkage function
      Shrink shrinking(*z,lambda/rho);
      for(int i = 0; i < maxiter; i++) {
        if(!stop()) {
          std::cout << "Iter: " << i << std::endl; 
          for(int j = 0; j < maxInner; j++) {
            int maxCGIter = niter + maxCG;
            z->linearCombo(rhos,*u,-rhos);
            CGRegAlg cg(L,Reg,m,d,*z,rnorm,nrnorm,niter,rtol,nrtol,maxCGIter,true);
            cg.run();
            A.Forward(false,&m,z);   // 
            z->linearCombo(1.,*u,1.);   // z = Am + u;
            shrinking.run();          // shrink(z,lambda/rho);
          }
          A.Forward(false,&m,u);   // put true back later
          u->linearCombo(1.,*z,-1.);

        }
        else break;
      } 
}



}