// CGAlg.H
// Created by YM
/*
Steal the idea from RVL. Then copied CGStep.h
Please check the following link for reference
http://www.trip.caam.rice.edu/software/rvl/umin/doc/html/cgalg_8hh_source.html


*/

#ifndef CGALG_H
#define CGALG_H 1

#include "Space.h"
#include "Vector.h"
#include "Operator.h"
#include "Algorithm.h"
#include <iostream>
#include <cmath>
namespace SEP {

class CGStepper: public Algorithm {
  public:

    CGStepper(Vector &mod, Vector &data,Vector &_g, Vector &_G, bool _verbose = false)
        : x(mod), r(data), g(_g),G(_G), verbose(_verbose) {

      s.clone(&mod); s.zeroElement(); 
      S.clone(&data); S.zeroElement();
      iter = 0;
    }

    virtual void run();

    ~CGStepper() {}

  private:
    int iter;
    Vector &x, &r;
    Vector &g, &G;
    Vector s, S; // updates to model and gradient
    bool verbose;

};

class CGRegStepper:public Algorithm {
  public:
    CGRegStepper(Vector &mod, Vector &data1, Vector &data2,Vector &_g, Vector &_G1, Vector &_G2, bool _verbose = false)
        :x(mod),r1(data1),r2(data2),g(_g),G1(_G1),G2(_G2),verbose(_verbose) {
      s.clone(&mod);     s.zeroElement();
      S1.clone(&data1); S1.zeroElement();
      S2.clone(&data2); S2.zeroElement();
      iter = 0;
    }
    virtual void run();

    ~CGRegStepper() {}
  private:
    int iter;
    Vector &x, &r1, &r2;
    Vector &g, &G1, &G2;
    Vector s, S1, S2;
    bool verbose;

};




class CGAlg: public Algorithm{

  public:
    CGAlg(Operator & _A, 
        Vector & _x,
        Vector & _rhs,
        double & _rnorm,
        double & _nrnorm,
        int    & _iter,
        double _rtol = 1.e-6,
        double _nrtol = 1.e-6,
        int    _maxiter = 10,
        bool   _verbose = false,
        float * _data_res = NULL
        )
      :A(_A), x(_x), b(_rhs), rnorm(_rnorm), nrnorm(_nrnorm),iter(_iter), rtol(_rtol),nrtol(_nrtol), maxiter(_maxiter),verbose(_verbose)
      {
   ///  rnorm is the norm of residue in data space
   /// nrnorm is the norm of gradient in model space
        r.clone(&b);
        r.addInv();
        A.Forward(true,&x,&r); 
        g.clone(&x);   g.zeroElement();
        G.clone(&r); G.zeroElement();
        data_res = _data_res;
      }

    ~CGAlg() {}

    bool query();
    void disp();
    void run();

  private:

    Operator       & A;    // Operator
    Vector         & x;    // solution vector
    Vector         & b;
    Vector         r;
    Vector         g, G;
    double         & rnorm;
    double         & nrnorm;
    int            & iter;
    double         rtol;   // tolerance of residual norm
    double         nrtol;  // tolerance of gradient norm
    int            maxiter;
    bool           verbose;
    float         * data_res;
};


class CGRegAlg: public Algorithm{

  public:
    CGRegAlg(Operator & _A1, Operator &_A2, 
        Vector & _x,
        Vector & _rhs1, Vector &_rhs2,
        double & _rnorm,
        double & _nrnorm,
        int    & _iter,
        double _rtol = 1.e-6,
        double _nrtol = 1.e-6,
        int    _maxiter = 10,
        bool   _verbose = false,
        float * _data_res = NULL
        )
      :A1(_A1), A2(_A2), x(_x), b1(_rhs1), b2(_rhs2),rnorm(_rnorm), nrnorm(_nrnorm),iter(_iter), rtol(_rtol),nrtol(_nrtol), maxiter(_maxiter),verbose(_verbose)
      {
   ///  rnorm is the norm of residue in data space
   /// nrnorm is the norm of gradient in model space
        r1.clone(&b1); r2.clone(&b2);
        r1.addInv(); r2.addInv();
        A1.Forward(true,&x,&r1);
        A2.Forward(true,&x,&r2); 
        g.clone(&x);   g.zeroElement();
        G1.clone(&r1); G1.zeroElement();
        G2.clone(&r2); G2.zeroElement();
        data_res = _data_res;
      }

    ~CGRegAlg() {}

    bool query();
    void disp();
    void run();

  private:

    Operator       & A1, &A2;    // Operator
    Vector         & x;    // solution vector
    Vector         & b1, &b2;
    Vector         r1,r2;
    Vector         g, G1,G2;
    double         & rnorm;
    double         & nrnorm;
    int            & iter;
    double         rtol;   // tolerance of residual norm
    double         nrtol;  // tolerance of gradient norm
    int            maxiter;
    bool           verbose;
    float         * data_res;
};



}

#endif