// CGLSAlg.H
// Created by YM
/*
Steal the idea from RVL.
Please check the following link for reference
http://www.trip.caam.rice.edu/software/rvl/umin/doc/html/cgalg_8hh_source.html



*/

#ifndef CGLSALG_H
#define CGLSALG_H 1

#include "Space.h"
#include "Vector.h"
#include "Operator.h"
#include "Algorithm.h"
#include <iostream>
#include <cmath>
namespace SEP {

class CGLSStep: public Algorithm {
  public:

    CGLSStep(Operator & _A, Vector & _x, Vector  & _b, 
         double & _rnorm, double & _nrnorm)
      :A(_A), x(_x), b(_b), rnorm(_rnorm),nrnorm(_nrnorm) {
      s.clone(&x);
      /*
      q.clone(&b);q.zeroElement();
      r.clone(&b);
      */
      
      q.clone(&b);q.zeroElement();
      r.clone(&b);r.addInv();
      A.Forward(true,&x,&r); r.addInv();
      
      rnorm = std::sqrt(r.innerProd(r));
      A.Adjoint(false,&s,&r);
      p.clone(&s);
      gamma = s.innerProd(s);
      nrnorm = std::sqrt(gamma);

    }

    void run() {

      A.Forward(false,&p,&q);
      double qtq = q.innerProd(q);
      double alpha = gamma/qtq;
      x.linearCombo(1.,p,(float)alpha);
      r.linearCombo((float)1.,q,(float)(alpha*(-1.)));
      A.Adjoint(false,&s,&r); 
      double newgamma = s.innerProd(s);
      double beta = newgamma/gamma;
      p.linearCombo((float)beta,s,1.);
      gamma = newgamma;
      rnorm = std::sqrt(r.innerProd(r));
      nrnorm = std::sqrt(gamma);

    }

    ~CGLSStep() {}

  private:
    // reference to external objects

    Operator & A;
    Vector & x;
    Vector  & b;
    double & rnorm;
    double & nrnorm;

    // need to construct the following for CGLSStep
    Vector r;
    Vector s;
    Vector q;
    Vector p;
    double gamma;
};



class CGLSAlg: public Algorithm{

  public:
    CGLSAlg(Operator & _A, 
        Vector & _x,
        Vector & _rhs,
        double & _rnorm,
        double & _nrnorm,
        double _rtol = 1.e-6,
        double _nrtol = 1.e-6,
        int    _maxiter = 10,
        bool   _verbose = false)
      :A(_A), x(_x), rhs(_rhs), rnorm(_rnorm), nrnorm(_nrnorm), rtol(_rtol),nrtol(_nrtol), maxiter(_maxiter),verbose(_verbose),step(A,x,rhs,rnorm,nrnorm)
      { //x.zeroElement();
       niter = 0;}

    ~CGLSAlg() {}

    bool query() {
      if (rnorm < rtol) return true;
      if (nrnorm < nrtol) return true;
      if (niter >= maxiter) return true;
      return false;
    } 
    void disp() {
        niter++;
        if(verbose) {
          std::cout<<"Iter = " << niter << ", rnorm = "<<rnorm << ", nrnorm = " << nrnorm << std::endl;
        }
    }

    void run() {
      while (!query()) {
        step.run();
        disp();
      }
     
    }

  private:

    Operator       & A;    // Operator
    Vector         & x;    // solution vector
    Vector         & rhs;
    double         & rnorm;
    double         & nrnorm;
    double         rtol;   // tolerance of residual norm
    double         nrtol;  // tolerance of gradient norm
    int            niter;
    int            maxiter;
    bool           verbose;
    CGLSStep       step;

};


}

#endif