#include "CGStep.h"
#include "math.h"
#include <iostream>
#include <algorithm>

void SEP::CGStep::Alloc_step(Vector &mod, Vector &data) {
  // create s=0 same size as mod
  // create ss=0 same size as data

  s.clone(&mod);
  s.zeroElement(); 

  ss.clone(&data);
  ss.zeroElement();
}

bool SEP::CGStep::Step(int iter, Vector &m, Vector &g, Vector &rr, Vector &gg, double &res, bool verbose){

  /// As a certain iteration, this updates model m with gradient g. The previous
  /// updates to m and g are rr and gg. res is the residual (Am-data).
  /// If you want an output of iter #, alpha, beta (step scale), add optional 
  /// verbose = true. By default verbose = false.

  double alpha,beta=0,small=1e-20;  // building array size 10 here? 

  double gg_gg= (gg).innerProd(gg); 
  double gg_rr=-(gg).innerProd(rr);

  if(iter==0){
    alpha = gg_rr/gg_gg;
    if(fabs(gg_gg) < small){
      fprintf(stderr,"Zero dot product in first CG step %g %g\n",gg_gg,gg_rr);
      return false;
    }
  }
  else{
    double ss_ss=(ss).innerProd(ss);
    double gg_ss=(gg).innerProd(ss); 
    if(fabs(gg_gg) < small || fabs(ss_ss) < small){
      fprintf(stderr,"Zero dot product in CG step %g %g / %g %g \n",gg_gg,gg_rr,ss_ss,gg_ss);
      return false;
    }

    double determ=gg_gg*ss_ss*std::max( 1.-(gg_ss/gg_gg)*(gg_ss/ss_ss) , 1e-12 );
    double ss_rr=-(ss).innerProd(rr);
    alpha = (ss_ss*gg_rr - gg_ss*ss_rr)/determ;
    beta = (-gg_ss*gg_rr + gg_gg*ss_rr)/determ;
  }

  if(verbose){
    std::cout << "iter: " << iter << " , alpha = " << alpha << ", beta = " << beta << "\n";
  }

  (s).linearCombo(beta,g,alpha);   //Update solution step alpha*g+beta*s
  (ss).linearCombo(beta,gg,alpha); //Update residual step

  (m).linearCombo(1.0,s,1.0);      //Update solution
  (rr).linearCombo(1.0,ss,1.0);    //Update residual

   res = (rr).innerProd(rr);
  
  return true;

}
