#ifndef DATAINCOREFLOAT_H
#define DATAINCOREFLOAT_H 1
#include "dataContainer.h"
#include <boost/multi_array.hpp> 
//#include <omp.h>
#include "myomp.h"

typedef boost::multi_array<float, 1> float_1d;
typedef boost::multi_array<float, 2> float_2d;
typedef boost::multi_array<float, 3> float_3d;
typedef boost::multi_array<float, 4> float_4d;
typedef boost::multi_array<float, 5> float_5d;
typedef boost::multi_array<float, 6> float_6d;
typedef boost::multi_array<float, 7> float_7d;

namespace SEP{

class dataInCoreFloat: public dataContainer {
  public:
    dataInCoreFloat() {} // Default constructor
    void setup_omp();
    std::vector<long long> *get_b_omp() {return &b_omp;}
    std::vector<long long> *get_n_omp() {return &n_omp;}
    float *getFloatPtr() {return vals;}
    void set_n123(long long n123) {this->n123=n123;}
    int get_nth() {return nth;}
    void random(dataContainer *x) const;
    virtual void my_type() = 0; // needed for this to be abstract class
    virtual ~dataInCoreFloat(){
      cleanInCoreFloat();
    }
  public:
    float *vals;
  private:
     std::vector<long long> b_omp,n_omp;
     void cleanInCoreFloat();
     long long n123;
     int nth;
     int vector_beg;
};

class data1DFloat: public dataInCoreFloat{
  public:
    data1DFloat(long long n){ 
      base_1d(n); 
      strcpy(name,"1dfloat\n");
    }
    virtual void *returnData(){return (void*)array;}
    virtual void my_type(){ std::cerr<<"I am dataInCoreFloat1D "<<std::endl;}
    virtual ~data1DFloat(){ clean_1d();}
  private:
    void base_1d(long long n);
    void clean_1d();
    float_1d *array;
};

class data2DFloat: public dataInCoreFloat{
  public:
    data2DFloat(long long n1, long long n2){ 
      base_2d(n1,n2);
    }
    data2DFloat(std::vector<long long>n){ 
      base_2d(n[0],n[1]);
    }
    data2DFloat(long long *n){ 
      base_2d(n[0],n[1]);
    }
    virtual void *returnData(){return (void*)array;}
    virtual void my_type(){ std::cerr<<"I am dataInCoreFloat2D "<<std::endl;}
    virtual ~data2DFloat(){ delete array; array = NULL;}
  private:
    void base_2d(long long n1,long long n2);
    float_2d *array;
};

class data3DFloat: public dataInCoreFloat{
  public:
    data3DFloat(long long n1, long long n2, long long n3){ 
      base_3d(n1,n2,n3);
    }
      data3DFloat(std::vector<long long>n){ 
      base_3d(n[0],n[1],n[2]);
    }
    data3DFloat(long long *n){ 
      base_3d(n[0],n[1],n[2]);
    }
    virtual void *returnData(){return (void*)array;}
    virtual void my_type(){ std::cerr<<"I am dataInCoreFloat3D "<<std::endl;}
    virtual ~data3DFloat(){ delete array; array = NULL;}
  private:
    void base_3d(long long n1,long long n2, long long n3);
    float_3d *array;
};

class data4DFloat: public dataInCoreFloat{
  public:
    data4DFloat(long long n1, long long n2, long long n3, long long n4){ 
      base_4d(n1,n2,n3,n4);
    }
    data4DFloat(std::vector<long long>n){ 
      base_4d(n[0],n[1],n[2],n[3]);
    }
    data4DFloat(long long *n){ 
      base_4d(n[0],n[1],n[2],n[3]);
    }
    virtual void *returnData(){return (void*)array;}
    virtual void my_type(){ std::cerr<<"I am dataInCoreFloat4D "<<std::endl;}
    virtual ~data4DFloat(){ delete array; array = NULL;}
  private:
    void base_4d(long long n1,long long n2, long long n3, long long n4);
    float_4d *array;
};

class data5DFloat: public dataInCoreFloat{
  public:
    data5DFloat(long long n1, long long n2, long long n3, long long n4, long long n5){ 
      base_5d(n1,n2,n3,n4,n5);
    }
    data5DFloat(std::vector<long long>n){ 
      base_5d(n[0],n[1],n[2],n[3],n[4]);
    }
    data5DFloat(long long *n){
      base_5d(n[0],n[1],n[2],n[3],n[4]);
    }
    virtual void *returnData(){return (void*)array; array = NULL;}
    virtual void my_type(){ std::cerr<<"I am dataInCoreFloat5D "<<std::endl;}
    virtual ~data5DFloat(){ delete array; array = NULL;}

  private:
    void base_5d(long long n1,long long n2, long long n3, long long n4, long long n5);
    float_5d *array;
};

class data6DFloat: public dataInCoreFloat{
  public:
    data6DFloat(long long n1, long long n2, long long n3, long long n4, long long n5, long long n6){ 
      base_6d(n1,n2,n3,n4,n5,n6);
    }
    data6DFloat(std::vector<long long>n){ 
      base_6d(n[0],n[1],n[2],n[3],n[4],n[5]);
    }
    data6DFloat(long long *n){ 
      base_6d(n[0],n[1],n[2],n[3],n[4],n[5]);
    }
    virtual void *returnData(){return (void*)array;}
    virtual void my_type(){ std::cerr<<"I am dataInCoreFloat6D "<<std::endl;}
    virtual ~data6DFloat(){ delete array; array = NULL;}
  private:
    void base_6d(long long n1,long long n2, long long n3, long long n4, long long n5, long long n6);
    float_6d *array;
};

class data7DFloat: public dataInCoreFloat{
  public:
    data7DFloat(long long n1, long long n2, long long n3, long long n4, long long n5, long long n6, long long n7){ 
      base_7d(n1,n2,n3,n4,n5,n6,n7);
    }
    data7DFloat(std::vector<long long>n){ 
      base_7d(n[0],n[1],n[2],n[3],n[4],n[5],n[6]);
    }
    
    data7DFloat(long long *n){ 
      base_7d(n[0],n[1],n[2],n[3],n[4],n[5],n[6]);
    }
    virtual void *returnData(){return (void*)array;}
    virtual void my_type(){ std::cerr<<"I am dataInCoreFloat7D "<<std::endl;}
    virtual ~data7DFloat(){ delete array; array = NULL;}
  private:
    void base_7d(long long n1,long long n2, long long n3, long long n4, long long n5, long long n6, long long n7);
    float_7d *array;
};

}
#endif
