// HBAlg.H
// Created by YM
/*
Using Hyperbolic norm.
Implement cg stepper here, as we are going to do plane search several times.

*/

#ifndef HBALG_H
#define HBALG_H 1

#include "Space.h"
#include "Vector.h"
#include "Operator.h"
#include "MultiOperator.h"
#include "Algorithm.h"
#include "CGAlg.h"
#include "IRLS_reg.h"
#include <iostream>
#include <cmath>

namespace SEP {

class CGstep_ps: public Algorithm {
  public:

    CGstep_ps(Vector &mod, Vector &_rdata, Vector &_rreg,Vector &_g, Vector &_Gdata, Vector &_Greg,long long _nd, long long _nr, float _eps,bool _verbose = false)
        : x(mod), rdata(_rdata),rreg(_rreg), g(_g),Gdata(_Gdata),Greg(_Greg), verbose(_verbose) {

      s.clone(&mod); s.zeroElement(); 
      Sdata.clone(&rdata); Sdata.zeroElement();
      Sreg.clone(&rreg); Sreg.zeroElement();
      hp.clone(&rreg);   hp.zeroElement();
      hpp.clone(&rreg);  hpp.zeroElement();
      nd = _nd;
      nr = _nr;
      eps = _eps;
      iter = 0;
      maxps= 10;
      tol_ps = 1.e-5;
    }

    void run() {     // x is the current model
      // r is the current residue
      // g is the gradient in model space
      // G is the gradient in residue space G = Ag
      Vector rdtemp, rmtemp;    // temperary residue vector
      rdtemp.clone(&rdata); rmtemp.clone(&rreg);
      Vector GGreg,GSreg, SSreg;
      GGreg.clone(&rreg); GSreg.clone(&rreg); SSreg.clone(&rreg);
      Prod(&GGreg,&GSreg,&SSreg);

      double alpha=0.,beta=0.,small=1e-20;  // building array size 10 here? 

      if(iter==0){ 
        // steepest descent
        double alpha_temp = 1.;
        for(int i = 0; i < maxps; i++) {
          FirstDeriv(&hp,&rmtemp);
          SecondDeriv(&hpp,&rmtemp);
          double GR = Gdata.innerProd(rdtemp) + Greg.innerProd(hp);
          double GG = Gdata.innerProd(Gdata) + GGreg.innerProd(hpp);
          double alpha_temp = -GR/GG;
          rdtemp.linearCombo(1.,Gdata,alpha_temp);
          rmtemp.linearCombo(1.,Greg,alpha_temp);
          alpha = alpha + alpha_temp;
          if(std::abs(alpha_temp) < tol_ps) break;
        }
      }
      else{
        // Complicated computation
        double alpha_temp = 1.; double beta_temp = 1.;
        for(int i = 0; i < maxps; i++) {
          FirstDeriv(&hp,&rmtemp);
          SecondDeriv(&hpp,&rmtemp);
          double GR = Gdata.innerProd(rdtemp) + Greg.innerProd(hp);
          double SR = Sdata.innerProd(rdtemp) + Sreg.innerProd(hp);
          double GG = Gdata.innerProd(Gdata) + GGreg.innerProd(hpp);
          double GS = Gdata.innerProd(Sdata) + GSreg.innerProd(hpp);
          double SS = Sdata.innerProd(Sdata) + SSreg.innerProd(hpp);
          double determ = GG*SS*std::max( 1.-(GS/GG)*(GS/SS) , 1e-12 );
          //double determ = GG*SS - GS*GS;
          alpha_temp = -(SS*GR - GS*SR)/determ;
          beta_temp  = -(-GS*GR + GG*SR)/determ;
          rdtemp.linearCombo(1.,Gdata,alpha_temp);
          rdtemp.linearCombo(1.,Sdata,beta_temp);
          rmtemp.linearCombo(1.,Greg,alpha_temp);
          rmtemp.linearCombo(1.,Sreg,beta_temp);
          alpha = alpha + alpha_temp;
          beta = beta + beta_temp;
          if(std::abs(alpha_temp) + std::abs(beta_temp)< tol_ps) break;
        }
      }
  
      if(verbose){
        std::cout << "iter: " << iter << " , alpha = " << alpha << ", beta = " << beta << "\n";
      }

      // updating s S, m r

      s.linearCombo(beta,g,alpha);   //Update solution step alpha*g+beta*s
      Sdata.linearCombo(beta,Gdata,alpha); //Update residual step
      Sreg.linearCombo(beta,Greg,alpha);

      x.linearCombo(1.0,s,1.0);      //Update solution
      rdata.linearCombo(1.0,Sdata,1.0);    //Update residual
      rreg.linearCombo(1.0,Sreg,1.0);
      iter = iter + 1;
      //if(iter == 30) iter = 0;  // restart every 10 steps
      //if(std::abs(alpha) < std::abs(beta)*1.e-5) {iter = 0;std::cout<<"--------Restarting!------"<<std::endl;}
    }

    void Prod(Vector *GG, Vector *GS, Vector *SS) {
      float *__restrict__ GG_val = ((dataInCoreFloat*)GG->getDataContainer())->getFloatPtr();
      float *__restrict__ GS_val = ((dataInCoreFloat*)GS->getDataContainer())->getFloatPtr();
      float *__restrict__ SS_val = ((dataInCoreFloat*)SS->getDataContainer())->getFloatPtr();
      float *__restrict__ G_val = ((dataInCoreFloat*)Greg.getDataContainer())->getFloatPtr();
      float *__restrict__ S_val = ((dataInCoreFloat*)Sreg.getDataContainer())->getFloatPtr();

      for(long long i = 0; i < nr; i++) {
        GG_val[i] = G_val[i]*G_val[i];
        GS_val[i] = G_val[i]*S_val[i];
        SS_val[i] = S_val[i]*S_val[i];
      }
    }    // End of ProdGS function


    void FirstDeriv(Vector *hp, Vector *r) {
      float *__restrict__ hp_val = ((dataInCoreFloat*)hp->getDataContainer())->getFloatPtr();
      float *__restrict__ r_val = ((dataInCoreFloat*)r->getDataContainer())->getFloatPtr();

      for(long long i = 0; i < nr; i++) {
        float temp = r_val[i]*r_val[i] + 1.;
        hp_val[i] = eps*r_val[i]/std::sqrt(temp);
        //hp_val[i] = r_val[i];
      }

    }  // End of First Deriv function

    void SecondDeriv(Vector *hpp, Vector *r) {
      float *__restrict__ hpp_val = ((dataInCoreFloat*)hpp->getDataContainer())->getFloatPtr();
      float *__restrict__ r_val = ((dataInCoreFloat*)r->getDataContainer())->getFloatPtr();

      for(long long i = 0; i < nr; i++) {
        float temp = r_val[i]*r_val[i] + 1.;
        hpp_val[i] = eps/(temp*std::sqrt(temp));
        //hpp_val[i] = 1.;
      }

    }  // End of First Deriv function

    void SetPar(int _maxps,double _tol_ps) {
      maxps = _maxps;
      tol_ps = _tol_ps;
    }
    ~CGstep_ps() {}

  private:
    int iter, maxps;
    Vector &x, &rdata,&rreg;
    Vector &g, &Gdata,&Greg;
    Vector s, Sdata,Sreg; // updates to model and gradient
    Vector hp,hpp;
    long long nd, nr;
    float eps;
    float tol_ps;
    bool verbose;

};


//////////////////////////////////////////////
//////////////////////////////////////////////

class HB_Solver: public Algorithm {

  public:

    HB_Solver(Operator &_L, Operator &_A,
        Vector & _m, Vector &_d,
        inCoreFloat * _bigspace,
        int    & _niter,
        float * _data_res = NULL)
      :L(_L),A(_A), m(_m), d(_d), niter(_niter) {

      data_res = _data_res;
      niter = 0;
      mspace   = (inCoreFloat *)m.getSpace();
      dspace   = (inCoreFloat *)L.getRange();
      regspace = (inCoreFloat *)A.getRange();
      bigspace   =  _bigspace;
      nreg  = regspace->get_n123();
      ndata = dspace->get_n123();

      rdata = new Vector(dspace);     // residue in data space
      rreg  = new Vector(regspace);       // residue in regularization space
      hpreg = new Vector(regspace);
      g     = new Vector(mspace);       // gradient in model space
      Gdata = new Vector(dspace);       // graident in data space
      Greg  = new Vector(regspace);     //
      // default iteration parameters
      eps = 1.;
      rnorm = 1.; nrnorm = 1.; rtol = 1.e-6; nrtol = 1.e-6;
      maxiter=1;  maxCG = 1; 

    }  // Constructor

    void SetPar(float _eps, int _maxiter, double _rtol, double _nrtol,int _maxps,float _pstol) {
      eps     = _eps;
      maxiter = _maxiter;
      rtol    = _rtol;
      nrtol   = _nrtol;
      maxps   = _maxps;
      pstol   = _pstol;
    }

    bool stop() {
      if (rnorm < rtol) return true;
      if (nrnorm < nrtol) return true;
      return false;
    }


    void run() {
      std::cout << "Hello, this is hyperbolic CD solver!" << std::endl;
      std::cout << "MaxIter = " << maxiter << std::endl;
      // initialize Vector and Operators
      rdata->linearCombo(0.,d,-1.);
      L.Forward(true,&m,rdata);
      A.Forward(false,&m,rreg);
      CGstep_ps step_ps(m, *rdata, *rreg,*g,*Gdata, *Greg, ndata,nreg,eps,false);
      step_ps.SetPar(maxps,pstol);
      for(int i = 0; i < maxiter; i++) {
        if(stop()) break;
        std::cout << "Iter: " << i << " of HBCD Solver." << std::endl;
        FirstDeriv(hpreg,rreg);  // hpdata = h'(rdata)
        L.Adjoint(false,g,rdata); 
        A.Adjoint(true,g,hpreg);       // g = L'r_d + (WA)' h'(r_reg)
        L.Forward(false,g,Gdata);
        A.Forward(false,g,Greg);       // G = [Lg; WAg]
        step_ps.run();
        rnorm = std::sqrt(rdata->innerProd(*rdata)); // get norm or residue
        nrnorm = std::sqrt(g->innerProd(*g));        // get norm of gradient
        if(data_res != NULL)
          data_res[niter] = rnorm;
        std::cout << "CG step: " << niter << ". Residue = " << rnorm <<". Gradient =" << nrnorm << std::endl; 
        niter++;
      }
    }

    void FirstDeriv(Vector *hp, Vector *r) {
      float *__restrict__ hp_val = ((dataInCoreFloat*)hp->getDataContainer())->getFloatPtr();
      float *__restrict__ r_val = ((dataInCoreFloat*)r->getDataContainer())->getFloatPtr();

      for(long long i = 0; i < nreg; i++) {
        float temp = r_val[i]*r_val[i] + 1.;
        hp_val[i] = eps*r_val[i]/std::sqrt(temp);
        //hp_val[i] = r_val[i];
      }

    }  // End of First Deriv function

    ~HB_Solver() {}

  private:
    Operator       & L;
    Operator       & A;    // Operator
    Vector         & m;    // solution vector
    Vector         & d;
    int            &niter;

    float   * data_res;

    inCoreFloat    *mspace, *dspace,*regspace,*bigspace;
    long long      ndata, nreg;      // length of rd, rm

    Vector         *b, *rdata,*rreg;  
    Vector         *r, *g, *Gdata, *Greg;     // g is delta_m, G is delta_r
    Vector         *hpreg;

    int            maxiter, maxCG, maxps;
    float          eps;
    double         rnorm, nrnorm;
    double         rtol, nrtol, pstol;


};

}

#endif