#include "IRLSAlg.h"
#include <cmath>
#include <iostream>
#include <algorithm>


namespace SEP {

void IRLSReg::init(Space *mod, Space *data, float _eps) {



}

bool IRLSReg::Forward(bool add, Vector *model, Vector *data) {

  if(!add) data->zeroElement();
  float *__restrict__ m = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ d = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();
  float *__restrict__ w = ((dataInCoreFloat*)W->getDataContainer())->getFloatPtr();
  for(long long i = 0; i < n; i++) {
    d[i] += w[i]*m[i];

  }

  return true;
}

bool IRLSReg::Adjoint(bool add, Vector *model, Vector *data) {

  if(!add) model->zeroElement();
  float *__restrict__ m = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ d = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();
  float *__restrict__ w = ((dataInCoreFloat*)W->getDataContainer())->getFloatPtr();
  for(long long i = 0; i < n; i++) {
    m[i] += w[i]*d[i];
  }

  return true;
}

bool IRLSReg::SetW(Vector *residue,bool verbose) {
 
  float *__restrict__ w = ((dataInCoreFloat*)W->getDataContainer())->getFloatPtr();
  float *__restrict__ r = ((dataInCoreFloat*)residue->getDataContainer())->getFloatPtr();
  long long temp = 0;
  for(long long i = 0; i < n; i++) {
    if(std::abs(r[i]) > eps) {
      w[i] = 1./std::sqrt(std::abs(r[i]));
      temp++;
    }
    else 
      w[i] = upper;
  }
  //if (verbose)
    std::cout << "# in L1 zone is:" << temp << " , total is: "<<n << std::endl;

  return true;
}

/*
Below is the code for IRLS

*/

void IRLS_Solver::init() {
      niter = 0;
      mspace   = (inCoreFloat *)m.getSpace(); mlength = mspace->get_n123();
      dspace   = (inCoreFloat *)L.getRange();
      regspace = (inCoreFloat *)A.getRange();
      nreg  = regspace->get_n123();
      ndata = dspace->get_n123();

      rreg  = new Vector(regspace);
      rdata = new Vector(dspace);
      g     = new Vector(mspace);
      Greg  = new Vector(regspace);
      Gdata = new Vector(dspace);
      
      // default iteration parameters
      rnorm = 1.; nrnorm = 1.; rtol = 1.e-6; nrtol = 1.e-6;
      maxiter=1;  maxCG = 1; 
}


void IRLS_Solver::SetPar(int _maxiter, int _maxCG, double _rtol, double _nrtol) {
  maxiter = _maxiter;
  maxCG   = _maxCG;
  rtol    = _rtol;
  nrtol   = _nrtol;
}

bool IRLS_Solver::stop() {
  if (rnorm < rtol) return true;
  if (nrnorm < nrtol) return true;
  return false;
}


void IRLS_Solver::run() {
  std::cout << "Hello, this is IRLS solver!" << std::endl;
  std::cout << "MaxIter = " << maxiter <<", maxCG = " << maxCG << std::endl;
  // initialize Vector and Operators
  ChainOP WA(A,Wr);  // WA = Wr*A
   int cgiter = 0;
   
  for(int i = 0; i < maxiter; i++) {
    //if(stop()) break;
    std::cout << "Iter: " << i << " of IRLS Solver." << std::endl;
    rreg->linearCombo(0.,dreg,-1.);
    A.Forward(true,&m,rreg);     // rreg = A*m
    if(i !=0)((IRLSReg *)&Wr)->SetW(rreg);  // modify Wr<=f(rreg)
    CGRegAlg cg(L,WA,m,data,dreg,rnorm,nrnorm,cgiter,rtol,nrtol,maxCG*(i+1),true);
    cg.run();
  }
}




} // End of namespace
