// IRLSAlg.H
// Created by YM
/*
Using CGstep function, only handle weighting in the regularization space
How to use CCStepper?
    CGStepper(Vector &mod, Vector &data,Vector &_g, Vector &_G, bool _verbose = false);
    CGStepper.run();

  This is actually IRLS regularized solver

*/

#ifndef IRLSALG_H
#define IRLSALG_H 1

#include "Space.h"
#include "DataInCoreFloat.h"
#include "inCoreFloat.h"
#include "Vector.h"
#include "Operator.h"
#include "MultiOperator.h"
#include "Algorithm.h"
#include "CGAlg.h"
#include <iostream>
#include <cmath>

namespace SEP {


class IRLSReg: public Operator {
  public:
    IRLSReg(Space *mod, Space *data, float _eps = 1.0) {
  setDomain((inCoreFloat *)mod);
  setRange((inCoreFloat *)data);
  eps  = _eps;
  upper = 1./std::sqrt(eps);
  std::cout <<"hell0?" <<std::endl;
  n = ((inCoreFloat*)mod)->get_n123();
  std::cout <<"hell0?" << n <<std::endl;
  wspace = new inCoreFloat1D(n);
  std::cout <<"hell0?" <<std::endl;
  W = new Vector(wspace);
  float *__restrict__ w = ((dataInCoreFloat*)W->getDataContainer())->getFloatPtr();
  for(long long i = 0; i< n; i++) w[i] = 1.;

      init(mod,data,_eps);
    }
    ~IRLSReg() {delete W; delete wspace;}

    bool Forward(bool add, Vector *model, Vector *data);
    bool Adjoint(bool add, Vector *model, Vector *data);
    bool SetW(Vector *residue, bool verbose = true);

  private:
    inCoreFloat * wspace;
    Vector * W;
    float eps, upper;
    long long n;
    void init(Space *mod, Space *data, float _eps);
};



class IRLS_Solver: public Algorithm {

  public:

    IRLS_Solver(Operator &_L, Operator &_A, Operator &_Wr,
        Vector & _m, Vector &_data, Vector &_dreg,
        int    & _niter,
        float * _data_res = NULL)
      :L(_L),A(_A), Wr(_Wr), m(_m), data(_data), dreg(_dreg),niter(_niter) {

      init();


    }  // Constructor

    void SetPar(int _maxiter, int _maxCG, double _rtol, double _nrtol);
    bool stop();
    void run();
    ~IRLS_Solver() {}

  private:
    Operator       & L;
    Operator       & A;    // Operator
    Operator       & Wr;
    Vector         & m;    // solution vector
    Vector         & data, &dreg;
    int            &niter;
    long long      mlength;


    inCoreFloat    *mspace, *dspace,*regspace;
    long long      ndata, nreg;      // length of rd, rm

    Vector         *rdata,*rreg, *r, *g, *Greg, *Gdata;// g is delta_m, G is delta_r

    int            maxiter, maxCG;
    double         rnorm, nrnorm;
    double         rtol, nrtol;
    void init();

};




}

#endif
