#ifndef IRLS_REG_H
#define IRLS_REG_H 1
#include "Operator.h"
#include "DataInCoreFloat.h"
#include "inCoreFloat.h"
typedef long long llong;

namespace SEP {

class IRLSL1: public Operator {
  public:
    IRLSL1(Space *mod, Space *data, float eps,Vector *_mask=NULL) {
      setDomain(mod);
      setRange(data);
      this->eps  = eps;
      upper = 1./std::sqrt(eps);
      n = ((inCoreFloat*)mod)->get_n123();
      wspace = new inCoreFloat1D(n);
      W = new Vector(wspace);
      float *__restrict__ w = ((dataInCoreFloat*)W->getDataContainer())->getFloatPtr();
      for(long long i = 0; i< n; i++) {
        w[i] = 1.;
      }
      mask = _mask;
    }

    ~IRLSL1() {delete W; delete wspace;}

    bool Forward(bool add, Vector *model, Vector *data) {
      if(!add) data->zeroElement();
      float *__restrict__ m = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
      float *__restrict__ d = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();
      float *__restrict__ w = ((dataInCoreFloat*)W->getDataContainer())->getFloatPtr();
      for(long long i = 0; i < n; i++) {
        d[i] += w[i]*m[i];
        // d[i] += m[i];
      }
      return true;
    }

    bool Adjoint(bool add, Vector *model, Vector *data) {
      if(!add) model->zeroElement();
      float *__restrict__ m = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
      float *__restrict__ d = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();
      float *__restrict__ w = ((dataInCoreFloat*)W->getDataContainer())->getFloatPtr();
      for(long long i = 0; i < n; i++) {
        m[i] += w[i]*d[i];
        //m[i] += d[i];
      }
      return true;
    }

    bool SetW(Vector *residue) {
      float *__restrict__ w = ((dataInCoreFloat*)W->getDataContainer())->getFloatPtr();
      float *__restrict__ r = ((dataInCoreFloat*)residue->getDataContainer())->getFloatPtr();
      float *__restrict__ mas = ((dataInCoreFloat*)mask->getDataContainer())->getFloatPtr();
      long long temp = 0;
      for(long long i = 0; i < n; i++) {
        if(std::abs(r[i]) > eps) {
          w[i] = 1./std::sqrt(std::abs(r[i]));
          temp++;
        }
        else 
          w[i] = upper;
      }

     std::cout << "# in L1 zone is:" << temp << " , total is: "<<n << std::endl;
    }

  private:
    inCoreFloat * wspace;
    Vector * W;
    float eps, upper;
    long long n;
    Vector *mask;
};



class IRLSL2: public Operator {
  public:
    IRLSL2(Space *mod, Space *data) {
      setDomain(mod);
      setRange(data);
    }

    ~IRLSL2() {}

    bool Forward(bool add, Vector *model, Vector *data) {
      if(!add) data->zeroElement();
      data->linearCombo(1.,*model,1.);
      return true;
    }

    bool Adjoint(bool add, Vector *model, Vector *data) {
      if(!add) model->zeroElement();
      model->linearCombo(1.,*data,1.);
      return true;
    }


  private:

};




}

#endif