#ifndef Space_H
#define Space_H 1
#include "dataContainer.h"
#include "handleIO.h"
#include "except.h"
#include <boost/smart_ptr/shared_ptr.hpp>
#include <boost/smart_ptr/make_shared.hpp>
#include "axis.h"
#include <vector>


namespace SEP {


class Space: public Writeable {
	private:
		boost::shared_ptr<int> key;
	public:
		Space() { 
			key = boost::make_shared<int>(1);			
		} // default constructor
		virtual ~Space() {} // destructor

        virtual dataContainer * buildDataContainer()const{
            return 0; // must return something
		}
		// build a data container for a vector in the spaces

		boost::shared_ptr<int> get_key() const{ return this->key;}

		virtual void set_key(const boost::shared_ptr<int> & key){
			this->key = key;
		} 

		virtual long long get_n123() = 0;
		// specific instances of space have dimensions 
		
		virtual bool isCompatible(dataContainer * otherContainer) const = 0;
		// check compatibility of a vector with this space

		virtual bool isCompatible(Space * otherSpace, bool makeSame=false) const = 0;
		// Check if spaces are compatible. By default, do not give them the
		// same key if they are compatible, but if makeSame = true, compatible
		// spaces are given the same key.

		virtual void zeroElement(dataContainer * x) const = 0;

		virtual void random(dataContainer * x) const = 0;
		// zero element of the space

		virtual bool operator ==(const Space & otherSpace) const
		{
				if (this == &otherSpace){
					return true;
				} else if(this->key == otherSpace.get_key()){
					return true;
				} 
                                return false;
		}
		// check if two vector MultiSpace are equal 

		virtual void linearCombo(float a, dataContainer  * x, float b, dataContainer * y) const = 0;
		// do a linear combination x = ax + by

		virtual double innerProd(dataContainer  * x, dataContainer  * y) const = 0;
		// take the inner product of x and y

 //Put back later *************************
		//virtual void norm(int p, dataContainer  * x) const = 0;
		// calculate the p-norm of a vector (p=2 Euclidean)

		virtual void addInv(dataContainer * x) const = 0;
		// get the additive inverse of x (multiply by -1) 
		// and overwrite x

		virtual void addInv(dataContainer  * x, dataContainer * xNeg) const = 0;
		// xNeg = -1*x but leave x alone 

		virtual void scale(float a, dataContainer * x) const=0;
		// multiply vector x by scalar a and overwrite x

		virtual void scale(float a, dataContainer  * x, dataContainer * ax) const  = 0;

		// multiply vector x by scalar a and write it in ax


		
};
  
class MultiSpace: public Space {
	public:
	    Space *sp1,*sp2;
  
        MultiSpace():sp1(0),sp2(0) {} // default constructor
		MultiSpace(Space *s1, Space *s2);
		virtual ~MultiSpace() {} // destructor
        	Space *getSpace1()const{return sp1;}
        	Space *getSpace2()const{return sp2;}

		virtual MultiDataContainer * buildDataContainer() const;
		
	//	virtual MultiSpace clone_space() const;

		virtual long long get_n123();
		
		virtual void zeroElement(dataContainer * x) const;

		virtual void linearCombo(float a, dataContainer  * x, float b, dataContainer * y) const ;

		virtual double innerProd(dataContainer  * x, dataContainer  * y) const ;

		//hide virtual float norm(int p, dataContainer  * x) const ;

		virtual void addInv(dataContainer * x) const ;

		virtual void addInv(dataContainer  * x, dataContainer * xNeg) const;

		virtual void scale(float a, dataContainer * x) const;

		virtual void scale(float a, dataContainer  * x, dataContainer * ax) const ;
		virtual bool isCompatible(Space * otherSpace, bool makeSame=false) const;
		virtual bool isCompatible(dataContainer * otherContainer) const;
       	virtual void random(dataContainer * x) const;

		// zero element of the space

};



}
#endif
