#ifndef SPLITBREGMAN_H
#define SPLITBREGMAN_H 1
#include "Algorithm.h"
#include "CGAlg.h"
#include "Operator.h"
#include "MultiOperator.h"
#include "DataInCoreFloat.h"
#include "inCoreFloat.h"


#include <iostream>
namespace SEP {

class Shrink: public Algorithm {

  public:
    Shrink(Vector & _z, double _threshold):z(_z) {
    threshold = _threshold;
    inCoreFloat *temp = (inCoreFloat *)z.getSpace();
    d = ((dataInCoreFloat*)z.getDataContainer())->getFloatPtr();
    n123 = temp->get_n123();
    }
    void run() {
      double temp;
      float pos = 1., neg = 1.;
      double zero = 0.;
      long long zeros = 0;
      for(long long i = 0; i < n123; i++) {
        temp = d[i];
        if (temp >0) {
          d[i] = pos*std::max(zero,temp - threshold);
        }
        else {
          d[i] = neg*std::max(zero,-temp - threshold);
        }
        if(d[i] < 1.e-9 and d[i] > -1.e-9) zeros++;
      }
      std::cout <<"zeros(z) = " << 100.*((double)zeros)/(double)n123 <<std::endl;
    }
    long long zeros;
  private:
    Vector &z;
    double threshold;
    long long n123;
    float *__restrict__ d;



};


class SplitBregman: public Algorithm{

  public:

    SplitBregman(Operator & _L, Operator & _A,
        Vector & _m,
        Vector & _d,
        Vector & _sol,
        inCoreFloat  * _bspace,
        double _lambda,
        double _rho,
        int    &_niter,
        float  * _data_res=NULL)
      :L(_L),A(_A),m(_m),d(_d),sol(_sol),niter(_niter)
      { 
        // initilizer, could be complicated
        data_res = _data_res;
        lambda = _lambda; rho = _rho;
        niter = 0;
        mspace   = (inCoreFloat *)m.getSpace();
        dspace   = (inCoreFloat *)L.getRange();
        zspace   = (inCoreFloat *)A.getRange();
        bspace   =  _bspace;

        z = new Vector(zspace);
        u = new Vector(zspace);
        b = new Vector(bspace);
        first  = new Vector(dspace);
        second = new Vector(zspace);

        n1 = dspace->get_n123();
        n2 = zspace->get_n123();
        ml = mspace->get_n123();
        std::cout <<"n1="<<n1<<", n2="<<n2<<std::endl;
        rhos = std::sqrt(rho);
        maxiter = 10; rtol = 1.e-6; nrtol = 1.e-6; maxCG = 5; maxInner = 2;//Default
        rnorm = 1.; nrnorm = 1.;
        // Creating 
        b->zeroElement();
        float *__restrict__ bd = ((dataInCoreFloat*)b->getDataContainer())->getFloatPtr();
        float *__restrict__ dd = ((dataInCoreFloat*)d.getDataContainer())->getFloatPtr();
        for(long long i = 0; i < n1; i++) {
          bd[i] = dd[i];
        }

      }
    void SetPar(int _maxiter,double _rtol, double _nrtol) {
      maxiter = _maxiter;
      rtol    = _rtol;
      nrtol   = _nrtol;
    }
    void SetSPPar(int _maxInner, int _maxCG) {
      maxInner = _maxInner; 
      maxCG    = _maxCG;
    }

    bool stop() {
      if (rnorm < rtol) return true;
      if (nrnorm < nrtol) return true;
      return false;
    } 

    void run() {
      // Setup the operators
      ScaleOP scale(zspace,zspace,rhos);
      RowChain   Reg(A,scale);
      ColOP   M(L,Reg,bspace);
      // Setup the shrinkage function
      Shrink shrinking(*z,lambda/rho);
      for(int i = 0; i < maxiter; i++) {
        if(!stop()) {
          std::cout << "Iter: " << i << std::endl; 
          for(int j = 0; j < maxInner; j++) {
            Combine();          // b = [d;sqrt(rho)*(z-u)]
            int maxCGIter = niter + maxCG;
            CGAlg cg(M,m,*b,rnorm,nrnorm,niter,rtol,nrtol,maxCGIter,true,data_res,first,second);
            cg.run();
            A.Forward(false,&m,z);   // 
            z->linearCombo(1.,u,1.);   // z = Am + u;
            shrinking.run();          // shrink(z,lambda/rho);
          }
          A.Forward(false,&m,u);   // put true back later
          u->linearCombo(1.,*z,-1.);
          Copying(i,&m,&sol);

        }
        else break;
      } 
    }

    void Combine() {
      float *__restrict__ bd = ((dataInCoreFloat*)b->getDataContainer())->getFloatPtr();
      float *__restrict__ zd = ((dataInCoreFloat*)z->getDataContainer())->getFloatPtr();
      float *__restrict__ ud = ((dataInCoreFloat*)u->getDataContainer())->getFloatPtr();
      for(long long i = 0; i < n2; i++) 
        bd[i+n1] = (zd[i] - ud[i])*rhos;

    }

    void Copying(long long index,Vector *small, Vector *big) {
      float *__restrict__ sdata = ((dataInCoreFloat*)small->getDataContainer())->getFloatPtr();
      float *__restrict__ bdata = ((dataInCoreFloat*)big->getDataContainer())->getFloatPtr();
      long long shift = index*ml;
      for(long long i = 0; i < ml; i++) 
        bdata[shift + i] = sdata[i];

    }
    ~SplitBregman(){delete u; delete z; delete b; delete first; delete second;}


  private:
    Operator       & L;
    Operator       & A;    // Operator
    Vector         & m;    // solution vector
    Vector         & d;
    Vector         & sol;
    int            &niter;
    Vector         *u;
    Vector         *z;
    Vector         *b;
    Vector  *first, *second;

    inCoreFloat    *mspace, *dspace,*zspace,*bspace;


    double         lambda;
    double          rho;
    double          rhos;
    double         rtol;   // tolerance of residual norm
    double         nrtol;  // tolerance of gradient norm
    double         rnorm, nrnorm;
    int            maxiter, maxCG, maxInner;
    bool           verbose;

    long long  n1, n2,ml;
    float   * data_res;


};

}

#endif