#ifndef TWOOPER_H
#define TWOOPER_H 1
#include "Space.h"
#include "handleIO.h"
#include "inCoreFloat.h"
#include "Vector.h"
#include <iostream>

namespace SEP {

class TwoOperator: public Operator {

  public:
    TwoOperator(float a1,Operator *L1,float a2, Operator *L2,Space *dom,Space *ran) {
      this->L1 = L1; this->a1 = a1;
      this->L2 = L2; this->a2 = a2;
      setDomain(dom); setRange(ran);
      range1 = (inCoreFloat *)L1->getRange();
      range2 = (inCoreFloat *)L2->getRange();
      data1 = new Vector(range1);
      data2 = new Vector(range2);
      size1 = range1->get_n123();
      size2 = range2->get_n123();
      size = size1 + size2;
      std::cout <<"size1="<<size1 << ", size2="<<size2<<std::endl;
    }
    //void SetRange(Space *rang) {range = rang;}

    virtual bool Forward(bool add, Vector *model, Vector *data) {
      if(!add) data->zeroElement();
      L1->Forward(false,model,data1);
      L2->Forward(false,model,data2);
      Combine(data);
      return true;
    }
    virtual bool Adjoint(bool add, Vector *model, Vector *data) {
      Split(data);
      L1->Adjoint(add,model,data1);
      L2->Adjoint(true,model,data2);
      return true;
    }

    void Combine(Vector *data) {
      float *__restrict__ d1 = ((dataInCoreFloat*)data1->getDataContainer())->getFloatPtr();
      float *__restrict__ d2 = ((dataInCoreFloat*)data2->getDataContainer())->getFloatPtr();
      float *__restrict__ d = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();
      for(long long i = 0; i < size1; i++) {
        d[i] += a1*d1[i];
      }
      for(long long i = 0; i < size2; i++) {
        d[i+size1] += a2*d2[i];
      }
    }
    void Split(Vector *data) {
      float *__restrict__ d1 = ((dataInCoreFloat*)data1->getDataContainer())->getFloatPtr();
      float *__restrict__ d2 = ((dataInCoreFloat*)data2->getDataContainer())->getFloatPtr();
      float *__restrict__ d = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();
      for(long long i = 0; i < size1; i++) {
        d1[i] = a1*d[i];
      }
      for(long long i = 0; i < size2; i++) {
        d2[i] = a2*d[i+size1];
      }
    }

   ~TwoOperator() {delete data1; delete data2;}
  private:
    Operator *L1;
    Operator *L2;
    float a1;
    float a2;
    inCoreFloat * range1;
    inCoreFloat * range2;
    Vector *data1;
    Vector *data2;
    float *d1;
    float *d2;
    long long size1, size2,size;
};


}

#endif