#include "inCoreFloat.h"
#include "Vector.h"

SEP::Vector::Vector(Space *sp1) {
  sp=sp1;
  d=sp->buildDataContainer();
}

SEP::Vector::Vector(Vector *x) {
  sp=x->getSpace();
  d=sp->buildDataContainer();
  sp->linearCombo(0.,d,1.,x->getDataContainer());
} 

SEP::Vector::Vector(Space *sp1, const dataContainer &otherContainer){
  sp=sp1;
  if (sp->isCompatible(getDataContainer())) {
    *d = otherContainer;
  }
  else {
    SEPException e;
    e<<"The data container is not compatible with the space!";
    throw e;
  }
}

SEP::Vector SEP::Vector::clone() { 
  Vector y(this); 
  return y;
}

void SEP::Vector::clone(Vector *in){ 
  sp=in->getSpace();
  d=sp->buildDataContainer();
  zeroElement();
  linearCombo( 1.0, *in, 1.0);
}


void SEP::Vector::zeroElement() {
  /// zero out the data Container
  ///
  sp->zeroElement((this->getDataContainer())); 
}

void SEP::Vector::linearCombo(float a, const Vector &x, float b) { 
  /// Calculate this = a*this + b*x.
  /// Calling linearCombo function in Space class.
  if (inSameSpace(x)){
    sp->linearCombo(a, (this->getDataContainer()), b, (x.getDataContainer()));
  }
  else {
    SEPException e;
    e<<"The two vectors you intended to combine are not in the same space!";
    throw e;
  }
}

double SEP::Vector::innerProd(const Vector &x) {
  /// Calculates the inner product of x and y as a double. 
  /// Calling innerProd in Space class.
  if (inSameSpace(x)){
    double ip = sp->innerProd((x.getDataContainer()), (this->getDataContainer()));
    return ip;
  }
  else {
    SEPException e;
    e<<"The two vectors you are trying to take the inner product of are not in the same space!";
    throw e;
  }
}

void SEP::Vector::scale(float a) {
  /// Calculates x = ax (using openmp and vectorization)
  /// Calling scale function in Space class.
  dataContainer* x=this->getDataContainer();
  sp->scale(a, x);
} 

void SEP::Vector::scale(float a, dataContainer *ax) {
   sp->scale(a, (this->getDataContainer()), ax);
}

void SEP::Vector::random() {
  sp->random(this->getDataContainer());
}

void SEP::Vector::addInv() {
  sp->addInv((this->getDataContainer()));
}


void SEP::Vector::addInv(const Vector &xNeg) {
  if (inSameSpace(xNeg)) 
    sp->addInv((this->getDataContainer()), (xNeg.getDataContainer()));
  else {
    SEPException e;
    e<<"The output and input vectors are not in the same space!";
    throw e;
  }
}

