#ifndef VECTOR_H
#define VECTOR_H 1
#include "Space.h"
#include "handleIO.h"
#include "inCoreFloat.h"
#include <vector>
/*

*/
namespace SEP {

class Vector: public Writeable {
  private:
    Space *sp;
    dataContainer *d;
    
  public:
    Vector(){sp=NULL;d=NULL;}  // Default constructor
    Vector(Space *sp1);
    Vector(Vector *x);// construct Vector from another Vector, space is the same,
                      // data Container is COPIED.
    Vector(Space *sp1, const dataContainer &otherContainer);
    virtual ~Vector(){ delete d; d = NULL;}
    Vector clone();// Space is the same, data Container is COPIED!
    void clone(Vector *in);
//    //Return pointer to data storage (depends on data container type)
//   void *returnData() const { return d->returnData();}  
    virtual dataContainer *getDataContainer() const {return d;}
    virtual Space *getSpace() const { return sp; }
    virtual bool inSpace(Space *sp1) const { return (sp==sp1); }
    virtual bool inSameSpace(const Vector &x) const { return (inSpace(x.sp)); }
    virtual void zeroElement();
    virtual void linearCombo(float a, const Vector &x, float b);
    virtual double innerProd(const Vector &x);
    virtual void scale(float a);
    virtual void scale(float a, dataContainer *ax);
    virtual void random();
    virtual void addInv();
    virtual void addInv(const Vector &xNeg);


/*Add back later
    void scale(float a, const Vector &ax) {      
      if (inSameSpace(ax)) 
        sp->scale(a, (this->getDataContainer()), (ax.getDataContainer()));
      else {
        SEPException e;
        e<<"The output and input vectors are not in the same space!";
        throw e;
      }
    } 
    */
};




class MultiVector: public Vector{
  private:
    MultiSpace *sp;
    dataContainer *d1, *d2;
  public:
    MultiVector(){}
    MultiVector(Vector *v1, Vector *v2){
      Space *sp1 = v1->getSpace();
      Space *sp2 = v2->getSpace();
      this->sp = new MultiSpace(sp1, sp2);

      d1 = sp1->buildDataContainer(); // ****should this actually copy vectors?
      sp1->linearCombo(float(0), d1, float(1), v1->getDataContainer());
      d2 = sp2->buildDataContainer();
      sp2->linearCombo(float(0), d2, float(1), v2->getDataContainer());

    }

    MultiVector(Vector *v1, Vector *v2, MultiSpace *s){
     // ******compatibility checks etc... then create multivector****
      sp = s;
      Space *sp1 = s->getSpace1();
      Space *sp2 = s->getSpace2();
      d1 = sp1->buildDataContainer(); // ****should this actually copy vectors?
      sp1->linearCombo(float(0), d1, float(1), v1->getDataContainer());
      d2 = sp2->buildDataContainer();
      sp2->linearCombo(float(0), d2, float(1), v2->getDataContainer());
    }

    MultiVector(Space *s1, Space *s2){
      Space *sp1=s1;
      this->d1=sp1->buildDataContainer();
      Space *sp2=s2;
      this->d2=sp2->buildDataContainer();
    }

    MultiVector(MultiVector *m){
      sp = m-> getSpace();
      Space *sp1 = sp->getSpace1();
      d1 = sp1->buildDataContainer();
      sp1->linearCombo(float(0),d1,float(1),m->getDataContainer1());
      Space *sp2 = sp->getSpace2();
      d2 = sp2->buildDataContainer();
      sp2->linearCombo(float(0),d2,float(1),m->getDataContainer2());
    }

    MultiVector(Space *s1, const dataContainer &otherContainer1, Space *s2, const dataContainer &otherContainer2){
      Space *sp1 = s1;

      if(sp1->isCompatible(getDataContainer())){
        *d1 = otherContainer1;
      } else{
        SEPException e;
        e<<"The data container is not compatible with the space!";
        throw e;
      }
      Space *sp2 = s2;
      if(sp2->isCompatible(getDataContainer())){
        *d2 = otherContainer2;
      } else{
        SEPException e;
        e<<"The data container is not compatible with the space!";
        throw e;
      }

    }

    virtual ~MultiVector() {}

    MultiDataContainer *getDataContainer() const{
        MultiDataContainer *temp=new MultiDataContainer(d1, d2);
        return temp; 
    };

    dataContainer *getDataContainer1() const{ return d1;}
    dataContainer *getDataContainer2() const{ return d2;}
    MultiSpace *getSpace() const{ return sp;}
    bool inSpace (const MultiSpace *m) const{};
    bool inSameSpace(const MultiVector &x) const{ //if not a const this compiles, or add const{} to inSpace
	const MultiSpace *multsp = x.getSpace();
	return (inSpace(multsp));
        //std::cout << "WARNING: Called inSameSpace on MultiVector, needs implementation\n";
	//return true; // **********************just to compile code*****
    };
    virtual MultiVector clone(void) {

    }

    virtual void zeroElement(void) {
      sp->zeroElement((this->getDataContainer())); 
    }

    virtual void linearCombo(float a, const Vector &x, float b) {

      MultiVector * xm = (MultiVector *)(& x);
      sp->linearCombo(a, this->getDataContainer(), b,xm->getDataContainer());

    }
    virtual double innerProd(const Vector &x){ //inherit from vector
      MultiVector * xm = (MultiVector *)(& x);
      return sp->innerProd(this->getDataContainer(),xm->getDataContainer());

    }
    virtual void scale(float a) {
      sp->scale(a,this->getDataContainer());
    }
    virtual void scale(float a, dataContainer *ax) {
      sp->scale(a,this->getDataContainer(),ax);
    }

    virtual void random() {
      sp->random(this->getDataContainer());
    }
    virtual void addInv() {
      sp->addInv(this->getDataContainer());
    }
    virtual void addInv(const Vector &xNeg) {
      dataContainer * xm = ((MultiVector *)(& xNeg))->getDataContainer();
      sp->addInv(this->getDataContainer(),xm);
    }
};

};
#endif

