#include "inCoreFloat.h"
#include "vectorize.h"
#include "DataInCoreFloat.h"
#include <iostream> //****just to test
#include <cstdlib> //****just to test
typedef long long llong;

void SEP::inCoreFloat::linearCombo(float a,  dataContainer *x, float b, dataContainer *y) const{
   /// Calculate x = ax + by. Note that this can use vectorization/openmp.
   ///
   dataInCoreFloat *s=(dataInCoreFloat*)x;
   dataInCoreFloat *t=(dataInCoreFloat*)y;
   std::vector<llong> *b_omp=s->get_b_omp(),*n_omp=s->get_n_omp();
   float *__restrict__ my=s->getFloatPtr();
   float *__restrict__ your=t->getFloatPtr();
   for(llong i=0; i < b_omp->at(0); i++) {
	 my[i]=my[i]*a+your[i]*b;
   }
   int ith=0;
   #pragma omp parallel
   ith=omp_get_thread_num();
   llong j=0;
   SIMD_constant c0,c1;
   SIMD_Float16::setConstant(c0,a);
   SIMD_Float16::setConstant(c1,b);
   SIMD_Float16 mine,yours;
   for(llong i=b_omp->at(ith); j< n_omp->at(ith); j++, i+=BLOCK_SIZE){
     mine.load(&my[i]);
     yours.loadu(&your[i]);  //Can't guarantee the same allignment
     mine*=c0;
     yours*=c1;
     mine+=yours;
     mine.stream(&my[i]);
   }
	
   //Finish it up 
   int nth=s->get_nth()-1;
   for(llong i=b_omp->at(nth)+(n_omp->at(nth))*BLOCK_SIZE; i< n123; i++) my[i]=my[i]*a+b*your[i];
}


void  SEP::inCoreFloat::addInv(dataContainer *x) const{
   /// Sets x = -x, the additive inverse (uses openmp and vectorization)
   ///
   dataInCoreFloat *s=(dataInCoreFloat*)x;
   std::vector<llong> *b_omp=s->get_b_omp(),*n_omp=s->get_n_omp();
   float *__restrict__ my=s->getFloatPtr();
      int ith=0;
   #pragma omp parallel
   ith=omp_get_thread_num();
   for(llong i=0; i < b_omp->at(0); i++)  my[i]=my[i]*-1;;
   llong j=0;
   SIMD_constant c0;
   SIMD_Float16::setConstant(c0,-1.);
   SIMD_Float16 mine;
   for(llong i=b_omp->at(ith); j < n_omp->at(ith); j++, i+=BLOCK_SIZE){
     mine.load(&my[i]);
     mine*=c0;
     mine.stream(&my[i]);
   }
   //Finish it up 
      int nth=s->get_nth()-1;

   for(llong i=b_omp->at(nth)+(n_omp->at(nth))*BLOCK_SIZE; i< n123; i++) my[i]=-my[i];

}
void  SEP::inCoreFloat::addInv( dataContainer *x, dataContainer *xneg)const {
   /// Calculates xneg = -x (using openmp and vectorization)
   ///
   dataInCoreFloat *s=(dataInCoreFloat*)x;
   dataInCoreFloat *t=(dataInCoreFloat*)xneg;
   std::vector<llong> *b_omp=s->get_b_omp(),*n_omp=s->get_n_omp();
   float *__restrict__ my=t->getFloatPtr();
   float *__restrict__ you=s->getFloatPtr();
      int ith=0;
   #pragma omp parallel
   ith=omp_get_thread_num();
   for(llong i=0; i < b_omp->at(0) ; i++)  my[i]=you[i]*-1;;

   llong j=0;
   SIMD_constant c0;
   SIMD_Float16::setConstant(c0,-1.);
   SIMD_Float16 mine;
   for(llong i=b_omp->at(ith); j < n_omp->at(ith); j++, i+=BLOCK_SIZE){
     mine.load(&you[i]);
     mine*=c0;
     mine.stream(&my[i]);
   }
   //Finish it up 
      int nth=s->get_nth()-1;

   for(llong i=b_omp->at(nth)+(n_omp->at(nth))*BLOCK_SIZE; i< n123; i++) my[i]=-you[i];
}

void  SEP::inCoreFloat::scale(float a, dataContainer *x,dataContainer *y) const {
   /// Calculates y = ax (uses openmp and vectorization)
   ///
   dataInCoreFloat *s=(dataInCoreFloat*)x;
      dataInCoreFloat *t=(dataInCoreFloat*)y;
   std::vector<llong> *b_omp=s->get_b_omp(),*n_omp=s->get_n_omp();
   float *__restrict__ my=t->getFloatPtr();
   float *__restrict__ their=s->getFloatPtr();
   for(llong i=0; i < b_omp->at(0); i++)  my[i]=their[i]*a; 
   int ith;
   #pragma omp parallel
   ith=omp_get_thread_num();
   llong j=0;
   SIMD_constant c0;
   SIMD_Float16::setConstant(c0,a);
   SIMD_Float16 mine;
   for(llong i=b_omp->at(ith); j < n_omp->at(ith); j++, i+=BLOCK_SIZE){
     mine.loadu(&their[i]);
     mine*=c0;
     mine.stream(&my[i]);
   }
   //Finish it up 
   int nth=s->get_nth()-1;

   for(llong i=b_omp->at(nth)+(n_omp->at(nth))*BLOCK_SIZE; i< n123; i++) my[i]=their[i]*a;
	// added *a in line above here, made no difference
}

void  SEP::inCoreFloat::scale(float a, dataContainer *x) const {
   /// Calculates x = ax (using openmp and vectorization)
   ///
   dataInCoreFloat *s=(dataInCoreFloat*)x;
   std::vector<llong> *b_omp=s->get_b_omp(),*n_omp=s->get_n_omp();
   float *__restrict__ my=s->getFloatPtr();
   for(llong i=0; i < b_omp->at(0); i++)  my[i]=my[i]*a;
   int ith;

   #pragma omp parallel
   ith=omp_get_thread_num();
   llong j=0;
   SIMD_constant c0;
   SIMD_Float16::setConstant(c0,a);
   SIMD_Float16 mine;
   for(llong i=b_omp->at(ith); j < n_omp->at(ith); j++, i+=BLOCK_SIZE){
     mine.load(&my[i]);
     mine*=c0;
     mine.stream(&my[i]);
   }
   //Finish it up 
   int nth=s->get_nth()-1;

   for(llong i=b_omp->at(nth)+(n_omp->at(nth))*BLOCK_SIZE; i< n123; i++) my[i]*=a;
}
void  SEP::inCoreFloat::zeroElement(dataContainer *x) const{
   /// Fills in x with all zero data of the right dimensions
   /// (uses openmp and vectorization)
   dataInCoreFloat *s=(dataInCoreFloat*)x;
   std::vector<llong> *b_omp=s->get_b_omp(),*n_omp=s->get_n_omp();
   float *__restrict__ my=s->getFloatPtr();
   int ith;
   #pragma omp parallel
   ith=omp_get_thread_num();
   for(llong i=0; i < b_omp->at(0); i++)  my[i]=0;
   llong j=0;
   SIMD_constant c0;
   SIMD_Float16 mine;
   mine.load(&my[0]);
   SIMD_Float16::setConstant(c0,0);
   mine*=c0;
   for(llong i=b_omp->at(ith); j <n_omp->at(ith); j++, i+=BLOCK_SIZE){
     mine.stream(&my[i]);
   }
   //Finish it up 
     
    
      int nth=s->get_nth()-1;
      
   for(llong i=b_omp->at(nth)+(n_omp->at(nth))*BLOCK_SIZE; i< n123; i++) my[i]=0.;
}
void  SEP::inCoreFloat::random(dataContainer *x) const{
   /// Fills x with random data selected from a uniform distribution
   /// between -0.5 and 0.5 (uses openmp and vectorization)
   dataInCoreFloat *s=(dataInCoreFloat*)x;
   std::vector<llong> *b_omp=s->get_b_omp(),*n_omp=s->get_n_omp();
   float *__restrict__ my=s->getFloatPtr();
   int ith;
   #pragma omp parallel
   ith=omp_get_thread_num();
   for(llong i=0; i < b_omp->at(0); i++) my[i]=(float)rand()/(float)RAND_MAX-.5;
   llong j=0;
   SIMD_constant c0;
   SIMD_Float16 mine;
   SIMD_Float16::setConstant(c0,0);
   mine*=c0;
   for(llong i=b_omp->at(ith); j <n_omp->at(ith); j++, i+=BLOCK_SIZE){
     for(llong k=0; k < BLOCK_SIZE; k++) my[i+k]=(float)rand()/(float)RAND_MAX-.5;
   }
   //Finish it up 
      int nth=s->get_nth()-1;

   for(llong i=b_omp->at(nth)+(n_omp->at(nth))*BLOCK_SIZE; i< n123; i++) my[i]=(float)rand()/(float)RAND_MAX-.5;
}


double SEP::inCoreFloat::innerProd(dataContainer *x, dataContainer *y) const{
   /// Calculates the inner product of x and y as a double (uses openmp, 
   /// vectorization, and smaller sub-sums to increase accuracy).
	dataInCoreFloat *s=(dataInCoreFloat*)x;
	dataInCoreFloat *t=(dataInCoreFloat*)y;
	std::vector<llong> *b_omp=s->get_b_omp(), *n_omp=s->get_n_omp();
	float *__restrict__ my=s->getFloatPtr();
	float *__restrict__ your=t->getFloatPtr();
	double ip = 0.0; // total sum
	double dip = 0.0; // keep track of double sub sum 
	int floatGroup = 0; // keep track of number of entries added in double
	int maxFloatGroup = 500;
	//std::cout << "maxFloatGroup is " << maxFloatGroup << "\n"; //******test
/* */	for(llong i=0; i<b_omp->at(0); i++){
		floatGroup = floatGroup + 1;
		//if (i == 0) std::cout << "In first loop \n"; //***just for test
		dip = dip + double(my[i]*your[i]);
		if(floatGroup >= maxFloatGroup){
			ip = ip + float(dip); 
			dip = 0.0;
			floatGroup = 0;
		}
	}
	int ith=0;
	#pragma omp parallel
	ith=omp_get_thread_num();
	llong j=0;
	SIMD_Float16 mine, yours, prod;
	float *__restrict__ prodv = new float[n123]; 
	//std::cout << "Past first for loop \n"; //******just for test
	for(llong i=b_omp->at(ith); j< n_omp->at(ith); j++, i+=BLOCK_SIZE){
		//std::cout << "In the middle for loop \n"; // ***** just for test
		mine.load(&my[i]);
		yours.loadu(&your[i]);
                prod.mult(mine,yours);
		prod.stream(&prodv[i]);
	}
	for(llong i=b_omp->at(ith); i<(b_omp->at(ith))+(n_omp->at(ith))*BLOCK_SIZE; i++){
		floatGroup += 1;
		dip += prodv[i];
		if(floatGroup >= maxFloatGroup){
			ip+=dip;
			floatGroup = 0;
			dip=0;
       		}
	}
	delete[] prodv;
/* */
	//std::cout << "Past second loop \n"; //***just for test
	// Finish it up
	int nth=s->get_nth()-1;
/* */	for(llong i=b_omp->at(nth)+(n_omp->at(nth))*BLOCK_SIZE; i< n123; i++) {
	//for(llong i=0; i<n123; i++){
		floatGroup = floatGroup + 1;
		dip = dip + double(my[i]*your[i]);
		if(floatGroup >= maxFloatGroup){
			ip = ip + dip; 
			dip = 0.0;
			floatGroup = 0;

		}
	}
	ip = ip + dip; // add on any additional terms in sum
	
	return ip;
}

void  SEP::inCoreFloat1D::base_1d(llong n1){
   set_n123((llong) n1); 
   push_back_n(n1);
}
namespace SEP{
dataContainer  *inCoreFloat1D::buildDataContainer() const{
  llong a=get_n(0);
  data1DFloat *x=(data1DFloat*) new data1DFloat(a);
  //std::cout << "built from 1d\n"; 
  return x;

  }
}
void  SEP::inCoreFloat2D::base_2d(llong n1,llong n2){
  llong n123=n1;push_back_n(n1); 
  n123*=(llong) n2; push_back_n(n2); 
	//std::cout << "built from 2d\n"; 
  set_n123(n123);
}
void  SEP::inCoreFloat3D::base_3d(llong n1,llong n2,llong n3){
  llong n123=n1;push_back_n(n1); 
  n123*=(llong) n2; push_back_n(n2); 
  n123*=(llong) n3; push_back_n(n3); 
  set_n123(n123);
}
void  SEP::inCoreFloat4D::base_4d(llong n1,llong n2,llong n3, llong n4){
  llong n123=n1;push_back_n(n1);  
  n123*=(llong) n2; push_back_n(n2); 
  n123*=(llong) n3; push_back_n(n3); 
  n123*=(llong) n4; push_back_n(n4); 
  set_n123(n123);
}
void  SEP::inCoreFloat5D::base_5d(llong n1,llong n2,llong n3, llong n4,llong n5){
  llong n123=n1;push_back_n(n1);  
  n123*=(llong) n2; push_back_n(n2); 
  n123*=(llong) n3; push_back_n(n3); 
  n123*=(llong) n4;push_back_n(n4); 
  n123*=(llong) n5; push_back_n(n5); 
  set_n123(n123);
}
void  SEP::inCoreFloat6D::base_6d(llong n1,llong n2,llong n3, llong n4,llong n5,llong n6){
  llong n123=n1; push_back_n(n1); 
  n123*=(llong) n2;push_back_n(n2);
  n123*=(llong) n3; push_back_n(n3);
  n123*=(llong) n4; push_back_n(n4);
  n123*=(llong) n5;push_back_n(n5);
  n123*=(llong) n6; push_back_n(n6);
  set_n123(n123);
}
void  SEP::inCoreFloat7D::base_7d(llong n1,llong n2,llong n3, llong n4,llong n5,llong n6, llong n7){
  llong n123=n1; push_back_n(n1);
  n123*=(llong) n2; push_back_n(n2);
  n123*=(llong) n3; push_back_n(n3);
  n123*=(llong) n4; push_back_n(n4);
  n123*=(llong) n5; push_back_n(n5);
  n123*=(llong) n6; push_back_n(n6);
  n123*=(llong) n7; push_back_n(n7);
     push_back_n(n1);

  set_n123(n123);
}
 bool SEP::inCoreFloat7D::isCompatible(dataContainer *x) const{
   inCoreFloat7D *y=(inCoreFloat7D*)x;
   std::vector<llong> n2; n2= y->get_ns();
   std::vector<llong> n3=this->get_ns();

   if(n3[0]==n2[0] && n3[1]==n2[1] && n2[2]==n3[2] && n2[3]==n3[3] &&
     n3[4]==n2[4] && n3[5]==n2[5] && n3[6]==n2[6]) return true;
   return false;
 
 }

 bool SEP::inCoreFloat7D::isCompatible(Space *otherSpace, bool makeSame) const{
// *******should actually check the type? *********
   inCoreFloat7D *os=(inCoreFloat7D*)otherSpace;
   std::vector<llong> n2 = os->get_ns();
   std::vector<llong> n3 = this->get_ns();
   if(this == os) return true;
   if(get_key() == os->get_key()) return true;
       
   if(n3[0]==n2[0] && n3[1]==n2[1] && n3[2]==n2[2] && n3[3]==n2[3] && n3[4]==n2[4] && n3[5]==n2[5] && n3[6]==n2[6]){
     // If you choose makeSame to be true, the key for these spaces will be the same
     if(makeSame == true) os->set_key(get_key());
     return true;
   }   
   return false;
 }

 bool SEP::inCoreFloat6D::isCompatible(dataContainer *x) const{
   inCoreFloat6D *y=(inCoreFloat6D*)x;
  std::vector<llong> n2; n2= y->get_ns();
      std::vector<llong> n3=get_ns();

   if(n3[0]==n2[0] && n3[1]==n2[1] && n2[2]==n3[2] && n2[3]==n3[3] &&
     n3[4]==n2[4] && n3[5]==n2[5] ) return true;
   return false;
 }

 bool SEP::inCoreFloat6D::isCompatible(Space *otherSpace, bool makeSame) const{
// *******should actually check the type? *********
   inCoreFloat6D *os=(inCoreFloat6D*)otherSpace;
   std::vector<llong> n2 = os->get_ns();
   std::vector<llong> n3 = this->get_ns();
   if(this == os) return true;
   if(get_key() == os->get_key()) return true;       
   if(n3[0]==n2[0] && n3[1]==n2[1] && n3[2]==n2[2] && n3[3]==n2[3] && n3[4]==n2[4] && n3[5]==n2[5]){
     // If you choose makeSame to be true, the key for these spaces will be the same
     if(makeSame == true) os->set_key(get_key());     
     return true;
   }
   return false;
 }

  bool SEP::inCoreFloat5D::isCompatible(dataContainer *x) const{
   inCoreFloat5D *y=(inCoreFloat5D*)x;
   std::vector<llong> n3=get_ns();
   std::vector<llong> n2; n2= y->get_ns();
   if(n3[0]==n2[0] && n3[1]==n2[1] && n2[2]==n3[2] && n2[3]==n3[3] &&
     n3[4]==n2[4]  ) return true;
   return false;
 }

 bool SEP::inCoreFloat5D::isCompatible(Space *otherSpace, bool makeSame) const{
// *******should actually check the type? *********
   inCoreFloat5D *os=(inCoreFloat5D*)otherSpace;
   std::vector<llong> n2 = os->get_ns();
   std::vector<llong> n3 = this->get_ns();
   if(this == os) return true;
   if(get_key() == os->get_key()) return true;     
   if(n3[0]==n2[0] && n3[1]==n2[1] && n3[2]==n2[2] && n3[3]==n2[3] && n3[4]==n2[4]){
     // If you choose makeSame to be true, the key for these spaces will be the same
     if(makeSame == true) os->set_key(get_key());     
     return true;
   }   
   return false;
 }

 bool SEP::inCoreFloat4D::isCompatible(dataContainer *x) const{
   inCoreFloat4D *y=(inCoreFloat4D*)x;
   std::vector<llong> n3=get_ns();

   std::vector<llong> n2; n2= y->get_ns();
   if(n3[0]==n2[0] && n3[1]==n2[1] && n2[2]==n3[2] && n2[3]==n3[3] ) return true;
   return false;
 }

 bool SEP::inCoreFloat4D::isCompatible(Space *otherSpace, bool makeSame) const{
// *******should actually check the type? *********
   inCoreFloat4D *os=(inCoreFloat4D*)otherSpace;
   std::vector<llong> n2 = os->get_ns();
   std::vector<llong> n3 = this->get_ns();
   if(this == os) return true;
   if(get_key() == os->get_key()) return true;     
   if(n3[0]==n2[0] && n3[1]==n2[1] && n3[2]==n2[2] && n3[3]==n2[3]){
     // If you choose makeSame to be true, the key for these spaces will be the same
     if(makeSame == true) os->set_key(get_key());
     return true;
   }  
   return false;
 }

 bool SEP::inCoreFloat3D::isCompatible(dataContainer *x) const{
   inCoreFloat3D *y=(inCoreFloat3D*)x;
       std::vector<llong> n3=get_ns();

  std::vector<llong> n2; n2= y->get_ns();
   if(n3[0]==n2[0] && n3[1]==n2[1] && n2[2]==n3[2]  ) return true;
   return false;
 }

 bool SEP::inCoreFloat3D::isCompatible(Space *otherSpace, bool makeSame) const{
// *******should actually check the type? *********
   inCoreFloat3D *os=(inCoreFloat3D*)otherSpace;
   std::vector<llong> n2 = os->get_ns();
   std::vector<llong> n3 = this->get_ns();
   if(this == os) return true;
   if(get_key() == os->get_key()) return true;     
   if(n3[0]==n2[0] && n3[1]==n2[1] && n3[2]==n2[2]){
     // If you choose makeSame to be true, the key for these spaces will be the same
     if(makeSame == true) os->set_key(get_key());
     return true;
   }
   return false;
 }

 bool SEP::inCoreFloat2D::isCompatible(dataContainer *x) const{
   inCoreFloat2D *y=(inCoreFloat2D*)x;
   std::vector<llong> n3=get_ns();
   std::vector<llong> n2; n2= y->get_ns();
   if(n3[0]==n2[0] && n3[1]==n2[1]  ) return true;
   return false;
 }

  bool SEP::inCoreFloat2D::isCompatible(SEP::Space *otherSpace, bool makeSame) const{
// *******should actually check the type? *********
   inCoreFloat2D *os=(inCoreFloat2D*)otherSpace;
   std::vector<llong> n2 = os->get_ns();
   std::vector<llong> n3 = this->get_ns();
   if(this == os) return true;
   if(get_key() == os->get_key()) return true;  
   if(n3[0]==n2[0] && n3[1]==n2[1]){
     // If you choose makeSame to be true, the key for these spaces will be the same
     if(makeSame == true) os->set_key(get_key());
     return true;
   }   
   return false;
 }

bool SEP::inCoreFloat1D::isCompatible(dataContainer *x) const{
  inCoreFloat1D *y=(inCoreFloat1D*)x;
  std::vector<llong> n3=get_ns();

  std::vector<llong> n2; n2= y->get_ns();
  if(n3[0]==n2[0]) return true;
  return false;
}

bool SEP::inCoreFloat1D::isCompatible(SEP::Space *otherSpace, bool makeSame) const{
 // *******should actually check the type? *********
   inCoreFloat1D *os=(inCoreFloat1D*)otherSpace;
   std::vector<llong> n2 = os->get_ns();
   std::vector<llong> n3 = this->get_ns();
   if(this == os) return true;
   if(get_key() == os->get_key()) return true;
   if(n3[0]==n2[0]){
     // If you choose makeSame to be true, the key for these spaces will be the same
     if(makeSame == true) os->set_key(get_key());
     return true;
   }
   return false;
}
