#ifndef INCOREFLOAT_H
#define INCOREFLOAT_H 1
#include "Space.h"
#include "hypercube.h"
#include "DataInCoreFloat.h"
//#include <omp.h>
#include "myomp.h"
#include "axis.h"
#include <iostream>
#include "s_reg.h"

typedef long long llong;

namespace SEP {


class inCoreFloat: public s_reg {
  public:
  inCoreFloat(){ init=false;}
  inCoreFloat(char *tag,std::vector<axis> axes);

  virtual void linearCombo(float a, dataContainer  * x, float b, dataContainer * y) const;
  virtual double innerProd(dataContainer * x, dataContainer * y) const;
  virtual void addInv(dataContainer * x) const;
  virtual void addInv(dataContainer  * x, dataContainer * xNeg) const ;
  virtual void scale(float a, dataContainer * x)const ;
  virtual void scale(float a, dataContainer * x,dataContainer * ax)const ;
  virtual void zeroElement(dataContainer * x) const ;
  virtual void random(dataContainer * x) const ;
  virtual void print_it(){std::cerr<<"incorefloat"<<std::endl;}
  virtual ~inCoreFloat(){ }

  void set_n123(llong n123){ this->n123=n123;}
  llong get_n123(){return n123;}
  void set_n(std::vector<llong> n) {this->n = n;}
  void push_back_n(llong i){ n.push_back(i);}
  std::vector<llong> get_ns() const{
    return n;
  }
  llong get_n(int iax) const{ return n[iax];} 

  private:
    bool init;
    llong n123;
    std::vector<llong> b_omp,e_omp;
    int beg_vector;
    int nomp;
    std::vector<llong> n;
    std::vector<llong> o; 
    std::vector<llong> d; 


};

class inCoreFloat1D: public inCoreFloat{
  public:
    inCoreFloat1D(){ }
    inCoreFloat1D(llong n){ base_1d(n);}
    virtual bool isCompatible(dataContainer *x) const;
    virtual bool isCompatible(Space *otherSpace, bool makeSame=false) const;
    virtual void print_it(){std::cerr<<"incorefloat1d"<<std::endl;}
    virtual dataContainer *buildDataContainer()const;
    void base_1d(llong n1);

};

class inCoreFloat2D: public inCoreFloat{
  public:
    inCoreFloat2D(){ }
    inCoreFloat2D(llong *n){ 
      base_2d(n[0],n[1]);
    }
    inCoreFloat2D(std::vector<llong> n){ 
      base_2d(n[0],n[1]);
    }
    inCoreFloat2D(llong n1, llong n2){
      base_2d(n1,n2);
    }
    inCoreFloat2D(std::vector<axis> axes){
      this->init_ndim(axes);
      base_2d(get_axis(1).n,get_axis(2).n);
    }
    inCoreFloat2D(char *tag){
      tag_init(tag);
      base_2d(get_axis(1).n,get_axis(2).n);
    }


  //inCoreFloat(char *tag,std::vector<axis> axes);

    virtual bool isCompatible(dataContainer *x) const;
    virtual bool isCompatible(Space *otherSpace, bool makeSame=false) const;
    virtual dataContainer *buildDataContainer() const{
      dataContainer *x=(dataContainer*) new data2DFloat(get_ns());
      return x;
    }

  private:
    void base_2d(llong n1,llong n2);

};

class inCoreFloat3D: public inCoreFloat{
  public:
    inCoreFloat3D(){}
    inCoreFloat3D(std::vector<llong> n){ 
      base_3d(n[0],n[1],n[2]);
    }
    inCoreFloat3D(llong *n){ 
      base_3d(n[0],n[1],n[2]);
    }
    inCoreFloat3D(llong n1, llong n2, llong n3){
      base_3d(n1,n2,n3);
    }
    virtual bool isCompatible(dataContainer *x) const;
    virtual bool isCompatible(Space *otherSpace, bool makeSame=false) const;
    virtual dataContainer *buildDataContainer() const{
      dataContainer *x=(dataContainer*) new data3DFloat(get_ns());
      return x;
    }

  private:
    void base_3d(llong n1,llong n2, llong n3);

};

class inCoreFloat4D: public inCoreFloat{
  public:
    inCoreFloat4D(){ }
    inCoreFloat4D(std::vector<llong> n){ 
      base_4d(n[0],n[1],n[2],n[3]);
    }
    inCoreFloat4D(llong *n){ 
      base_4d(n[0],n[1],n[2],n[3]);
    }
    inCoreFloat4D(llong n1, llong n2, llong n3, llong n4){
      base_4d(n1,n2,n3,n4);
    }
    virtual bool isCompatible(dataContainer *x) const;
    virtual bool isCompatible(Space *otherSpace, bool makeSame=false) const;
    virtual dataContainer *buildDataContainer() const{
      dataContainer *x=(dataContainer*) new data4DFloat(get_ns());
      return x;
    }

  private:
    void base_4d(llong n1,llong n2, llong n3, llong n4);

};
class inCoreFloat5D: public inCoreFloat{
  public:
    inCoreFloat5D(){ }
    inCoreFloat5D(std::vector<llong> n){ 
      base_5d(n[0],n[1],n[2],n[3],n[4]);
    }
    inCoreFloat5D(llong *n){ 
      base_5d(n[0],n[1],n[2],n[3],n[4]);
    }
    inCoreFloat5D(llong n1, llong n2, llong n3, llong n4, llong n5){
      base_5d(n1,n2,n3,n4,n5);
    }
    virtual bool isCompatible(dataContainer *x) const;
    virtual bool isCompatible(Space *otherSpace, bool makeSame=false) const;
    virtual dataContainer *buildDataContainer() const{
      dataContainer *x=(dataContainer*) new data5DFloat(get_ns());
      return x;
    }

  private:
    void base_5d(llong n1,llong n2, llong n3, llong n4, llong n5);

};
class inCoreFloat6D: public inCoreFloat{
  public:
    inCoreFloat6D(){ }
    inCoreFloat6D(std::vector<llong> n){ 
      base_6d(n[0],n[1],n[2],n[3],n[4],n[5]);
    }
    inCoreFloat6D(llong *n){ 
      base_6d(n[0],n[1],n[2],n[3],n[4],n[5]);
    }
    inCoreFloat6D(llong n1, llong n2, llong n3, llong n4, llong n5, llong n6){
      base_6d(n1,n2,n3,n4,n5,n6);
    }
    virtual bool isCompatible(dataContainer *x) const;
    virtual bool isCompatible(Space *otherSpace, bool makeSame=false) const;
    virtual dataContainer *buildDataContainer() const{
      dataContainer *x=(dataContainer*) new data6DFloat(get_ns());
      return x;
    }

  private:
    void base_6d(llong n1,llong n2, llong n3, llong n4, llong n5, llong n6);

};
class inCoreFloat7D: public inCoreFloat{
  public:
    inCoreFloat7D(){ }
    inCoreFloat7D(std::vector<llong> n){ 
      base_7d(n[0],n[1],n[2],n[3],n[4],n[5],n[6]);
    }
    inCoreFloat7D(llong *n){ 
      base_7d(n[0],n[1],n[2],n[3],n[4],n[5],n[6]);
    }
    inCoreFloat7D(llong n1, llong n2, llong n3, llong n4, llong n5, llong n6, llong n7){
      base_7d(n1,n2,n3,n4,n5,n6,n7);
    }
    virtual dataContainer *buildDataContainer() const{
      dataContainer *x=(dataContainer*) new data7DFloat(get_ns());
      return x;
    }
    virtual bool isCompatible(dataContainer *x) const;
    virtual bool isCompatible(Space *otherSpace, bool makeSame=false) const;

  private:
    void base_7d(llong n1,llong n2, llong n3, llong n4, llong n5, llong n6, llong n7);

};

}

#endif
