#include "spraysum.h"
#include <iostream>
#include <cstdlib>
#include <algorithm> 


bool SEP::Spraysum::Forward(bool add, Vector *model, Vector *data){
  /// Forward operator for spray and sum
  ///
  if(!add) data->zeroElement();  // add=False we will erase the data

  // Initilization
  Vector *model_vec = model;
  Vector *data_vec = data;

  // Do the operator one by one
  Vector *prev = model_vec;
  Vector *next = NULL;
  for(int i = 0; i < op_number; i++) {
    
    if(i==op_number - 1){
      next = data_vec;
    }
    else {
      next = new Vector(inter_space[i]);
    }
    spraysum1d(prev, next, i, true);
    if(i > 0){delete prev;}
    prev = next;
  
  }
  return true;
}

bool SEP::Spraysum::Adjoint(bool add, Vector *model, Vector *data) {
  /// For Adjoint operator, we just need to call Forward operator
  ///
  if(!add) model->zeroElement();  // add=False we will erase the data

  // Initilization
  Vector *model_vec = model;
  Vector *data_vec = data;
  //init(in_vec,out_vec);

  // Do the operator one by one
  Vector *prev = data_vec;
  Vector *next = NULL;
  for(int i = op_number-1; i >= 0; i--) {
  
    if (i==0) {
      next = model_vec;
    }
    else {
      next = new Vector(inter_space[i-1]);
    }
    spraysum1d(prev, next, i, false);
    if(i<op_number - 1){delete prev;}
    prev = next;
  
  }
  return true;
}


void SEP::Spraysum::spraysum1d(Vector *in_vec, Vector *out_vec, int iop, bool direction) {
  /// 1D spray and sum operator
  /// op_type == true means spray, op_type == false means sum
  if((op_direction[iop] == false) and (direction == true)) {
    Sum(in_vec, out_vec, iop);
  }
  if((op_direction[iop] == true) and (direction == true)) {
    Spray(in_vec, out_vec, iop);
  }
  if((op_direction[iop] == false) and (direction == false)) {
    Spray(in_vec, out_vec, iop);
  }
  if((op_direction[iop] == true) and (direction == false)) {
    Sum(in_vec, out_vec, iop);
  }
}

void SEP::Spraysum::Sum(Vector *in_vec, Vector *out_vec, int iop) {
  /// Sum function sums over a particular axis
  /// [big](middle)[small]
  /// middle represents the summation axis
  /// break the vector into 3 blocks, and loop over each of them

  float *__restrict__ in = ((dataInCoreFloat*)in_vec->getDataContainer())->getFloatPtr();
  float *__restrict__ out = ((dataInCoreFloat*)out_vec->getDataContainer())->getFloatPtr();

  long long big = bigs[iop];
  long long middle = middles[iop];
  long long small = smalls[iop];
  long long block = blocks[iop];
  for(long long ibig = 0; ibig < big;  ibig++) {
    for(long long imid = 0; imid < middle; imid++) {
      for(long long ismall = 0; ismall < small; ismall++) {
        out[ismall + ibig * small] += in[ismall + imid * small + ibig * block];
      }
    }
  }

}

void SEP::Spraysum::Spray(Vector *in_vec, Vector *out_vec, int iop) {
  /// Spray function sums over a particular axis
  /// [big](middle)[small]
  /// middle represents the spreading axis
  /// break the vector into 3 blocks, and loop over each of them

  float *__restrict__ in = ((dataInCoreFloat*)in_vec->getDataContainer())->getFloatPtr();
  float *__restrict__ out = ((dataInCoreFloat*)out_vec->getDataContainer())->getFloatPtr();

  long long big = bigs[iop];
  long long middle = middles[iop];
  long long small = smalls[iop];
  long long block = blocks[iop];
  for(long long ibig = 0; ibig < big;  ibig++) {
    for(long long imid = 0; imid < middle; imid++) {
      for(long long ismall = 0; ismall < small; ismall++) {
        out[ismall + imid * small + ibig * block] += in[ismall + ibig * small];
      }
    }
  }

}




void SEP::Spraysum::init() {
  //inCoreFloat *in_space  = (inCoreFloat *)in_vec->getSpace();
  //inCoreFloat *out_space = (inCoreFloat *)out_vec->getSpace();
  inCoreFloat *in_space = (inCoreFloat *)domain;
  inCoreFloat *out_space = (inCoreFloat *)range;
  std::vector<long long> in_dims = in_space->get_ns();
  std::vector<long long> out_dims = out_space->get_ns();
  std::vector<long long> temp_dims = in_dims;
  init_dim = in_dims;
  dimension = (int)in_dims.size();
  op_number = 0;


  for(int i = 0; i < dimension; i++) {
    if(in_dims[i]==out_dims[i])
      continue;
    if(in_dims[i] == 1 and out_dims[i] != 1) {
      // sprading along this direction
      op_dim.push_back(i);
      op_number++;
      op_direction.push_back(true);
      temp_dims[i] = out_dims[i];
      set_bms(temp_dims,i);
    }
    if(in_dims[i] != 1 and out_dims[i] == 1) {
      // sprading along this direction
      op_dim.push_back(i);
      op_number++;
      op_direction.push_back(false);
      set_bms(temp_dims,i);
      temp_dims[i] = out_dims[i];
    }
    // Now construct the space with stupid method
    if (dimension == 1) {
      inCoreFloat *temp = new inCoreFloat1D(temp_dims[0]);
      inter_space.push_back(temp);
    }
    if (dimension == 2) {
      inCoreFloat *temp = new inCoreFloat2D(temp_dims);
      inter_space.push_back(temp);
    }
    if (dimension == 3) {
      inCoreFloat *temp = new inCoreFloat3D(temp_dims);
      inter_space.push_back(temp);
    }
    if (dimension == 4) {
      inCoreFloat *temp = new inCoreFloat4D(temp_dims);
      inter_space.push_back(temp);
    }
    if (dimension == 5) {
      inCoreFloat *temp = new inCoreFloat5D(temp_dims);
      inter_space.push_back(temp);
    }
    if (dimension == 6) {
      inCoreFloat *temp = new inCoreFloat6D(temp_dims);
      inter_space.push_back(temp);
    }
    if (dimension == 7) {
      inCoreFloat *temp = new inCoreFloat7D(temp_dims);
      inter_space.push_back(temp);
    }

  } // end of for

}

void SEP::Spraysum::set_bms(std::vector<long long> temp_dims, int index) {

  long long big = 1, middle = 1, small = 1, block = 1;

  for(int i = 0; i < index; i++)
    big = big*temp_dims[i];
  middle = temp_dims[index];
  for(int i = index + 1; i < dimension; i++)
    small = small*temp_dims[i];
  block = small*middle;

  bigs.push_back(big);
  middles.push_back(middle);
  smalls.push_back(small);
  blocks.push_back(block);
}

void SEP::Spraysum::OP_info() {
  std::cout << "We are working at " << dimension << "D vector." << std::endl;
  std::cout << "We need to use " << op_number << " steps to complete operation." << std::endl;

  std::cout << "Starting vector dimension: "<< std::endl;
  for(int i = 0; i < dimension - 1; i++){
    std::cout << init_dim[i] << " == ";
  }
  std::cout << init_dim[dimension-1] << std::endl;

  for(int i = 0; i < op_number; i++) {
    std::cout << "At operation " << i + 1;
    std::cout << ", working at dim " << op_dim[i]+1;
    if (op_direction[i] == true)
      std::cout << " with spray operator." << std::endl;
    if (op_direction[i] == false)
      std::cout << " with sum operator." << std::endl;
    std::vector<long long> temp = inter_space[i]->get_ns();
    for(int j = 0; j < dimension-1; j++)
      std::cout << temp[j] << " == ";
    std::cout<< temp[dimension-1] <<std::endl;
  }

}

void SEP::Spraysum::clean() {
  
  op_direction.clear();
  op_dim.clear();
  bigs.clear();
  middles.clear();
  smalls.clear();
  blocks.clear();
  init_dim.clear();
  for(int i = 0; i < op_number; i++) {
    inCoreFloat *temp = inter_space[i];
    delete temp; temp = NULL;
  }
  inter_space.clear();

}