/* SEP Unit Test 11
    Purpose: Test MultiDataContainer, MultiSpace, MultiVector
*/
#include <iostream>
#include <cstdlib>
#include "../inCoreFloat.h"
#include "../DataInCoreFloat.h"
#include "../Vector.h"
#include <assert.h>

using namespace std;
using namespace SEP;

int main(){
  try{
    cout << "TEST 12: Create a MultiSpace from two spaces, check compatibility and getSpace methods ... \n"; 

    int n1 = 1000; // set dimensions of spaces to test
    int n2 = 27;
    inCoreFloat *oneD = new inCoreFloat1D(n1);
    inCoreFloat *oneD2 = new inCoreFloat1D(n2);
    // create a multispace from two spaces
    MultiSpace *oneD3 = new MultiSpace(oneD, oneD2);
    // get dimensions of multispace
    long long n1n2 = oneD3->get_n123();
    if(int(n1n2) - (n1+n2) > 1e-6){
      cout << "ERROR: dimension of multispace should be " << n1+n2 << " but really is " << n1n2 << "\n";
    }
    // get and check compatibility of sub-spaces 
    Space *oneDTest = oneD3->getSpace1();
    Space *oneD2Test = oneD3->getSpace2();
    bool check1 = oneD->isCompatible(oneDTest);
    bool check2 = oneD2->isCompatible(oneD2Test);
    if(check1 == false || check2 == false){
      cout << "ERROR: subspaces making up multispace are incompatible with original spaces used to create multispace \n";
    }


    // create data containers and fill with random stuff
    Vector x1(oneD);
    x1.random();
    Vector x2(oneD2);
    x2.random();
    MultiVector x3(&x1, &x2);
    float vecDiff = 0;
    float *a1 = ((dataInCoreFloat*)x1.getDataContainer())->getFloatPtr();
    float *a2 = ((dataInCoreFloat*)x2.getDataContainer())->getFloatPtr();
    float *a3p1 = ((dataInCoreFloat*)((x3.getDataContainer())->getData1()))->getFloatPtr();
    float *a3p2 = ((dataInCoreFloat*)((x3.getDataContainer())->getData2()))->getFloatPtr();
    for(int i=0; i<n1; ++i){
      vecDiff += abs(a3p1[i]-a1[i]);
    }
    for(int i=0; i<n2; ++i){
      vecDiff += abs(a3p2[i]-a2[i]);
    }
    if(vecDiff/n1n2 > 1e-6){
      cout << "ERROR: MultiVector not filled in properly by two smaller vectors. \n";
      cout << "First ten entries of each part (sub vector, multi vector): \n";
      for(int i=0; i<20; ++i){
        cout << a1[i] << " , " << a3p1[i] << "\n";
      }
      cout << "Next part: \n";
      for(int i=0; i<20; ++i){
        cout << a2[i] << " , " << a3p2[i] << "\n";
      }
    }
    
    // clean up
    delete oneD;
    delete oneD2;
    delete oneD3;
  }
  catch (SEPException &e) {
    e.write(cout);
    exit(EXIT_FAILURE);
  }
}
