/** SEP Unit test 16
    Purpose: test CGLS
    Created by Yinbin 2014-07-22
    Last modified:
*/


#include <iostream>
#include <cstdlib>
#include "seplib.h"
#include <assert.h>
#include<stdbool.h>
#include<string>

#include "../inCoreFloat.h"
#include "../Vector.h"
#include "../Operator.h"
#include "../MultiOperator.h"
#include "../spraysum.h"

#include "../Algorithm.h"
#include "../CGLSAlg.h"
#include "../CGAlg.h"
#include "../matmul.h"

using namespace std;
using namespace SEP;

int main() {

  try {
    long long m1 = 100,m2=100, n = 100;
    inCoreFloat1D mspace(n);
    inCoreFloat1D d1space(m1), d2space(m2);
    inCoreFloat1D dspace(m1+m2);
    inCoreFloat2D op1space(m1,n),op2space(m2,n);

    Vector mvec(&mspace), inversion(&mspace);
    Vector d1vec(&d1space), d2vec(&mspace), dvec(&dspace);
    Vector Op1vec(&op1space), Op2vec(&op2space);
    dvec.zeroElement();
    mvec.random(); inversion.zeroElement();

    Op1vec.random();
    Matmul testOP(&mspace,&d1space,&Op1vec);
    ScaleOP L2(&mspace,&mspace,1.e-3);
    Operator &L1 = testOP;  // test passing by reference


    ColOP L(L1,L2,&dspace);

    // Construct data
    L.Forward(false,&mvec,&dvec);
    L.DotTest(true);


    L1.Forward(false,&mvec,&d1vec);
    L2.Forward(false,&mvec,&d2vec);
    double res = dvec.innerProd(dvec);
    cout << res << "is the data" <<endl;
    double rnorm=1., nrnorm=1.;
    int maxiter = 6; 
    //CGLSStep stepper(OP,model,data,rnorm,nrnorm);
    int iter = 0;
    cout << "Running with CG methods in GEE." << endl;
    CGAlg cg(L,inversion,dvec,rnorm,nrnorm,iter,1.e-6,1.e-6,6,true);
    cg.run();
    cout << "Running standard CGLS methods. should see exact the same output." <<endl;
    inversion.zeroElement();
    L.Forward(false,&mvec,&dvec);
    rnorm = 1.; nrnorm = 1.;
    CGLSAlg cg1(L,inversion,dvec,rnorm,nrnorm,1.e-6,1.e-6,6,true);
    cg1.run();


    cout << "Testing CG regularized Solver. " << endl;
    iter=0; rnorm = 1.; nrnorm = 1.;
    inversion.zeroElement();
    CGRegAlg cg2(L1,L2,inversion,d1vec,d2vec,rnorm,nrnorm,iter,1.e-6,1.e-6,6,true);
    cg2.run();
    return 0;
  }
  catch (SEPException &e) {
    e.write(cout);
    exit(EXIT_FAILURE);
  }


}