/** SEP Unit test 18
    Purpose: test operator that chains vectors&operators
    Created by Yinbin 2014-08-21
    Last modified:
*/


#include <iostream>
#include <cstdlib>
#include "seplib.h"
#include "../inCoreFloat.h"
#include "../DataInCoreFloat.h"
#include "../Vector.h"
#include "../Operator.h"
#include "../MultiOperator.h"
#include "../spraysum.h"
#include <assert.h>
#include<stdbool.h>
#include<string>
#include "../matmul.h"

using namespace std;
using namespace SEP;

int main() {

  try {
     cout << "========================================"<<endl;
     cout << "                                        "<<endl;
     cout << "========================================"<<endl;
     cout << "TEST 18: Test MultiOperator.\n";

    // Preprocessing, construct space and vectors.
    long long m = 40, l = 60, n = 50, k = 30;

    inCoreFloat1D mspace(n);
    inCoreFloat1D fspace(l),fspace2(k);
    inCoreFloat1D dspace(m);
    inCoreFloat1D bigspace(l+k);
    inCoreFloat2D L1space(l,n);
    inCoreFloat2D L2space(m,l);
    inCoreFloat2D L3space(k,n);

    Vector mvec(&mspace);
    Vector fvec(&fspace),f2vec(&fspace2);
    Vector dvec(&dspace);
    MultiVector mfvec(&fvec,&f2vec);
    Vector L1vec(&L1space); L1vec.random();
    Vector L2vec(&L2space); L2vec.random();
    Vector L3vec(&L3space); L3vec.random();

    // Building element operators
    Matmul L1(&mspace,&fspace,&L1vec);
    Matmul L2(&fspace,&dspace,&L2vec);
    Matmul L3(&mspace,&fspace2,&L3vec);
    ScaleOP alphaOP(&dspace,&dspace,0.5);

    // Building transpose operators
    cout << "Testing d=L' m operator: " << endl;
    TransOP Transpose(L1);
    Transpose.DotTest(true); 

    // Building row block operators
    cout << "Testing d=L2*L1 m operator: " << endl;
    ChainOP RowOP(L1,L2);
    RowOP.DotTest(true);    
    cout << "Testing d=alpha*L2*L1 m operator: " << endl;
    ChainOP ROP(RowOP,alphaOP);
    ROP.DotTest(true);

    // Building Column operator
    ColOP colOP(L1,L3,&mfvec);
    
    cout << "Testing [L1;L2] column operators:" <<endl;
    //colOP.Forward(true,&mvec,&mfvec);
    colOP.DotTest(true); 

  }
  catch (SEPException &e) {
    e.write(cout);
    exit(EXIT_FAILURE);
  }


}