/** SEP Unit test 19
    Purpose: test regularized solver. CG, Split-Bregman, IRLS, HBCD
    Created by Yinbin 2014-07-22
    Last modified:    2014-08-27
*/

#include <iostream>
#include <cstdlib>
#include "seplib.h"
#include <assert.h>
#include<stdbool.h>
#include<string>


#include "../inCoreFloat.h"
#include "../Vector.h"
#include "../Operator.h"

#include "../Algorithm.h"
#include "../CGAlg.h"
#include "../IRLSAlg.h"
#include "../matmul.h"

using namespace std;
using namespace SEP;



int main() {

  try {

    long long m = 100;
    long long n = 200;
    double    alpha = 0.02;
    float eps = 1.e-4;

    // Set up model space and two data space
    inCoreFloat1D mspace(n);
    inCoreFloat1D dspace1(m);
    inCoreFloat1D &dspace2 = mspace;
    inCoreFloat2D Lspace(m,n);
    inCoreFloat2D Regspace(m,m);

    // Initialize the vector in model&data space
    Vector mvec(&mspace), inversion(&mspace);
    Vector dvec1(&dspace1), dvec2(&dspace2);
    Vector Lvec(&Lspace);
    mvec.random();
    Lvec.random();


    // Generate basic operators
    Matmul   L(&mspace,&dspace1,&Lvec);
    ScaleOP  Reg(&mspace,&dspace2,alpha);
    IRLSReg Wr(&mspace,&mspace,1.);

    // Joining the Operator, and create the data.
    L.Forward(false,&mvec,&dvec1);
    dvec2.zeroElement();
    inversion.zeroElement();

    // Setting up CG algorithm
    double rnorm=1., nrnorm=1.;
    int maxiter = 10, iter = 0;
    L.DotTest(true);
    Reg.DotTest(true);
    Wr.DotTest(true);
    IRLS_Solver irls(L,Reg, Wr,inversion,dvec1,dvec2,iter);
    irls.SetPar(maxiter, 5, 1.e-3, 1.e-3);
    irls.run();
    return 0;
    
  }
  catch (SEPException &e) {
    e.write(cout);
    exit(EXIT_FAILURE);
  }


}