/** SEP Unit Test 2
    Purpose: test Vector clone and linear combination functions.
*/

#include <iostream>
#include <cstdlib>
#include "../Vector.h"
#include "../inCoreFloat.h"
#include "../DataInCoreFloat.h"

using namespace std;
using namespace SEP;

int main() {
	try {
		cout << "TEST 2: Clone a vector (deep copy), check space pointer and dataContainer entries of clone, test linear combination of two vectors.\n";

		/* Create a vector space, a vector, assign its elements to be consecutive integers, then clone the array */
		int n = 1000000; // sets dimension of space to test
		inCoreFloat1D *oneD = new inCoreFloat1D(n); // create n-D space 
		Vector v1(oneD); // create vector with a data container from space
 		dataInCoreFloat *x= ((dataInCoreFloat*)v1.getDataContainer());
		float *array1=x->getFloatPtr();
		double indexDbl = 1.0; // will be used to fill array with consecutive #
		for (int i=0; i<n; i++){
			array1[i] = float(indexDbl);
			indexDbl = indexDbl + double(1.0);
		}	
    		Vector v2 = v1.clone(); // make a clone of v1 (deep copy)
		//dataInCoreFloat *x2 = ((dataInCoreFloat*)v2.getDataContainer());
		//float *array2 = x2->getFloatPtr();
		float *array2 = ((dataInCoreFloat*)v2.getDataContainer())->getFloatPtr();
		float maxDiff = 0.0; // will take max over vector
		int maxDiffIn = 0; // will keep last index of large relative err.
		for (int i=0; i<n; i++) {
			if(abs((array2[i]-array1[i])/array1[i]) >= maxDiff){
				maxDiffIn = i;
				maxDiff = abs((array2[i]-array1[i])/array1[i]);
			}
		}
		if (maxDiff > 1.0e-6) {
			cout << "ERROR: There was a problem copying the entries of the dataContainer when cloning a vector in test 2. \n";
			cout << "The maximum relative difference was "<< maxDiff << " and\n";
			cout << " it occured in spot " << maxDiffIn << ".\n";
			cout << "What's actually and should be stored there is: \n";
			cout << array2[maxDiffIn] << " instead of " << array1[maxDiffIn] << "\n";
			cout << "In first ten spots, it should and actually stores:  \n";
			for (int i=0; i<min(n,10); i++) {
				cout << array1[i] << " - " << array2[i] << "\n";
			} 
		}    

		// Test whether the clone function assigns the vector to the right space 
	    	bool spaceCheck1 = v1.inSpace(oneD); // check if v1 points to right space
	    	if (spaceCheck1 == 0){ // If v1 doesn't point to right space
	    		cout << "ERROR: The inSpace function on Vector 1 doesn't point to the correct space in test 2! \n";
			cout << "Because of this, it is unknown whether the clone of Vector 1 points to the correct space in test 2. \n";
	    	} else{ // check if v2 points to right space
	    		bool spaceCheck2 = v2.inSpace(oneD);
			if(spaceCheck2 == 0){
				cout << "Vector clone function isn't assigning clone vector to correct space in test 2. \n";
			}
	    	}

	    	/* Test whether the linear combination function works */
	    	// Overwrite v2 with integers 2 to n+1
	    	indexDbl = 2.0;
	    	for (int i=0; i<n; i++) {
			array2[i] = float(indexDbl);
			indexDbl = indexDbl + 1.0;
	   	}
	    	// This should perform v2 = 1*v1 + 2*v2
	    	v2.linearCombo(2., v1, 1.);
	    	float *array22 = ((dataInCoreFloat*)v2.getDataContainer())->getFloatPtr();
	    	maxDiff = 0.000001;
	    	maxDiffIn = 0;
	    for (int i=0; i<n; i++){
		if(float(abs(double(array22[i]) - double(5)-double(3*i)))/abs(array22[i]) >= maxDiff) {
			maxDiffIn = i;
			maxDiff = float(abs(double(array22[i])-double(5)-double(3*i)))/abs(array22[i]);
			}
	    }
	    if(maxDiff > 1e-6){
		cout << "Linear combination was not done properly in test 2.\n";
		cout << "Max relative error was " << maxDiff << " in spot " << maxDiffIn << ".\n";
		cout << "What's actually stored there is " << array22[maxDiffIn] << "\n";
		cout << "What should be there is " << 5+3*maxDiffIn << "\n";
		cout << "The linear combination actually returned (left), should have returned (right) first 10 entries. \n" ;
	    	for (int i=0; i<min(n,10); i++) {
			cout << array22[i] << " - " << 5+3*i << "\n";
		}
	    }
           delete oneD;
	}
	catch (SEPException &e) {
	    e.write(cout);
	    exit(EXIT_FAILURE);
   	}
}
