/** SEP Unit test 21
    Purpose: test for MultiSpace
    Created by YM 2014-11-10
    Last modified: 2015-01-13
*/

#include <iostream>
#include <cstdlib>
#include "../Space.h"
#include "../Vector.h"
#include "../Operator.h"

using namespace std;
using namespace SEP;

int main() {

  try {
     cout << "========================================"<<endl;
     cout << "                                        "<<endl;
     cout << "========================================"<<endl;
     cout << "TEST 21: Test MultiSpace.\n";



/* Test out creating a vector space, a vector, and assigning its elements to be consecutive integers, check compatibility of spaces */
     int n = 1605; // sets dimension of space to test
     inCoreFloat1D  *space1 = new inCoreFloat1D(1500);
     inCoreFloat1D  *space2 = new inCoreFloat1D(n); // create n-D space
     inCoreFloat1D  *space3 = new inCoreFloat1D(n); // create compatible space

     Vector v1(space1), u1(space1);
     Vector v2(space2), u2(space2);
     Vector v3(space3), u3(space3);
     v1.random(); v2.random(); v3.random();
     u1.random(); u2.random(); u3.random();
     double inn = v1.innerProd(u1)+v2.innerProd(u2)+v3.innerProd(u3);

     cout << "========================================"<<endl;
     std::cout << "Joining V=[v1 v2]' and U = [U1 U2]'" << std::endl;
     MultiVector V(&v1,&v2), U(&u1,&u2);
     std::cout << "Vbig = [v3 V]' and Ubig = [u3 U]" << std::endl;
     MultiVector Vbig(&v3,&V),Ubig(&u3,&U);

     cout << "========================================"<<endl;
     std::cout << "Computing norm(V+U)^2" << std::endl;
     V.linearCombo(1.,U,1.);
     double test = V.innerProd(V);
     std::cout << "Inner Prod = " << test << std::endl;

     std::cout << "Computing norm(v1+u1)^2+norm(v2+u2)^2" << std::endl;
     v1.linearCombo(1.,u1,1.);
     v2.linearCombo(1.,u2,1.);
     v3.linearCombo(1.,u3,1.);
     test = v1.innerProd(v1) + v2.innerProd(v2);
     std::cout << "Should be the same = " << test << std::endl;




     cout << "========================================"<<endl;
     std::cout << "<v1,u1> + <v2,u2> + <v3,u3> =" <<inn<< std::endl;
     test = Vbig.innerProd(Ubig);
     std::cout << "<Vbig, Ubig> = " << test << " Should be the same."<< std::endl;
 

     cout << "========================================"<<endl;
     Vbig.linearCombo(1.,Ubig,1.);
     test = Vbig.innerProd(Vbig) - v3.innerProd(v3);
     std::cout << "norm(Vbig+Ubig)^2 - norm(v2,u3) = " << test << std::endl;
     std::cout << "Should be equal to norm(V+U)^2" << std::endl;


     return 0;
  }
  catch (SEPException &e) {
    e.write(cout);
    exit(EXIT_FAILURE);
  }

}