/** SEP Unit Test 3
    Purpose: test Vector scale function.
*/

#include <iostream>
#include <cstdlib>
#include "../Vector.h"
#include "../inCoreFloat.h"

using namespace std;
using namespace SEP;

int main() {
  try {  
	cerr << "TEST 3: Scale a vector and overwrite self, scale a vector and overwrite another. \n";   

	int n = 20; // set dimension of space 
	float c = 3.0; // amount to scale vector by
	double dblc = double(c); // for use in double computations, same scale
	inCoreFloat1D sp1(n); // create n-D space
		Vector v1(&sp1), v11(&sp1); // create vector from space
	float *array1 = ((dataInCoreFloat*)v1.getDataContainer())->getFloatPtr();
	
/* scale a vector and overwrite the vector with the scaled version */
	double indexDbl = 1.0; // fill in array1 with 1 to n
	for (int i=0; i<n; i++) {
		array1[i] = float(indexDbl);
		indexDbl = indexDbl + 1.0;
	}

	v1.scale(c); // scale v1 by c and overwrite v1 with c*v1
	float *array1now = ((dataInCoreFloat*)v1.getDataContainer())->getFloatPtr();
	float maxDiff = 0.0;
	int maxDiffIn = 0;
	int countOfBigErrors = 0;
 
	for (int i=0; i<n; i++) {
		if(abs((array1now[i] - float(dblc*(double(i)+1.0))))/float(dblc*(double(i)+1.0)) >= c-1.1){
			countOfBigErrors = countOfBigErrors + 1;
		}
		if(abs((array1now[i] - float(dblc*(double(i)+1.0))))/float(dblc*(double(i)+1.0)) >= maxDiff){
			maxDiffIn = i;
			maxDiff = abs((array1now[i]-float(dblc*(double(i)+double(1.0))))/float(dblc*(double(i)+double(1.0))));

//if(maxDiffIn < 10) fprintf(stderr,"Error at %d %f %f \n",i,array1now[i],(float)dblc*((double(i)+double(1.0))));
		}
	}	 

	if(maxDiff > 1e-6) {
		cerr << "ERROR: Vector scaled and overwriting itself did not fill in properly. \n";
		cerr << "Count of big errors " << countOfBigErrors << "\n";
		cerr << "The max relative error was " << maxDiff << "\n";
		cerr << "Occuring in spot " << maxDiffIn << ".\n";
		cerr << "What's actually there is " << array1now[maxDiffIn] << "\n";
		cerr << "What should be there is " << c*float(double(maxDiffIn)+double(1.0)) << "\n";
		cerr << "First ten entries, it actually holds left but should hold  right. \n";
		for (int i=0; i<min(n,50); i++) cerr << array1now[i] << " - " << float(dblc*(double(i)+1)) << "\n";
	}


/* Scale a vector by c again and overwrite the other vector*/
    	v1.scale(c, v11.getDataContainer());
 	float *array11now = ((dataInCoreFloat*)v11.getDataContainer())->getFloatPtr();
	maxDiff = 0.0;
	maxDiffIn = 0;
	for (int i=0; i<n; i++) {
		if(abs((array11now[i] - float(dblc*dblc*(double(i)+1.0))))/float(dblc*(double(i)+1.0)) >= maxDiff){
			maxDiffIn = i;
			maxDiff = abs((array11now[i]-float(dblc*dblc*(double(i)+1.0)))/float(dblc*dblc*(double(i)+1.0)));
		}
	}	
	if(maxDiff > 1e-6) {
		cerr << "ERROR: Vector scaled and overwriting another vector did not fill in properly. \n";
		cerr << "The max relative error was " << maxDiff << "\n";
		cerr << "Occuring in spot " << maxDiffIn << ".\n";
		cerr << "What's actually there is " << array11now[maxDiffIn] << "\n";
		cerr << "What should be there is " << dblc*dblc*float(double(maxDiffIn)+double(1)) << "\n";
		cerr << "First ten entries, it actually holds left but should hold  right. \n";
		for (int i=0; i<min(n,50); i++) cerr << array11now[i] << " - " << dblc*dblc*(double(i)+1.0) << "\n";
	}
   
    	
    



  }
  catch (SEPException &e) {
    e.write(cerr);
    exit(EXIT_FAILURE);
  }
}
