/** SEP Unit Test 4
    Purpose: test Vector addInv function.
*/

#include <iostream>
#include <cstdlib>
#include "../Vector.h"
#include "../inCoreFloat.h"

using namespace std;
using namespace SEP;

int main() {

  try {  
	cout << "TEST 4: Calculate additive inverse of vector and try overwriting the vector itself, then overwriting another vector with the inverse.\n";

	int n = 20000000; // choose dimension of space
    	inCoreFloat1D sp1(n); // create space and two vectors in space
    	Vector v1(&sp1), v11(&sp1);
    	// way of accessing values in vector
    	float *array1 = ((dataInCoreFloat*)v1.getDataContainer())->getFloatPtr();

	
/* test of overwriting a vector with its additive inverse */
	double indexDbl = 1.0;	// create a vector with entries 1 to n
	for (int i=0; i<n; i++){
		array1[i] = float(indexDbl);
		indexDbl = indexDbl + 1.0;
	}
    	// overwrite the vector with itself times -1
    	v1.addInv();
	float *array1now = ((dataInCoreFloat*)v1.getDataContainer())->getFloatPtr();
	int maxDiffIn = 0;
	float maxDiff = 0.0; // calculate max of the difference
	for (int i=0; i<n; i++) {
		if(abs(float(double(array1now[i]) + double(i)+1.0))/abs(float(double(i)+1.0)) >= maxDiff){
			maxDiffIn = i;
			maxDiff = abs(float(double(array1now[i]) + double(i)+1.0))/abs(float(double(i)+1.0));
		}
	}
  	if(maxDiff > 1e-6) {
		cout << "ERROR: Vector was not overwritten with its additive inverse properly in test 4.\n";
		cout << "The maximum relative difference was "<< maxDiff << " and\n";
		cout << " it occured in spot " << maxDiffIn << ".\n";
		cout << "What's actually stored there is " << array1now[maxDiffIn] << "\n";
		cout << "What should be there is " << -1*(double(maxDiffIn)+1) << "\n";		
		cout << "It actually holds (1st ten entries) the vector on the left, but should hold the vector on the right. \n";
		for(int i=0; i<min(n,10); i++) cout << array1now[i] << " , " << -1*(double(i)+1) << "\n"; 
	}  
    
/* test of overwriting another vector with the additive inverse */
    	v1.addInv(v11); // overwrite v11 with -1*v1 = original v1 integers 1 to n
    	float *array11now = ((dataInCoreFloat*)v11.getDataContainer())->getFloatPtr();
	maxDiffIn = 0;
	maxDiff = 0.0; // calculate max of the difference
	for (int i=0; i<n; i++) {
		if(abs(float(double(array11now[i]) - double(i)-1.0))/abs(float(double(i)+1.0)) >= maxDiff){
			maxDiffIn = i;
			maxDiff = abs(float(double(array11now[i]) - double(i)-1.0))/abs(float(double(i)+1.0));
		}
	}
  	if(maxDiff > 1e-6) {
		cout << "ERROR: New vector was not overwritten with other vector's additive inverse properly in test 4.\n";
		cout << "The maximum relative difference was "<< maxDiff << " and\n";
		cout << " it occured in spot " << maxDiffIn << ".\n";
		cout << "What's actually stored there is " << array11now[maxDiffIn] << "\n";
		cout << "What should be there is " << double(maxDiffIn)+1 << "\n";		

		cout << "It actually holds (first ten entries) the vector on the left, but should hold the vector on the right. \n";
		for(int i=0; i<min(n,10); i++) cout << array11now[i] << " , " << double(i)+1 << "\n"; 
	}  
	


  }
  catch (SEPException &e) {
    e.write(cout);
    exit(EXIT_FAILURE);
  }
}
