/** SEP Unit Test 5
    Purpose: test Vector linear combination, scale, addInv functions
    by comparing one-line version and multiple-line version of linear combination.
*/

#include <iostream>
#include <cstdlib>
#include "../Vector.h"
#include "../inCoreFloat.h"

using namespace std;
using namespace SEP;

int main() {

  try {  
	cout << "TEST 5: testing linear combination against scaling after adding (2(a+b)=2a+2b), scaling by -1 against additive inverse.\n";

	int n = 20; // dimension of space to test in

    	inCoreFloat1D oneD(n); // create space
    	Vector v1(&oneD), v2(&oneD); // vectors in oneD
    	// create array1,2,11,22, as a way to access elements of these vectors
    	float *array1 = ((dataInCoreFloat*)v1.getDataContainer())->getFloatPtr();
    	float *array2 =  ((dataInCoreFloat*)v2.getDataContainer())->getFloatPtr();
    	    	    
/* calculate the scale and add versus add then scale and check that these two ways are equal */

	double indexDbl = 1.0; 
	for (int i=0; i<n; i++){
		array1[i] = float(indexDbl); // number array1 as 1 to n
		array2[i] = float(indexDbl+1.0); // array2 as 1 to n
		indexDbl = indexDbl + 1.0;
	}


	// add with linear combo then scale by two
    	Vector v12 = v1.clone(); // v12 = v1
    	v12.linearCombo(1.0,v2,1.0); // v12 = v1+v2
    	v12.scale(2.0); // v12 = 2*(v1+v2) // was times 2
	float *array12 =  ((dataInCoreFloat*)v12.getDataContainer())->getFloatPtr();

	// scale by two and add with linear combo
	v2.linearCombo(2.0, v1, 2.0); // v2 = 2*v2+2*v1
		

	float maxDiff = 0.0; // calculate the sum of the differences in v2 and v12
	int maxDiffIn = 0;
	for(int i=1; i<n; i++){
		if(abs((array12[i]-array2[i])/array2[i]) >= maxDiff) {
			maxDiffIn = i;
			maxDiff = abs((array12[i]-array2[i])/array2[i]);
		}
	}
	if(maxDiff > 1e-6){
		cout << "ERROR: There was a problem showing that 2*v1+2*v2 != 2*(v1+v2) in test 5. \n";
		cout << "The maximum relative difference was "<< maxDiff << " and\n";
		cout << " it occured in spot " << maxDiffIn << ".\n";
		cout << "2*(v1+v2) has " << array12[maxDiffIn] << "\n";
		cout << "2*v1+2*v2 has " << array2[maxDiffIn] << "\n";
		cout << "The first ten entries of these two vectors are: \n";
		for (int i=0; i<min(n,10); i++) cout << array12[i] << " - " << array2[i] << "\n";
	}
	
    	
/* Test scale by -1 versus addInv */
	Vector v11 = v1.clone(); // v11 = v1
	v11.scale(-1.0); // v111 = -v1
	float *array11 =  ((dataInCoreFloat*)v11.getDataContainer())->getFloatPtr();
	v1.addInv(); // v1 = -v1

	maxDiff = 0.0; // calculate the sum of the differences in v11 and v1
	maxDiffIn = 0;
	for(int i=1; i<n; i++){
		if(abs((array11[i]-array1[i])/array1[i]) >= maxDiff) {
			maxDiffIn = i;
			maxDiff = abs((array11[i]-array1[i])/array1[i]);
		}
	}
	if(maxDiff > 1e-6){
		cout << "ERROR: There was a problem showing that addInv(v1) != -1*v1 in test 5. \n";
		cout << "The maximum relative difference was "<< maxDiff << " and\n";
		cout << " it occured in spot " << maxDiffIn << ".\n";
		cout << "v*-1 has " << array11[maxDiffIn] << "\n";
		cout << "addInv(v) has " << array1[maxDiffIn] << "\n";
		cout << "The first ten entries of these two vectors are: \n";
		for (int i=0; i<min(n,10); i++) cout << array11[i] << " - " << array1[i] << "\n";
	}

  }
  catch (SEPException &e) {
    e.write(cout);
    exit(EXIT_FAILURE);
  }
}
