/* SEP Unit Test 6
	Purpose: test the inner product and norm capabilities.
	Written by Eileen Martin, July 2013
*/

#include <iostream>
#include <cstdlib>
#include "../Vector.h"
#include "../inCoreFloat.h"

using namespace std;
using namespace SEP;

int main() {

  try {  
    cout << "Test 6: Take a dot product of a vector with a vector of all ones. Test that the symmetric operation gives the same answer. \n";
    int n = 2000000; // chose to use n dimensional spaces
    inCoreFloat1D sp(n), sp2(n); // create two dimension n vector spaces
    Vector v1(&sp), v11(&sp), v2(&sp2); // three length n vectors in two spaces
	
    float *array1 = ((dataInCoreFloat*)v1.getDataContainer())->getFloatPtr();
    float *array11 = ((dataInCoreFloat*)v11.getDataContainer())->getFloatPtr();
        
    //cout << "Initialize (length n) vector 1 with consecutive integers \n";
    array1[0] = 1.;
    //cout << array1[0] << "\n";
    for (int i=1; i<n; i++) {
	array1[i] = float(1. + double(array1[i-1]));
	//cout << array1[i] << "\n";
	}

    //cout << "Initialize (length n) vector 11 with all twos \n";
    for (int i=0; i<n; i++) {
	array11[i] = 2.;
	}
    
    // take v1 dot v11
    //cout << "Take inner product of vector 1 with vector 11. Should see "<< n << "*(" << n << "+1).\n";
    //cout << "If not, something is wrong with the inner product function. \n";
    //cout << v11.innerProd(v1) << "\n";
    double v1v11 = v11.innerProd(v1);
    double shouldBe = double(n)*(double(n)+1.0);
    double tolerance = 1e-4;
    if(abs((v1v11 - shouldBe)/shouldBe) > tolerance){
	cout << "ERROR: Dot product was not calculated correctly. \n";
	cout << "Dot product should be " << shouldBe << "\n";
	cout << "but was actually calculated to be " << v1v11 << "\n";
    }

    
    // take v11 dot v1	to check the symmetry of the inner product
    //cout << "Take inner product of vector 11 with vector 1. Should see " << n << "*("<< n << "+1)/2. \n";
    //cout << "If not, something is wrong with the inner product function. \n"; 
    //cout << v1.innerProd(v11) << "\n";
    double v11v1 = v1.innerProd(v11);
    if(abs((v11v1-v1v11)/v1v11) > tolerance){
	cout << "Error: Dot product is not symmetric. \n";
	cout << "One direction gives " << v1v11 << "\n";
	cout << "but the other direction gives " << v11v1 << "\n";
    }

  }
  catch (SEPException &e) {
    e.write(cout);
    exit(EXIT_FAILURE);
  }
}
