/** SEP Unit Test 9
    Purpose: test cgstep
*/

#include <iostream>
#include <cstdlib>
#include "../Vector.h"
#include "../inCoreFloat.h"
#include "../DataInCoreFloat.h"
#include "../matmul.h"
#include "../CGStep.h"

using namespace std;
using namespace SEP;

int main() {

  try {  
     cout << "TEST 9: Test cgstep.\n";

     int n1, n2, maxIter;
     n1 = 10;
     n2 = 10;
     maxIter = 500;

     double res=1; // just so we assume a nonzero residual to start

     inCoreFloat2D twoD(n1,n2);
     inCoreFloat1D oneD1(n1);
     inCoreFloat1D oneD2(n2);

     Vector F(&twoD), m_true(&oneD2), m(&oneD2), d(&oneD1), rr(&oneD1), g(&oneD2), gg(&oneD1);

     F.random(); // fill F operator with random entries
     m_true.random(); // random true model
     m.zeroElement(); // start out with a zero guess

     Matmul Fm(&oneD2, &oneD1, &F); // create matrix multiply object
     Fm.Forward(false,&m_true,&d); // true data is F*mtrue
     d.addInv(rr);  // initial residual is F*0 - F*mtrue = -d

     CGStep cg;
     cg.Alloc_step(m,rr); // create and allocate space for cg solver object

     int i=0;
     double resTol = 1e-6;
     while ((i<maxIter) && (res > resTol)) {
       Fm.Adjoint(false,&g,&rr);
       Fm.Forward(false,&g,&gg);
       cg.Step(i, m, g, rr, gg, res);
       i = i+1;
     }
      
      // Check solution error
      Vector diff = m.clone(); // deep copy of m (keeps m from being altered)
      diff.linearCombo(-1.0,m_true,1.0); // diff = m_true-m
      double diffNrmSq = diff.innerProd(diff); // ||m_true - m(original)||^2_2
      diffNrmSq /= m_true.innerProd(m_true); // for relative error sq.
      if(diffNrmSq > resTol){
        cout << "ERROR: Relative error of solution was " << diffNrmSq << std::endl;
	cout << "First " << min(n2,10) << " entries of solution were: \n";
        cout << "true solution,      calculated solution \n";

        dataInCoreFloat *x1= ((dataInCoreFloat*)m_true.getDataContainer());
        dataInCoreFloat *x2= ((dataInCoreFloat*)m.getDataContainer());
        float *array1=x1->getFloatPtr();
        float *array2=x2->getFloatPtr();
        for (int i=0; i< min(n2,10); i++) {
	  cout << array1[i] << "   ,   " << array2[i] << "\n";
        }    
        //delete[] x1, x2, array1, array2;    
      }      


  }
  catch (SEPException &e) {
    e.write(cout);
    exit(EXIT_FAILURE);
  }
}
