/** Kirchhoff migration demo
    Created by Yinbin 2014-08-12
    Last modified:
*/

#include <iostream>
#include <cstdlib>
#include "seplib.h"
#include "../base/Vector.h"
#include "../base/s_reg.h"
#include "../base/hypercube.h"
#include "../base/inCoreFloat.h"
#include "../base/DataInCoreFloat.h"
#include "../base/kirchhoff.h"
#include<stdbool.h>
#include<string>
#include "../base/hypercube.h"
extern "C" {
#include "sep3d.h"
}

using namespace std;
using namespace SEP;

int main(int argc, char *argv[]) {
  cout << "========================================"<<endl;
  cout << "                                        "<<endl;
  cout << "========================================"<<endl;
  cout << "Kirchhoff migration demonstration of the toy model."<<endl;
  initpar(argc,argv);  // Initialize the sep library

  // Preprocessing
  long long nz = 128;
  long long nx = 64;
  float dx = 20.;
  float x0 = 0.;
  float dz = 0.01;
  float z0 = 0.;

  axis da1=axis(nz,z0,dz);
  axis da2=axis(nx,x0,dx);

  std::vector<axis> data_axes;
  data_axes.push_back(da1);
  data_axes.push_back(da2);

  //inCoreFloat2D model(/*"in1"*/model_axes);
  inCoreFloat2D model("in1");
  inCoreFloat2D data(data_axes);

  long long nt=model.get_axis(1).n;
  float     t0=model.get_axis(1).o;
  float     dt=model.get_axis(1).d;
  long long ny=model.get_axis(2).n;
  float     y0=model.get_axis(2).o;
  float     dy=model.get_axis(2).d;

  inCoreFloat1D vel(nz);
  Vector *model_vec = new Vector(&model);
  Vector *data_vec = new Vector(&data);
  Vector *vrms_vec = new Vector(&vel);
  float *__restrict__ model_val = ((dataInCoreFloat*)model_vec->getDataContainer())->getFloatPtr();
  float *__restrict__ data_val = ((dataInCoreFloat*)data_vec->getDataContainer())->getFloatPtr();


  //float *__restrict__ vrms_val = ((dataInCoreFloat*)vrms_vec->getDataContainer())->getFloatPtr();
  sreed("in1",model_val,sizeof(float)*nz*nx);


  Kirchfast2d kirch(&model, &data);
  kirch.SetModelPar(z0,dz,x0,dx);
  kirch.SetDataPar(t0,dt,y0,dy);
  kirch.SetAperture(0.,45.);
  /* 
  For IO test, just use default velocity
  */


  
  // Set up the Kirchhoff migration class

  kirch.Forward(false,model_vec,data_vec);


  data.write_axes("out1");
  srite("out1",data_val,sizeof(float)*nz*nx);


  return 0;

}
