/* Code for the regularized solver, sec. 3.4 of GEE */

#include <iostream>
#include <cstdlib>

#include "../base/DataInCoreFloat.h" // also dataContainer
#include "../base/matmul.h" // also gives Operator.h
#include "../base/CGStep.h" // also gives stepper.h and thus vector.h
#include "../base/deriv2nd1D.h"
#include "../base/lint1.h"

using namespace std;
using namespace SEP;

int main() {
  try{
    cout << "Regularized solver, section 3.4 of GEE \n";

    int nm, nd, nr, forget;
    nm = 6; // dimension of model space (m)
    nd = 3; // dimension of data space (d, Fm)
    nr = 6; // dimension of regularization space (Am)

    double res = 1; // assume nonzero residual to start

    // spaces that model, regularization vector, and data live in
    inCoreFloat1D *modSp = new inCoreFloat1D(nm);
    inCoreFloat1D *regSp = new inCoreFloat1D(nr);
    inCoreFloat1D *dataSp = new inCoreFloat1D(nd);
    // space for long vector of data and regularization to live in
    MultiSpace *dataRegSp = new Multispace(dataSp, regSp);

    Vector m(modSp), d(dataSp), d0(dataRegSp), rr(dataRegSp), g(modSp), gg(modSp);


    // Create [F; A]m
    Lint1 F(modSp,dataSp); // linear interpolation operator
    Deriv2nd1D A(modSp,regSp); // second derivative operator

   // Check that these operators were created correctly
   F.DotTest(true);
   A.DotTest(true);

   // MultiOperator LintDeriv(F,A,'s'); // [F; A]

    // Set d (data) with random values?, put into d0 with vector of zeroes ****
    d.random(); 
    

    // Initialize zero model and residual = -[d; 0]
    m.zeroElement();
    d0.addInv(rr);

    // Create CG solver
    CGStep cg;
    cg.Alloc_step(m, rr);

    // ****will change later, need private s and ss
    Vector s(modSp), ss(dataRegSp);
    s.zeroElement(); 
    ss.zeroElement();

    forget = 1;
    unsigned int i = 0;
    unsigned int maxIter = 5; // ***increase later
    double resTol = 1e-6;
    while((i<maxIter) && (res > resTol)){
      cout << "iter: " << i << "\n"; // ****test only
      FAm.Adjoint(true,&g,&rr);
      FAm.Forward(false,&g,&gg);
      float gdg=g.innerProd(g);
      float ggd=gg.innerProd(gg);
      float ggr=gg.innerProd(rr);
      cout << "gradient dot product: " << gdg << ", " << ggd << ", "<<ggr << "\n"; // ***test only
      float mm=m.innerProd(m);
      cg.Step(i,m,g,rr,gg,res,s,ss);
      cout << "res: " << res << "\n"; // test only
      i = i+1;
    }

    // ******just as test for small example
    dataInCoreFloat *soln = ((dataInCoreFloat*)m.getDataContiner());
    float *array=soln->getFloatPtr();
    std::cout << "Solution is: \n";
    for(unsigned int i=0; i<nm; i++){
      cout << array[i] << "\n";
    }
    // *******end of output that's just for small example test

  } catch (SEPException &e){
    e.write(cout);
    exit(EXIT_FAILURE);
  }
}
