/** SEP wave propogation
    Purpose: 2d wave propogation
    Created by Yinbin 2014-07-21
    Last modified:
*/

#include <iostream>
#include <cstdlib>
#include "seplib.h"
#include "../base/Vector.h"
#include "../base/inCoreFloat.h"
#include "../base/DataInCoreFloat.h"
#include "../base/spraysum.h"
#include<stdbool.h>
#include<string>
using namespace std;
using namespace SEP;

// Function that could inject a source located at (x,z) with amplitude: amp
void InjectSource(Vector *u_vec, long long x, long long z, float amp) {
  float *__restrict__ u_val = ((dataInCoreFloat*)u_vec->getDataContainer())->getFloatPtr();
  inCoreFloat *u_dim=(inCoreFloat*)u_vec->getSpace();
  vector<long long> ns = u_dim->get_ns();
  long long nx = ns[0], nz=ns[1];
  u_val[x*nz + z] += amp;
/*
  for (long long ix = x-2; ix <= x+2; ix++) {
    for (long long iz = z - 2; iz <= z + 2; iz++) {
      float rr = sqrt((float)(ix-x)*(ix-x) + (float)(iz-z)*(iz-z));
      u_val[ix*nz + iz] += amp*0.5*(cos(rr*3.1415926/3) + 1.);
    }
  }
*/
}


// Function that generate velocity field
void InitialVel(Vector *velocity) {
  float *__restrict__ velocity_val = ((dataInCoreFloat*)velocity->getDataContainer())->getFloatPtr();
  inCoreFloat *dim=(inCoreFloat*)velocity->getSpace();
  long long n123 = dim->get_n123();
  vector<long long>ns = dim->get_ns();
  long long nx = ns[0], nz = ns[1];
  for(long long i = 0; i < nx; i++) {
    for(long long j = 0; j < nz; j++) {
      velocity_val[i*nz + j] = 2.2;
      if (i > j) velocity_val[i*nz + j] = 1.8;
    }
  }
}
// Finite difference operator
void FD2nd(Vector *model, Vector *data, Vector *velocity, float s) {
  // This will do finite difference from model, and save to data
  float *__restrict__ model_val = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ data_val = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();
  float *__restrict__ vel_val = ((dataInCoreFloat*)velocity->getDataContainer())->getFloatPtr();
  inCoreFloat *dim=(inCoreFloat*)model->getSpace();
  vector<long long> ns = dim->get_ns();
  long long nx = ns[0], nz=ns[1];


  for (long long ix = 2; ix < nx -2; ix ++) {
    for (long long iz = 2; iz < nz -2; iz ++) {
      float temp = model_val[ix*nz + iz]*(-5.)
                   +(model_val[(ix+1)*nz + iz]+model_val[(ix-1)*nz + iz])*4./3.
                   +(model_val[(ix+2)*nz + iz]+model_val[(ix-2)*nz + iz])*(-1./12.)
                   +(model_val[ix*nz + iz - 1]+model_val[ix*nz + iz + 1])*4./3.
                   +(model_val[ix*nz + iz - 2]+model_val[ix*nz + iz + 2])*(-1./12.);
            data_val[ix*nz + iz] = temp*s*vel_val[ix*nz + iz]*vel_val[ix*nz + iz];
    }
  }


}

void Copying(Vector *ucurr,Vector *output,long long it, long long n123){
  float *__restrict__ model_val = ((dataInCoreFloat*)ucurr->getDataContainer())->getFloatPtr();
  float *__restrict__ data_val = ((dataInCoreFloat*)output->getDataContainer())->getFloatPtr();
  long long shift = it*n123;
  for (long long i = 0; i < n123; i++) {
    data_val[shift + i] = model_val[i];
  }
}

void FreeBound(Vector *u, long long nx, long long nz){
  float *__restrict__ val = ((dataInCoreFloat*)u->getDataContainer())->getFloatPtr();

  // The upper bound
  for(long long iz = 2; iz < nz - 2; iz ++) {
    val[iz] = 1.*val[4*nz + iz];
    val[nz + iz] = 1.*val[3*nz + iz];
    val[(nx-1)*nz + iz] = 1.*val[(nx-5)*nz + iz];
    val[(nx-2)*nz + iz] = 1.*val[(nx-4)*nz + iz];
  }
  for(long long ix = 2; ix < nx - 2; ix ++) {
    val[ix*nz] = -1.*val[ix*nz + 4];
    val[ix*nz + 1] = -1.*val[ix*nz + 3];
    val[ix*nz+ (nz-1)] = -1.*val[ix*nz + (nz-5)];
    val[ix*nz + (nz-2)] = -1.*val[ix*nz + (nz-3)];
  }
}

int main(int argc, char *argv[]) {

  cout << "========================================"<<endl;
  cout << "                                        "<<endl;
  cout << "========================================"<<endl;
  cout << "Toy model for wave equation 2D."<<endl;

  // Prepocessing
  initpar(argc,argv);
  long long nt = 3200;
  long long nsnap = 40;
  long long ngap = nt/nsnap;
  long long nx = 1000;
  long long nz = 1000;
  long long wsize = 64;
  float     ot = 0.;
  float     ox = 0.;
  float     oz = 0.;
  float     dt = 0.0005;
  float     dx = 0.0035;
  float     dz = 0.0035;
  float     s  = dt*dt/(dx*dx);
  float     *amp = new float[wsize];
  inCoreFloat2D wave(nx,nz);
  inCoreFloat3D waves(nsnap,nx,nz);
  Vector *unext = new Vector(&wave);
  Vector *ucurr = new Vector(&wave);
  Vector *uprev = new Vector(&wave);
  Vector *vel   = new Vector(&wave);
  Vector *output = new Vector(&waves);
  uprev->zeroElement(); ucurr->zeroElement(); output->zeroElement();
  InitialVel(vel);
  sreed("in1",amp,sizeof(float)*wsize);


  // Generate random wavelet
  int nwave;
  fetch("in2","d",&nwave);
  cout<< "nwave=" << nwave <<endl;
  long long px[nwave];
  long long pn[nwave];
  long long pt[nwave];
  float pa[nwave];
  for(int i = 0; i < nwave; i++) {
    px[i] = rand() % 960 + 20;
    pn[i] = rand() % 960 + 20;
    pt[i] = 5 + i*100;
    pa[i] = (float)(rand()%100 - 50)/25.;
  }
  px[0] = 400;
  pn[0] = 500;
  pt[0] = 10;
  pa[0] = 1.;
  // Running wave propogation
  long long flag = 0;
  for(long long it = 0; it < nsnap; it++) {
    for (long long i = 0; i < ngap; i++) {
      // Marching in the time domain
      for (long long iter = 0; iter < nwave; iter++) {
        if(flag >=pt[iter] and flag <pt[iter] + wsize){
          InjectSource(ucurr,px[iter],pn[iter],pa[iter]*amp[flag - pt[iter]]);
        }
      }
      flag++;
      FD2nd(ucurr,unext,vel,s);
      unext->linearCombo(1.,ucurr,2.);
      unext->linearCombo(1.,uprev,-1.);
      FreeBound(unext, nx, nz);
      // finishing up
      uprev->linearCombo(0.,ucurr,1.);
      ucurr->linearCombo(0.,unext,1.);
    //ucurr->random();
    }
    Copying(ucurr,output,it,nx*nz);
  }
  cout << "computation done" << endl;
  // Write result to file
  float *__restrict__ data_val = ((dataInCoreFloat*)output->getDataContainer())->getFloatPtr();
  srite("out1",data_val,sizeof(float)*nz*nx*nsnap);

  // cleaning up
  delete unext;
  delete ucurr;
  delete uprev;
  delete vel;
  cout << "hello world!" << endl;
  return 0;
}